---- MODULE MemoirLL1InclCardUniqInvariance ----

(***************************************************************************)
(* This module proves that the InclusionInvariant, the                     *)
(* CardinalityInvariant, and the UniquenessInvariant are all inductive     *)
(* invariants of the `^\basicname^' spec.  Then, from this proof and the   *)
(* previous proof that the UnforgeabilityInvariant is an inductive         *)
(* invariant of the `^\basicname^' spec, it proves that the set of         *)
(* CorrectnessInvariants are inductive invariants of the `^\basicname^'    *)
(* spec.                                                                   *)
(***************************************************************************)

EXTENDS MemoirLL1UnforgeabilityInvariance

(***************************************************************************)
(* These three invariants cannot be ordered with respect to each other.    *)
(* The proof of InclusionInvariant' inductively depends upon               *)
(* UniquenessInvariant; the proof of UniquenessInvariant' inductively      *)
(* depends upon CardinalityInvariant; and the proof of                     *)
(* CardinalityInvariant' inductively depends upon InclusionInvariant.      *)
(* Therefore, we have to prove them co-inductively.                        *)
(***************************************************************************)

THEOREM InclusionCardinalityUniquenessInvariance == \*`^\label{THEOREM-InclusionCardinalityUniquenessInvariance}^'
  LL1Spec => []InclusionInvariant /\ []CardinalityInvariant /\ []UniquenessInvariant
  (*************************************************************************)
  (* This proof will require the LL1TypeInvariant and the                  *)
  (* UnforgeabilityInvariant.  Fortunately, the LL1TypeSafe theorem has    *)
  (* already proven that the `^\basicname^' spec satisfies its type        *)
  (* invariant, and the UnforgeabilityInvariance theorem has already       *)
  (* proven that the `^\basicname^' spec satisfies the                     *)
  (* UnforgeabilityInvariant.                                              *)
  (*************************************************************************)
  <1>1. LL1Spec => []LL1TypeInvariant
    BY LL1TypeSafe
  <1>2. LL1Spec => []UnforgeabilityInvariant
    BY UnforgeabilityInvariance
  (*************************************************************************)
  (* The top level of the proof is boilerplate TLA+ for an Inv1-style      *)
  (* proof.  First, we prove that the initial state satisfies the three    *)
  (* invariants.  Second, we prove that the LL1Next predicate inductively  *)
  (* preserves the three invariants.  Third, we use temporal induction to  *)
  (* prove that these two conditions satisfy the three invariants over all *)
  (* behaviors.                                                            *)
  (*************************************************************************)
  <1>3. LL1Init /\ LL1TypeInvariant /\ UnforgeabilityInvariant =>
        InclusionInvariant /\ CardinalityInvariant /\ UniquenessInvariant
    (***********************************************************************)
    (* First, we assume the antecedents.                                   *)
    (***********************************************************************)
    <2>1. HAVE LL1Init /\ LL1TypeInvariant
    (***********************************************************************)
    (* Then, we pick some symmetrick key for which LL1Init is true.        *)
    (***********************************************************************)
    <2>2. PICK symmetricKey \in SymmetricKeyType : LL1Init!(symmetricKey)!1
      BY <2>1 DEF LL1Init
    (***********************************************************************)
    (* To simplify the writing of the proof, we re-state the definitions   *)
    (* from LL1Init.                                                       *)
    (***********************************************************************)
    <2> initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
    <2> initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
    <2> initialHistoryStateBinding == Hash(BaseHashValue, initialStateHash)
    <2> initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
    <2> initialUntrustedStorage == [
            publicState |-> InitialPublicState,
            privateStateEnc |-> initialPrivateStateEnc,
            historySummary |-> BaseHashValue,
            authenticator |-> initialAuthenticator ]
    <2> initialTrustedStorage == [
            historySummary |-> BaseHashValue,
            symmetricKey |-> symmetricKey ]
    (***********************************************************************)
    (* We then assert the type safety of these definitions, with the help  *)
    (* of the LL1InitDefsTypeSafeLemma.                                    *)
    (***********************************************************************)
    <2>3. /\  initialPrivateStateEnc \in PrivateStateEncType
          /\  initialStateHash \in HashType
          /\  initialHistoryStateBinding \in HashType
          /\  initialAuthenticator \in MACType
          /\  initialUntrustedStorage \in LL1UntrustedStorageType
          /\  initialTrustedStorage \in LL1TrustedStorageType
      <3>1. symmetricKey \in SymmetricKeyType
        BY <2>2
      <3>2. QED
        BY <3>1, LL1InitDefsTypeSafeLemma
    (***********************************************************************)
    (* We hide the definitions, so they don't overwhelm the prover.  We'll *)
    (* pull them in as necessary below.                                    *)
    (***********************************************************************)
    <2> HIDE DEF initialPrivateStateEnc, initialStateHash, initialHistoryStateBinding,
             initialAuthenticator, initialUntrustedStorage, initialTrustedStorage
    (***********************************************************************)
    (* We'll prove each of the three invariants separately.  First, we'll  *)
    (* prove the InclusionInvariant.                                       *)
    (***********************************************************************)
    <2>4. InclusionInvariant
      (*********************************************************************)
      (* To prove the universally quantified expression, we take a new set *)
      (* of variables of the appropriate types.  For the TAKE step to be   *)
      (* meaningful to the prover, first we have to tell the prover to     *)
      (* expand the definition of InclusionInvariant, so it will see the   *)
      (* universally quantified expression therein.                        *)
      (*********************************************************************)
      <3> USE DEF InclusionInvariant
      <3>1. TAKE  input \in InputType,
                  historySummary \in HashType,
                  publicState \in PublicStateType,
                  privateStateEnc \in PrivateStateEncType
      (*********************************************************************)
      (* It suffices to prove that the the antecedent of the implication   *)
      (* is false, which merely requires showing that one of the conjuncts *)
      (* is false.  The proof is straightforward, since the history        *)
      (* summary in the NVRAM equals the base hash value, and the base     *)
      (* hash value cannot be constructed as the hash of any other values. *)
      (*********************************************************************)
      <3>2. SUFFICES ASSUME TRUE PROVE LL1NVRAM.historySummary # Hash(historySummary, input)
        OBVIOUS
      <3>3. LL1NVRAM.historySummary = BaseHashValue
        <4>1. LL1NVRAM = [ historySummary |-> BaseHashValue,
                        symmetricKey |-> symmetricKey ]
          BY <2>2
        <4>2. QED
          BY <4>1
      <3>4. historySummary \in HashDomain
        <4>1. historySummary \in HashType
          BY <3>1
        <4>2. QED
          BY <4>1 DEF HashDomain
      <3>5. input \in HashDomain
        <4>1. input \in InputType
          BY <3>1
        <4>2. QED
          BY <4>1 DEF HashDomain
      <3>6. QED
        BY <3>3, <3>4, <3>5, BaseHashValueUnique
    (***********************************************************************)
    (* Second, we'll prove the CardinalityInvariant.                       *)
    (***********************************************************************)
    <2>5. CardinalityInvariant
      (*********************************************************************)
      (* To prove the universally quantified expression, we take a new set *)
      (* of variables of the appropriate types.  For the TAKE step to be   *)
      (* meaningful to the prover, first we have to tell the prover to     *)
      (* expand the definition of CardinalityInvariant, so it will see the *)
      (* universally quantified expression therein.                        *)
      (*********************************************************************)
      <3> USE DEF CardinalityInvariant
      <3>1. TAKE historySummary \in HashType, stateHash \in HashType
      (*********************************************************************)
      (* To simplify the writing of the proof, we re-state the definition  *)
      (* from the CardinalityInvariant.                                    *)
      (*********************************************************************)
      <3> historyStateBinding == Hash(historySummary, stateHash)
      (*********************************************************************)
      (* We then assert the type safety of this definition, with the help  *)
      (* of the CardinalityInvariantDefsTypeSafeLemma.                     *)
      (*********************************************************************)
      <3>2. historyStateBinding \in HashType
          BY <3>1, CardinalityInvariantDefsTypeSafeLemma
      (*********************************************************************)
      (* To prove the implication, it suffices to assume the antecedent    *)
      (* and prove the consequent.                                         *)
      (*********************************************************************)
      <3>3. SUFFICES
            ASSUME
                /\  LL1NVRAMHistorySummaryUncorrupted
                /\  LL1HistoryStateBindingAuthenticated(historyStateBinding)
            PROVE
                HashCardinality(historySummary) <= HashCardinality(LL1NVRAM.historySummary)
        OBVIOUS
      (*********************************************************************)
      (* We hide the definition, so it doesn't overwhelm the prover.       *)
      (* We'll pull it in as necessary below.                              *)
      (*********************************************************************)
      <3> HIDE DEF historyStateBinding
      (*********************************************************************)
      (* The proof is simple.  First, we prove that the hash cardinality   *)
      (* of the history summary is zero.                                   *)
      (*********************************************************************)
      <3>4. HashCardinality(historySummary) = 0
        (*******************************************************************)
        (* There are two steps to proving that the hash cardinality of the *)
        (* history summary is zero.  First, we use the                     *)
        (* MACCollisionResistant property to prove that the history state  *)
        (* binding matches the initial history state binding defined in    *)
        (* LL1Init.                                                        *)
        (*******************************************************************)
        <4>2. historyStateBinding = initialHistoryStateBinding
          <5>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, initialAuthenticator)
            <6>1. LL1ObservedAuthenticators = {initialAuthenticator}
              BY <2>2 DEF initialAuthenticator, initialHistoryStateBinding,
                          initialStateHash, initialPrivateStateEnc
            <6>2. LL1HistoryStateBindingAuthenticated(historyStateBinding)
              BY <3>3
            <6>3. QED
              BY <6>1, <6>2 DEF LL1HistoryStateBindingAuthenticated
          <5>2. LL1NVRAM.symmetricKey \in SymmetricKeyType
            <6>1. LL1TypeInvariant
              BY <2>1
            <6>2. QED
              BY <6>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>3. symmetricKey \in SymmetricKeyType
            BY <2>2
          <5>4. historyStateBinding \in HashType
            BY <3>2
          <5>5. initialHistoryStateBinding \in HashType
            BY <2>3
          <5>6. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5, MACCollisionResistant
            DEF initialAuthenticator
        (*******************************************************************)
        (* Second, we use the HashCollisionResistant property to prove     *)
        (* that the history summary matches the BaseHashValue, which is    *)
        (* the initial history summary in LL1Init.                         *)
        (*******************************************************************)
        <4>3. historySummary = BaseHashValue 
          <5>1. historySummary \in HashDomain
            <6>1. historySummary \in HashType
              BY <3>1
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>2. stateHash \in HashDomain
            <6>1. stateHash \in HashType
              BY <3>1
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>3. BaseHashValue \in HashDomain
            BY BaseHashValueTypeSafe DEF HashDomain
          <5>4. initialStateHash \in HashDomain
            <6>1. initialStateHash \in HashType
              BY <2>3
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>5. QED
            BY <4>2, <5>1, <5>2, <5>3, <5>4, HashCollisionResistant
            DEF historyStateBinding, initialHistoryStateBinding
        (*******************************************************************)
        (* The conclusion follows directly, because the BaseHashValue has  *)
        (* cardinality zero.                                               *)
        (*******************************************************************)
        <4>4. QED
          BY <4>3, BaseHashCardinalityZero
      (*********************************************************************)
      (* Then, we prove that the hash cardinality of the history summary   *)
      (* in the NVRAM is zero.                                             *)
      (*********************************************************************)
      <3>5. HashCardinality(LL1NVRAM.historySummary) = 0
        <4>1. LL1NVRAM.historySummary = BaseHashValue
          <5>1. LL1NVRAM = [ historySummary |-> BaseHashValue,
                          symmetricKey |-> symmetricKey ]
            BY <2>2
          <5>2. QED
            BY <5>1
        <4>2. QED
          BY <4>1, BaseHashCardinalityZero
      (*********************************************************************)
      (* Since zero is less than or equal to zero, the                     *)
      (* CardinalityInvariant holds.                                       *)
      (*********************************************************************)
      <3>6. QED
        BY <3>4, <3>5
    (***********************************************************************)
    (* Third, we'll prove the UniquenessInvariant.                         *)
    (***********************************************************************)
    <2>6. UniquenessInvariant
      (*********************************************************************)
      (* To prove the universally quantified expression, we take a new set *)
      (* of variables of the appropriate types.  For the TAKE step to be   *)
      (* meaningful to the prover, first we have to tell the prover to     *)
      (* expand the definition of UniquenessInvariant, so it will see the  *)
      (* universally quantified expression therein.                        *)
      (*********************************************************************)
      <3> USE DEF UniquenessInvariant
      <3>1. TAKE stateHash1, stateHash2 \in HashType
      (*********************************************************************)
      (* To simplify the writing of the proof, we re-state the definitions *)
      (* from the UniquenessInvariant.                                     *)
      (*********************************************************************)
      <3> historyStateBinding1 == Hash(LL1NVRAM.historySummary, stateHash1)
      <3> historyStateBinding2 == Hash(LL1NVRAM.historySummary, stateHash2)
      (*********************************************************************)
      (* We then assert the type safety of these definitions, with the     *)
      (* help of the UniquenessInvariantDefsTypeSafeLemma.                 *)
      (*********************************************************************)
      <3>2. /\  historyStateBinding1 \in HashType
            /\  historyStateBinding2 \in HashType
        <4>1. LL1TypeInvariant
          BY <2>1
        <4>2. QED
          BY <3>1, <4>1, UniquenessInvariantDefsTypeSafeLemma
      (*********************************************************************)
      (* To prove the implication, it suffices to assume the antecedent    *)
      (* and prove the consequent.                                         *)
      (*********************************************************************)
      <3>3. SUFFICES
            ASSUME /\  LL1HistoryStateBindingAuthenticated(historyStateBinding1)
                   /\  LL1HistoryStateBindingAuthenticated(historyStateBinding2)
            PROVE stateHash1 = stateHash2
        OBVIOUS
      (*********************************************************************)
      (* We hide the definitions, so they don't overwhelm the prover.      *)
      (* We'll pull them in as necessary below.                            *)
      (*********************************************************************)
      <3> HIDE DEF historyStateBinding1, historyStateBinding2
      (*********************************************************************)
      (* Then, we'll prove that the two history state bindings are equal   *)
      (* to each other, because each one is equal to the initial state     *)
      (* binding.                                                          *)
      (*********************************************************************)
      <3>4. historyStateBinding1 = historyStateBinding2
        (*******************************************************************)
        (* To prove that history state binding 1 equals the initial        *)
        (* history state binding defined in LL1Init, we'll appeal to the   *)
        (* MACCollisionResistant property.                                 *)
        (*******************************************************************)
        <4>1. historyStateBinding1 = initialHistoryStateBinding
          (*****************************************************************)
          (* The main precondition for MACCollisionResistant is that       *)
          (* historyStateBinding1 is validated by an authenticator that    *)
          (* was generated with initialHistoryStateBinding.                *)
          (*****************************************************************)
          <5>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding1, initialAuthenticator)
            <6>1. LL1ObservedAuthenticators = {initialAuthenticator}
              BY <2>2 DEF initialAuthenticator, initialHistoryStateBinding,
                          initialStateHash, initialPrivateStateEnc
            <6>2. LL1HistoryStateBindingAuthenticated(historyStateBinding1)
              BY <3>3
            <6>3. QED
              BY <6>1, <6>2 DEF LL1HistoryStateBindingAuthenticated
          (*****************************************************************)
          (* We also have to prove the appropriate types, which are the    *)
          (* next four statements.                                         *)
          (*****************************************************************)
          <5>2. LL1NVRAM.symmetricKey \in SymmetricKeyType
            <6>1. LL1TypeInvariant
              BY <2>1
            <6>2. QED
              BY <6>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>3. symmetricKey \in SymmetricKeyType
            BY <2>2
          <5>4. historyStateBinding1 \in HashType
            BY <3>2
          <5>5. initialHistoryStateBinding \in HashType
            BY <2>3
          (*****************************************************************)
          (* Now, we can apply the MACCollisionResistant property, which   *)
          (* establishes that history state binding 1 equals the initial   *)
          (* history state binding.                                        *)
          (*****************************************************************)
          <5>6. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5, MACCollisionResistant
            DEF initialAuthenticator
        (*******************************************************************)
        (* To prove that history state binding 2 equals the initial        *)
        (* history state binding defined in LL1Init, we'll appeal to the   *)
        (* MACCollisionResistant property.  This is the exact same logic   *)
        (* we followed above for history state binding 1.                  *)
        (*******************************************************************)
        <4>2. historyStateBinding2 = initialHistoryStateBinding
          (*****************************************************************)
          (* The main precondition for MACCollisionResistant is that       *)
          (* historyStateBinding1 is validated by an authenticator that    *)
          (* was generated with initialHistoryStateBinding.                *)
          (*****************************************************************)
          <5>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding2, initialAuthenticator)
            <6>1. LL1ObservedAuthenticators = {initialAuthenticator}
              BY <2>2 DEF initialAuthenticator, initialHistoryStateBinding,
                          initialStateHash, initialPrivateStateEnc
            <6>2. LL1HistoryStateBindingAuthenticated(historyStateBinding2)
              BY <3>3
            <6>3. QED
              BY <6>1, <6>2 DEF LL1HistoryStateBindingAuthenticated
          (*****************************************************************)
          (* We also have to prove the appropriate types, which are the    *)
          (* next four statements.                                         *)
          (*****************************************************************)
          <5>2. LL1NVRAM.symmetricKey \in SymmetricKeyType
            <6>1. LL1TypeInvariant
              BY <2>1
            <6>2. QED
              BY <6>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>3. symmetricKey \in SymmetricKeyType
            BY <2>2
          <5>4. historyStateBinding2 \in HashType
            BY <3>2
          <5>5. initialHistoryStateBinding \in HashType
            BY <2>3
          (*****************************************************************)
          (* Now, we can apply the MACCollisionResistant property, which   *)
          (* establishes that history state binding 2 equals the initial   *)
          (* history state binding.                                        *)
          (*****************************************************************)
          <5>6. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5, MACCollisionResistant
            DEF initialAuthenticator
        (*******************************************************************)
        (* Because each history state binding is equal to the initial      *)
        (* state binding, the two history state bindings must equal each   *)
        (* other.                                                          *)
        (*******************************************************************)
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* Since the two state bindings are equal, it follows from the       *)
      (* collision resistance of the hash function that the two state      *)
      (* hashes are equal.                                                 *)
      (*********************************************************************)
      <3>5. QED
        <4>1. stateHash1 \in HashDomain
          <5>1. stateHash1 \in HashType
            BY <3>1
          <5>2. QED
            BY <2>1 DEF HashDomain
        <4>2. stateHash2 \in HashDomain
          <5>1. stateHash2 \in HashType
            BY <3>1
          <5>2. QED
            BY <2>1 DEF HashDomain
        <4>3. LL1NVRAM.historySummary \in HashDomain
          <5>1. LL1NVRAM.historySummary \in HashType
            <6>1. LL1TypeInvariant
              BY <2>1
            <6>2. QED
              BY <6>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>4. QED
          BY <3>4, <4>1, <4>2, <4>3, HashCollisionResistant
          DEF historyStateBinding1, historyStateBinding2
    <2>7. QED
      BY <2>4, <2>5, <2>6 DEF CorrectnessInvariants
  (*************************************************************************)
  (* For the induction step, we will need the type invariant and the       *)
  (* unforgeablity invariant to be true in both the unprimed and primed    *)
  (* states.                                                               *)
  (*************************************************************************)
  <1>4. ( /\ InclusionInvariant
          /\ CardinalityInvariant
          /\ UniquenessInvariant
          /\ [LL1Next]_LL1Vars
          /\ LL1TypeInvariant
          /\ LL1TypeInvariant'
          /\ UnforgeabilityInvariant
          /\ UnforgeabilityInvariant' )
      =>
          InclusionInvariant' /\ CardinalityInvariant' /\ UniquenessInvariant'
    (***********************************************************************)
    (* First, we assume the antecedents.                                   *)
    (***********************************************************************)
    <2>1. HAVE /\ InclusionInvariant
               /\ CardinalityInvariant
               /\ UniquenessInvariant
               /\ [LL1Next]_LL1Vars
               /\ LL1TypeInvariant
               /\ LL1TypeInvariant'
               /\ UnforgeabilityInvariant
               /\ UnforgeabilityInvariant'
    (***********************************************************************)
    (* The induction step includes two cases: stuttering and LL1Next       *)
    (* actions.  The stuttering case is a straightforward application of   *)
    (* the InclusionUnchangedLemma, the CardinalityUnchangedLemma, and the *)
    (* UniquenessUnchangedLemma.                                           *)
    (***********************************************************************)
    <2>2. CASE UNCHANGED LL1Vars
      <3>1. UNCHANGED <<LL1NVRAM, LL1ObservedOutputs>>
        BY <2>2 DEF LL1Vars
      <3>2. \A historyStateBinding \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
        <4>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
          BY <2>2 DEF LL1Vars
        <4>2. QED
          BY <4>1, UnchangedAuthenticatedHistoryStateBindingsLemma
      <3>3. InclusionInvariant'
        <4>1. InclusionInvariant /\ LL1TypeInvariant /\  LL1TypeInvariant'
          BY <2>1
        <4>2. QED
          BY <3>1, <3>2, <4>1, InclusionUnchangedLemma
      <3>4. CardinalityInvariant'
        <4>1. CardinalityInvariant /\ LL1TypeInvariant
          BY <2>1
        <4>2. QED
          BY <3>1, <3>2, <4>1, CardinalityUnchangedLemma
      <3>5. UniquenessInvariant'
        <4>1. UniquenessInvariant /\ LL1TypeInvariant
          BY <2>1
        <4>2. QED
          BY <3>1, <3>2, <4>1, UniquenessUnchangedLemma
      <3>6. QED
        BY <3>3, <3>4, <3>5
    (***********************************************************************)
    (* We break down the LL1Next case into separate cases for each action. *)
    (***********************************************************************)
    <2>3. CASE LL1Next
      (*********************************************************************)
      (* The LL1MakeInputAvailable case is a straightforward application   *)
      (* of the InclusionUnchangedLemma, the CardinalityUnchangedLemma,    *)
      (* and the UniquenessUnchangedLemma.                                 *)
      (*********************************************************************)
      <3>1. CASE LL1MakeInputAvailable
        <4>1. UNCHANGED <<LL1NVRAM, LL1ObservedOutputs>>
          BY <3>1 DEF LL1MakeInputAvailable
        <4>2. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>1 DEF LL1MakeInputAvailable
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        <4>3. InclusionInvariant'
          <5>1. InclusionInvariant /\ LL1TypeInvariant /\  LL1TypeInvariant'
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, InclusionUnchangedLemma
        <4>4. CardinalityInvariant'
          <5>1. CardinalityInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, CardinalityUnchangedLemma
        <4>5. UniquenessInvariant'
          <5>1. UniquenessInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, UniquenessUnchangedLemma
        <4>6. QED
          BY <4>3, <4>4, <4>5
      (*********************************************************************)
      (* The LL1PerformOperation case is the vast majority of this proof.  *)
      (*********************************************************************)
      <3>2. CASE LL1PerformOperation
        (*******************************************************************)
        (* We pick some input for which LL1PerformOperation is true.       *)
        (*******************************************************************)
        <4>1. PICK input1 \in LL1AvailableInputs : LL1PerformOperation!(input1)!1 
          BY <3>2 DEF LL1PerformOperation
        (*******************************************************************)
        (* To simplify the writing of the proof, we re-state the           *)
        (* definitions from the LL1PerformOperation action.                *)
        (*******************************************************************)
        <4> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
        <4> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
        <4> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
        <4> sResult == Service(LL1RAM.publicState, privateState, input1)
        <4> newPrivateStateEnc ==
                SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
        <4> newHistorySummary == Hash(LL1NVRAM.historySummary, input1)
        <4> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
        <4> newHistoryStateBinding == Hash(newHistorySummary, newStateHash)
        <4> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
       (********************************************************************)
       (* We then assert the type safety of these definitions, with the    *)
       (* help of the LL1PerformOperationDefsTypeSafeLemma.                *)
       (********************************************************************)
        <4>2. /\  stateHash \in HashType
              /\  historyStateBinding \in HashType
              /\  privateState \in PrivateStateType
              /\  sResult \in ServiceResultType
              /\  sResult.newPublicState \in PublicStateType
              /\  sResult.newPrivateState \in PrivateStateType
              /\  sResult.output \in OutputType
              /\  newPrivateStateEnc \in PrivateStateEncType
              /\  newHistorySummary \in HashType
              /\  newStateHash \in HashType
              /\  newHistoryStateBinding \in HashType
              /\  newAuthenticator \in MACType
          <5>1. input1 \in LL1AvailableInputs
            BY <4>1
          <5>2. LL1TypeInvariant
            BY <2>1
          <5>3. QED
            BY <5>1, <5>2, LL1PerformOperationDefsTypeSafeLemma
        (*******************************************************************)
        (* We hide the definitions, so they don't overwhelm the prover.    *)
        (* We'll pull them in as necessary below.                          *)
        (*******************************************************************)
        <4> HIDE DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
                newHistorySummary, newStateHash, newHistoryStateBinding, newAuthenticator
        (*******************************************************************)
        (* We prove some simple type statements up front, to avoid needing *)
        (* to repeat these multiple times below.                           *)
        (*******************************************************************)
        <4>3. /\  LL1NVRAM.historySummary \in HashType
              /\  LL1NVRAM.symmetricKey \in SymmetricKeyType
          <5>1. LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <5>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
        <4>4. /\  LL1NVRAM.historySummary' \in HashType
              /\  LL1NVRAM.symmetricKey' \in SymmetricKeyType
          <5>1. LL1TypeInvariant'
            BY <2>1
          <5>2. QED
            BY <5>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
        (*******************************************************************)
        (* We also prove that LL1NVRAMHistorySummaryUncorrupted predicate  *)
        (* is true.  This follows from the enablement conditions for       *)
        (* LL1PerformOperation.                                            *)
        (*******************************************************************)
        <4>5. LL1NVRAMHistorySummaryUncorrupted
          (*****************************************************************)
          (* We prove this directly from the definition of the             *)
          (* LL1NVRAMHistorySummaryUncorrupted predicate.  There is one    *)
          (* type assumption and one antecedent in the implication.        *)
          (*****************************************************************)
          <5>1. stateHash \in HashType
            BY <4>2
          <5>2. LL1HistoryStateBindingAuthenticated(historyStateBinding)
            (***************************************************************)
            (* The authenticator in the RAM is a witness, because          *)
            (* TMPPerformOperation ensures that it validates the history   *)
            (* state binding.                                              *)
            (***************************************************************)
            <6>1. ValidateMAC(
                      LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
              BY <4>1 DEF historyStateBinding, stateHash
            (***************************************************************)
            (* The UnforgeabilityInvariant then ensures that this          *)
            (* authenticator is in the set of observed authenticators.     *)
            (***************************************************************)
            <6>2. LL1RAM.authenticator \in LL1ObservedAuthenticators
              <7>1. UnforgeabilityInvariant
                BY <2>1
              <7>2. historyStateBinding \in HashType
                BY <4>2
              <7>3. QED
                BY <6>1, <7>1, <7>2 DEF UnforgeabilityInvariant, LL1HistoryStateBindingAuthenticated
            (***************************************************************)
            (* These two conditions satisfy the definition of              *)
            (* LL1HistoryStateBindingAuthenticated.                        *)
            (***************************************************************)
            <6>3. QED
              BY <6>1, <6>2 DEF LL1HistoryStateBindingAuthenticated
          <5>3. LL1NVRAM.historySummary = LL1RAM.historySummary
            BY <4>1
          <5>4. QED
            BY <5>1, <5>2, <5>3 DEF LL1NVRAMHistorySummaryUncorrupted, historyStateBinding 
        (*******************************************************************)
        (* We'll prove each of the three invariants separately.  First,    *)
        (* we'll prove that the InclusionInvariant holds in the primed     *)
        (* state.                                                          *)
        (*******************************************************************)
        <4>6. InclusionInvariant'
          (*****************************************************************)
          (* To prove the universally quantified expression, we take a new *)
          (* set of variables of the appropriate types.  For the TAKE step *)
          (* to be meaningful to the prover, first we have to tell the     *)
          (* prover to expand the definition of InclusionInvariant, so it  *)
          (* will see the universally quantified expression therein.       *)
          (*****************************************************************)
          <5> USE DEF InclusionInvariant
          <5>1. TAKE input2 \in InputType,
                     historySummary \in HashType,
                     publicState \in PublicStateType,
                     privateStateEnc \in PrivateStateEncType
          (*****************************************************************)
          (* To simplify the writing of the proof, we re-state the         *)
          (* definitions from the InclusionInvariant.                      *)
          (*****************************************************************)
          <5> inclStateHash == Hash(publicState, privateStateEnc)
          <5> inclHistoryStateBinding == Hash(historySummary, inclStateHash)
          <5> inclPrivateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, privateStateEnc)
          <5> inclSResult == Service(publicState, inclPrivateState, input2)
          <5> inclNewPrivateStateEnc ==
                  SymmetricEncrypt(LL1NVRAM.symmetricKey, inclSResult.newPrivateState)
          <5> inclNewStateHash == Hash(inclSResult.newPublicState, inclNewPrivateStateEnc)
          <5> inclNewHistoryStateBinding == Hash(LL1NVRAM.historySummary, inclNewStateHash)
          (*****************************************************************)
          (* We then assert the type safety of these definitions, with the *)
          (* help of the InclusionInvariantDefsTypeSafeLemma.              *)
          (*****************************************************************)
          <5>2. /\  inclStateHash \in HashType
                /\  inclHistoryStateBinding \in HashType
                /\  inclPrivateState \in PrivateStateType
                /\  inclSResult \in ServiceResultType
                /\  inclSResult.newPublicState \in PublicStateType
                /\  inclSResult.newPrivateState \in PrivateStateType
                /\  inclSResult.output \in OutputType
                /\  inclNewPrivateStateEnc \in PrivateStateEncType
                /\  inclNewStateHash \in HashType
                /\  inclNewHistoryStateBinding \in HashType
            <6>1. LL1TypeInvariant
              BY <2>1
            <6>2. QED
              BY <5>1, <6>1, InclusionInvariantDefsTypeSafeLemma
          (*****************************************************************)
          (* The InclusionInvariant states an implication.  To prove this, *)
          (* it suffices to assume the antecedent and prove the            *)
          (* consequent.                                                   *)
          (*****************************************************************)
          <5>3. SUFFICES
                ASSUME
                    /\  LL1NVRAM.historySummary' = Hash(historySummary, input2)
                    /\  LL1HistoryStateBindingAuthenticated(inclHistoryStateBinding)'
                PROVE
                    /\  inclSResult.output' \in LL1ObservedOutputs'
                    /\  LL1HistoryStateBindingAuthenticated(inclNewHistoryStateBinding)'
            OBVIOUS
          (*****************************************************************)
          (* We hide the definition of InclusionInvariant and the          *)
          (* definitions from the InclusionInvariant.                      *)
          (*****************************************************************)
          <5> HIDE DEF InclusionInvariant
          <5> HIDE DEF inclStateHash, inclHistoryStateBinding, inclPrivateState, inclSResult,
                  inclNewPrivateStateEnc, inclNewStateHash, inclNewHistoryStateBinding
          (*****************************************************************)
          (* Starting the InclusionInvariant proof proper, there are four  *)
          (* main steps.                                                   *)
          (*                                                               *)
          (* First, we prove that the history summary in the NVRAM matches *)
          (* the history summary that satisfies the antecedent condition,  *)
          (* and that the input supplied to the LL1PerformOperation action *)
          (* matches the input that satisfies the antecedent condition.    *)
          (*                                                               *)
          (* Second, we prove that the public and private state in the RAM *)
          (* matches the public and private state that satisfies the       *)
          (* antecedent condition.                                         *)
          (*                                                               *)
          (* The third and fourth steps each use the above results to      *)
          (* prove one of the conjuncts in the consequent of the           *)
          (* InclusionInvariant.                                           *)
          (*****************************************************************)
          <5>4. /\  LL1NVRAM.historySummary = historySummary
                /\  input1 = input2
            (***************************************************************)
            (* To prove the above equivalences, we will use the            *)
            (* HashCollisionResistant property.  To use this, we first     *)
            (* have to prove some simple types.                            *)
            (***************************************************************)
            <6>1. LL1NVRAM.historySummary \in HashDomain
              <7>1. LL1NVRAM.historySummary \in HashType
                BY <4>3
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>2. input1 \in HashDomain
              <7>1. input1 \in InputType
                <8>1. input1 \in LL1AvailableInputs
                  BY <4>1
                <8>2. LL1AvailableInputs \subseteq InputType
                  <9>1. LL1TypeInvariant
                    BY <2>1
                  <9>2. QED
                    BY <9>1 DEF LL1TypeInvariant
                <8>3. QED
                  BY <8>1, <8>2
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>3. historySummary \in HashDomain
              <7>1. historySummary \in HashType
                BY <5>1
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>4. input2 \in HashDomain
              <7>1. input2 \in InputType
                BY <5>1
              <7>2. QED
                BY <7>1 DEF HashDomain
            (***************************************************************)
            (* Then, we have to prove that the hashes are equal, which     *)
            (* follows from the definition of the new history summary      *)
            (* produced by LL1PerformOperation.                            *)
            (***************************************************************)
            <6>5. Hash(LL1NVRAM.historySummary, input1) = Hash(historySummary, input2)
              <7>1. newHistorySummary = Hash(LL1NVRAM.historySummary, input1)
                BY DEF newHistorySummary
              <7>2. LL1NVRAM.historySummary' = Hash(historySummary, input2)
                BY <5>3
              <7>3. LL1NVRAM.historySummary' = newHistorySummary
                <8>1. LL1NVRAM' = [ historySummary |-> newHistorySummary,
                                 symmetricKey |-> LL1NVRAM.symmetricKey ]
                  BY <4>1 DEF newHistorySummary, newPrivateStateEnc, sResult, privateState
                <8>2. QED
                  BY <8>1
              <7>4. QED
                BY <7>1, <7>2, <7>3
            <6>6. QED
              (*************************************************************)
              (* Ideally, this QED step should just read:                  *)
              (*                                                           *)
              (*  BY <<6>>1, <<6>>2, <<6>>3, <<6>>4, <<6>>5, HashCollisionResistant *)
              (*                                                           *)
              (* However, the prover seems to get a little confused in     *)
              (* this instance.  We make life easier for the prover by     *)
              (* definining some local variables and hiding their          *)
              (* definitions before appealing to the                       *)
              (* HashCollisionResistant assumption.                        *)
              (*************************************************************)
              <7> h1a == LL1NVRAM.historySummary
              <7> h2a == input1
              <7> h1b == historySummary
              <7> h2b == input2
              <7>1. h1a \in HashDomain
                BY <6>1
              <7>2. h2a \in HashDomain
                BY <6>2
              <7>3. h1b \in HashDomain
                BY <6>3
              <7>4. h2b \in HashDomain
                BY <6>4
              <7>5. Hash(h1a, h2a) = Hash(h1b, h2b)
                BY <6>5
              <7>6. h1a = h1b /\ h2a = h2b
                <8> HIDE DEF h1a, h2a, h1b, h2b
                <8>1. QED
                  BY <7>1, <7>2, <7>3, <7>4, <7>5, HashCollisionResistant
              <7>7. QED
                BY <7>6
          (*****************************************************************)
          (* The second main step in the InclusionInvariant proof is to    *)
          (* prove that the public and private state in the RAM matches    *)
          (* the public and private state that satisfies the antecedent    *)
          (* condition.                                                    *)
          (*****************************************************************)
          <5>5. /\  publicState = LL1RAM.publicState
                /\  inclPrivateState' = privateState
            (***************************************************************)
            (* Most of the work of proving this step is proving that the   *)
            (* public state in the RAM matches the public state that       *)
            (* satisfies the antecedent condition, and that the encrypted  *)
            (* private state in the RAM matches the encrypted private      *)
            (* state that satisfies the antecedent condition.              *)
            (***************************************************************)
            <6>1. /\ publicState = LL1RAM.publicState
                  /\ privateStateEnc = LL1RAM.privateStateEnc
              (*************************************************************)
              (* To prove the above equivalences, we will use the          *)
              (* HashCollisionResistant property.  To use this, we first   *)
              (* have to prove some simple types.                          *)
              (*************************************************************)
              <7>1. publicState \in HashDomain
                <8>1. publicState \in PublicStateType
                  BY <2>1 DEF LL1Refinement
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>2. /\  LL1RAM.publicState \in HashDomain
                    /\  LL1RAM.privateStateEnc \in HashDomain
                <8>1. /\  LL1RAM.publicState \in PublicStateType
                      /\  LL1RAM.privateStateEnc \in PrivateStateEncType
                  <9>1. LL1TypeInvariant
                    BY <2>1
                  <9>2. QED
                    BY <9>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>3. privateStateEnc \in HashDomain
                <8>1. privateStateEnc \in PrivateStateEncType
                  BY <5>1
                <8>2. QED
                  BY <8>1 DEF HashDomain
              (*************************************************************)
              (* Then, we'll show that the state hash defined in the       *)
              (* InclusionInvariant matches the state hash defined in      *)
              (* LL1PerformOperation.  This is the hash equivalence we     *)
              (* will use in our appeal to the HashCollisionResistant      *)
              (* property.                                                 *)
              (*                                                           *)
              (* This next step is quite involved, so the remainder of     *)
              (* this proof follows much further down.                     *)
              (*************************************************************)
              <7>4. inclStateHash' = stateHash
                (***********************************************************)
                (* We can show that the state hashes match using the       *)
                (* UniquenessInvariant.  This will require proving some    *)
                (* simple type statements, immediately below, and then     *)
                (* proving that the two history state bindings             *)
                (* corresponding to the two hashes are both authenticated. *)
                (***********************************************************)
                <8>1. inclStateHash' \in HashType
                  <9>1. publicState \in HashDomain
                    <10>1. publicState \in PublicStateType
                      BY <5>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>2. privateStateEnc \in HashDomain
                    <10>1. privateStateEnc \in PrivateStateEncType
                      BY <5>1
                    <10>2. QED
                      BY <10>1 DEF HashDomain
                  <9>3. QED
                    BY <9>1, <9>2, HashTypeSafe DEF inclStateHash
                <8>2. stateHash \in HashType
                  BY <4>2
                (***********************************************************)
                (* The first history state binding we have to prove to be  *)
                (* authenticated is the history state binding defined in   *)
                (* the InclusionInvariant.  This is by far the more        *)
                (* involved of the two.                                    *)
                (*                                                         *)
                (* Our strategy is to first prove that this history state  *)
                (* binding is authenticated in the primed state.  Then, we *)
                (* prove that this history state binding is not validated  *)
                (* by the new authenticator defined in the                 *)
                (* LL1PerformOperation action.  Since this new             *)
                (* authenticator is the only element that is in the primed *)
                (* set of observed authenticators but not in the unprimed  *)
                (* set of observed authenticators, it follows that the the *)
                (* history state binding is authenticated by an            *)
                (* authenticator in the unprimed set of authenticators, so *)
                (* it is authenticated in the unprimed state.              *)
                (***********************************************************)
                <8>3. LL1HistoryStateBindingAuthenticated(inclHistoryStateBinding)
                  (*********************************************************)
                  (* The history state binding defined by the              *)
                  (* InclusionInvariant is authenticated in the primed     *)
                  (* state by hypothesis.                                  *)
                  (*********************************************************)
                  <9>1. LL1HistoryStateBindingAuthenticated(inclHistoryStateBinding)'
                    BY <5>3
                  (*********************************************************)
                  (* We have to prove that the history state binding       *)
                  (* defined by the InclusionInvariant is not              *)
                  (* authenticated by the new authenticator defined in the *)
                  (* LL1PerformOperation action.                           *)
                  (*********************************************************)
                  <9>2. ~ValidateMAC(
                            LL1NVRAM.symmetricKey,
                            inclHistoryStateBinding,
                            newAuthenticator)
                    (*******************************************************)
                    (* Our strategy is to use proof by contradiction.      *)
                    (* First, we will show that if the history state       *)
                    (* binding defined by the InclusionInvariant were      *)
                    (* authenticated by the new authenticator defined in   *)
                    (* the LL1PerformOperation action, then this history   *)
                    (* state binding would equal the new history state     *)
                    (* binding defined in the LL1PerformOperation action.  *)
                    (* This then leads to the conclusion that the two      *)
                    (* history summaries in each of these history state    *)
                    (* bindings are equal.  However, we will show that     *)
                    (* these two history summaries cannot be equal,        *)
                    (* because one was generated from a hash of the other. *)
                    (*******************************************************)
                    <10>1. SUFFICES
                           ASSUME
                            ValidateMAC(
                                LL1NVRAM.symmetricKey,
                                inclHistoryStateBinding,
                                newAuthenticator)
                           PROVE FALSE
                      OBVIOUS
                    (*******************************************************)
                    (* In our contradictory universe, the history state    *)
                    (* binding defined by the InclusionInvariant equals    *)
                    (* the new history state binding defined in the        *)
                    (* LL1PerformOperation action, due to the collision    *)
                    (* resistance of the MAC.  We merely need to prove     *)
                    (* some types, and then we can employ the              *)
                    (* MACCollisionResistant property directly.            *)
                    (*******************************************************)
                    <10>2. inclHistoryStateBinding = newHistoryStateBinding
                      <11>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
                        BY <4>3
                      <11>2. inclHistoryStateBinding \in HashType
                        BY <5>2
                      <11>3. newHistoryStateBinding \in HashType
                        BY <4>2
                      <11>4. QED
                        BY <10>1, <11>1, <11>2, <11>3, MACCollisionResistant
                        DEF newAuthenticator
                    (*******************************************************)
                    (* In our contradictory universe, the history summary  *)
                    (* quantified by the InclusionInvariant equals the new *)
                    (* history summary defined in the LL1PerformOperation  *)
                    (* action, due to the collision resistance of the hash *)
                    (* function.  We merely need to prove some types, and  *)
                    (* then we can employ the HashCollisionResistant       *)
                    (* property directly.                                  *)
                    (*******************************************************)
                    <10>3. historySummary = newHistorySummary
                      <11>1. historySummary \in HashDomain
                        <12>1. historySummary \in HashType
                          BY <5>3
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>2. inclStateHash \in HashDomain
                        <12>1. inclStateHash \in HashType
                          BY <5>2
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>3. newHistorySummary \in HashDomain
                        <12>1. newHistorySummary \in HashType
                          BY <4>2
                        <12>2. QED
                         BY <12>1 DEF HashDomain
                      <11>4. newStateHash \in HashDomain
                        <12>1. newStateHash \in HashType
                          BY <4>2
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>5. QED
                        BY <10>2, <11>1, <11>2, <11>3, <11>4, HashCollisionResistant
                        DEF inclHistoryStateBinding, newHistoryStateBinding
                    (*******************************************************)
                    (* Back in the real universe, we will prove that the   *)
                    (* history summary quantified by the                   *)
                    (* InclusionInvariant cannot equal the new history     *)
                    (* summary defined in the LL1PerformOperation action.  *)
                    (* The proof relies on the property of hash            *)
                    (* cardinality, though not on the                      *)
                    (* CardinalityInvariant.  Basically, we prove that the *)
                    (* cardinality of the history summary in the new       *)
                    (* authenticator defined by LL1PerformOperation is one *)
                    (* greater than the hash cardinality of the history    *)
                    (* summary previously in the NVRAM.  Since the hash    *)
                    (* cardinalities differ, it follows that the history   *)
                    (* summaries differ.  The proof is tedious but         *)
                    (* straightforward.                                    *)
                    (*******************************************************)
                    <10>4. historySummary # newHistorySummary
                      (*****************************************************)
                      (* First, we prove a bunch of types that are needed  *)
                      (* by the hash cardinality assumptions or for        *)
                      (* proving basic arithmetic.                         *)
                      (*****************************************************)
                      <11>1. input1 \in InputType
                        <12>1. input1 \in LL1AvailableInputs
                          BY <4>1
                        <12>2. LL1AvailableInputs \subseteq InputType
                          <13>1. LL1TypeInvariant
                            BY <2>1
                          <13>2. QED
                            BY <13>1 DEF LL1TypeInvariant
                        <12>3. QED
                          BY <12>1, <12>2
                      <11>2. input1 \in HashDomain
                        BY <11>1 DEF HashDomain
                      <11>3. LL1NVRAM.historySummary \in HashDomain
                        <12>1. LL1NVRAM.historySummary \in HashType
                          BY <4>3
                        <12>2. QED
                          BY <12>1 DEF HashDomain
                      <11>4. HashCardinality(input1) \in Nat
                        BY <11>2, HashCardinalityTypeSafe
                      <11>5. HashCardinality(newHistorySummary) \in Nat
                        <12>1. newHistorySummary \in HashDomain
                          <13>1. newHistorySummary \in HashType
                            BY <4>2
                          <13>2. QED
                            BY <13>1 DEF HashDomain
                        <12>2. QED
                         BY <12>1, HashCardinalityTypeSafe
                      <11>6. HashCardinality(historySummary) \in Nat
                        <12>1. historySummary \in HashDomain
                          <13>1. historySummary \in HashType
                            BY <5>1
                          <13>2. QED
                            BY <13>1 DEF HashDomain
                        <12>2. QED
                          BY <12>1, HashCardinalityTypeSafe
                      (*****************************************************)
                      (* With the type statements out of the way, we can   *)
                      (* construct a simple inequality.  We do this in     *)
                      (* four linear steps.                                *)
                      (*****************************************************)
                      <11>7. HashCardinality(newHistorySummary) =
                                 HashCardinality(LL1NVRAM.historySummary) + HashCardinality(input1) + 1
                        BY <11>2, <11>3, HashCardinalityAccumulative DEF newHistorySummary
                      <11>8. HashCardinality(newHistorySummary) =
                                 HashCardinality(historySummary) + HashCardinality(input1) + 1
                        <12>1. LL1NVRAM.historySummary = historySummary
                          BY <5>4
                        <12>2. QED
                          BY <11>7, <12>1
                      <11>9. HashCardinality(newHistorySummary) = HashCardinality(historySummary) + 1
                        <12>1. HashCardinality(input1) = 0
                            BY <11>1, InputCardinalityZero
                        <12>2. QED
                            BY <11>5, <11>6, <11>8, <12>1
                      <11>10. HashCardinality(newHistorySummary) # HashCardinality(historySummary)
                        BY <11>5, <11>6, <11>9
                      (*****************************************************)
                      (* Since the hash cardinalities differ, it follows   *)
                      (* that the history summaries differ.                 *)
                      (*****************************************************)
                      <11>11. QED
                        BY <11>10
                    (*******************************************************)
                    (* The required contradiction follows from the         *)
                    (* previous two steps.                                 *)
                    (*******************************************************)
                    <10>5. QED
                      BY <10>3, <10>4
                  (*********************************************************)
                  (* Before completing the proof, we need to establish     *)
                  (* that the symmetric key in the NVRAM has not changed,  *)
                  (* because the LL1HistoryStateBindingAuthenticated       *)
                  (* predicate implicitly refers to this variable.         *)
                  (*********************************************************)
                  <9>3. UNCHANGED LL1NVRAM.symmetricKey
                    BY <2>1, SymmetricKeyConstantLemma
                  (*********************************************************)
                  (* We also need to establish that the new authenticator  *)
                  (* defined by the LL1PerformOperation action is the only *)
                  (* element that is in the primed set of observed         *)
                  (* authenticators but not in the unprimed set of         *)
                  (* observed authenticators.                              *)
                  (*********************************************************)
                  <9>4. LL1ObservedAuthenticators' =
                                 LL1ObservedAuthenticators \union {newAuthenticator}
                    BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                                newHistorySummary, newPrivateStateEnc, sResult, privateState
                  (*********************************************************)
                  (* The conclusion follows directly.                      *)
                  (*********************************************************)
                  <9>5. QED
                    BY <9>1, <9>2, <9>3, <9>4 DEF LL1HistoryStateBindingAuthenticated
                (***********************************************************)
                (* The second history state binding we have to prove to be *)
                (* authenticated is the history state binding defined in   *)
                (* the LL1PerformOperation action.  We need to show that   *)
                (* it is authenticated by some authenticator in the set of *)
                (* observed authenticators.                                *)
                (***********************************************************)
                <8>4. LL1HistoryStateBindingAuthenticated(historyStateBinding)
                  (*********************************************************)
                  (* The authenticator in the RAM is a witness, because    *)
                  (* TMPPerformOperation ensures that it validates the     *)
                  (* history state binding.                                *)
                  (*********************************************************)
                  <9>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
                    BY <4>1 DEF historyStateBinding, stateHash
                  (*********************************************************)
                  (* The UnforgeabilityInvariant then ensures that this    *)
                  (* authenticator is in the set of observed               *)
                  (* authenticators.                                       *)
                  (*********************************************************)
                  <9>2. LL1RAM.authenticator \in LL1ObservedAuthenticators
                    <10>1. UnforgeabilityInvariant
                      BY <2>1
                    <10>2. historyStateBinding \in HashType
                      BY <4>2
                    <10>3. QED
                      BY <9>1, <10>1, <10>2 DEF UnforgeabilityInvariant, LL1HistoryStateBindingAuthenticated
                  (*********************************************************)
                  (* These two conditions satisfy the definition of        *)
                  (* LL1HistoryStateBindingAuthenticated.                  *)
                  (*********************************************************)
                  <9>3. QED
                    BY <9>1, <9>2 DEF LL1HistoryStateBindingAuthenticated
                (***********************************************************)
                (* The conclusion follows fairly directly from the         *)
                (* UniquenessInvariant.  However, one small hitch is that  *)
                (* the UniquenessInvariant requires both history state     *)
                (* bindings to be defined using LL1NVRAM.historySummary,   *)
                (* but inclHistoryStateBinding is defined using            *)
                (* historySummary rather than LL1NVRAM.historySummary.     *)
                (* Therefore, we have to add in the fact that              *)
                (* LL1NVRAM.historySummary = historySummary.               *)
                (***********************************************************)
                <8>5. QED
                  <9>1. UniquenessInvariant
                    BY <2>1
                  <9>2. LL1NVRAM.historySummary = historySummary
                    BY <5>4
                  <9>3. LL1NVRAM.historySummary = LL1RAM.historySummary
                    BY <4>1
                  <9>4. QED
                    BY <8>1, <8>2, <8>3, <8>4, <9>1, <9>2, <9>3
                    DEF UniquenessInvariant, inclHistoryStateBinding, historyStateBinding
              (*************************************************************)
              (* The conclusion follows directly from applying the         *)
              (* HashCollisionResistant property.                          *)
              (*************************************************************)
              <7>5. QED
                BY <7>1, <7>2, <7>3, <7>4, HashCollisionResistant DEF inclStateHash, stateHash
            (***************************************************************)
            (* The remainder of this step follows trivially.               *)
            (***************************************************************)
            <6>2. publicState = LL1RAM.publicState
              BY <6>1
            <6>3. inclPrivateState' = privateState
              <7>1. privateStateEnc = LL1RAM.privateStateEnc
                BY <6>1
              <7>2. UNCHANGED LL1NVRAM.symmetricKey
                BY <2>1, SymmetricKeyConstantLemma
              <7>3. QED
                BY <7>1, <7>2 DEF inclPrivateState, privateState
            <6>4. QED
              BY <6>2, <6>3
          (*****************************************************************)
          (* The third main step in the InclusionInvariant proof is to     *)
          (* prove the first conjunct in the consequent of the             *)
          (* InclusionInvariant, namely that the primed output of the      *)
          (* service is in the primed set of observed outputs.             *)
          (*                                                               *)
          (* This follows fairly directly from the just-proven facts that  *)
          (* the variables that satisfy the antecedent conditions match    *)
          (* the corresponding values in the NVRAM and the input value     *)
          (* provided to the LL1PerformOperation action.                   *)
          (*****************************************************************)
          <5>6. inclSResult'.output \in LL1ObservedOutputs'
            <6>1. LL1ObservedOutputs' = LL1ObservedOutputs \union {inclSResult'.output}
              <7>1. LL1ObservedOutputs' = LL1ObservedOutputs \union {sResult.output}
                BY <4>1 DEF sResult, privateState
              <7>2. inclSResult'.output = sResult.output
                <8>1. inclSResult' = sResult
                  <9>1. /\ publicState = LL1RAM.publicState
                        /\ inclPrivateState' = privateState
                    BY <5>5
                  <9>2. input2 = input1
                    BY <5>4
                  <9>3. QED
                    BY <9>1, <9>2 DEF inclSResult, sResult
                <8>2. QED
                  BY <8>1
              <7>3. QED
                BY <7>1, <7>2
            <6>2. QED
              BY <6>1
          (*****************************************************************)
          (* The fourth main step in the InclusionInvariant proof is to    *)
          (* prove the second conjunct in the consequent of the            *)
          (* InclusionInvariant, namely that the new history state binding *)
          (* defined in the InclusionInvariant is authenticated in the     *)
          (* primed state.                                                 *)
          (*                                                               *)
          (* We need to show that there exists some authenticator in the   *)
          (* primed set of observed authenticators that authenticates this *)
          (* history state binding.  Our witness is the new authenticator  *)
          (* defined by the LL1PerformOperation action.                    *)
          (*****************************************************************)
          <5>7. LL1HistoryStateBindingAuthenticated(inclNewHistoryStateBinding)'
            (***************************************************************)
            (* The new authenticator defined in the LL1PerformOperation    *)
            (* action is unioned into the set of observed authenticators   *)
            (* by the LL1PerformOperation action.                          *)
            (***************************************************************)
            <6>1. newAuthenticator \in LL1ObservedAuthenticators'
              <7>1. LL1ObservedAuthenticators' =
                        LL1ObservedAuthenticators \union {newAuthenticator}
                BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                            newHistorySummary, newPrivateStateEnc, sResult, privateState
              <7>2. QED
                BY <7>1
            (***************************************************************)
            (* To prove that the new authenticator defined by the          *)
            (* LL1PerformOperation action authenticates the new history    *)
            (* state binding defined in the InclusionInvariant, we show    *)
            (* that this authenticator was generated using this history    *)
            (* state binding and using the same key.                       *)
            (***************************************************************)
            <6>2. ValidateMAC(LL1NVRAM.symmetricKey', inclNewHistoryStateBinding', newAuthenticator)
              (*************************************************************)
              (* First, we show that the new history state binding defined *)
              (* by the LL1PerformOperation action is equal to the primed  *)
              (* state of the new history state binding defined in the     *)
              (* InclusionInvariant.                                       *)
              (*************************************************************)
              <7>1. inclNewHistoryStateBinding' = newHistoryStateBinding
                (***********************************************************)
                (* The proof is fairly straightforward.  Using the         *)
                (* above-proven facts that the variables that satisfy the  *)
                (* antecedent conditions match the corresponding values in *)
                (* the NVRAM and the input value provided to the           *)
                (* LL1PerformOperation action, we show that the results of *)
                (* the service in the primed state of the                  *)
                (* InclusionInvariant are the same as the results of the   *)
                (* service in the LL1PerformOperation action.  From there, *)
                (* we show that the state hashes are equal and that the    *)
                (* history summaries are equal, which together imply that  *)
                (* the state bindings are equal.                           *)
                (***********************************************************)
                <8>1. UNCHANGED LL1NVRAM.symmetricKey
                  BY <2>1, SymmetricKeyConstantLemma
                <8>2. inclSResult' = sResult
                  <9>1. /\ publicState = LL1RAM.publicState
                        /\ inclPrivateState' = privateState
                    BY <5>5
                  <9>2. input2 = input1
                    BY <5>4
                  <9>3. QED
                    BY <9>1, <9>2 DEF inclSResult, sResult
                <8>3. inclSResult'.newPublicState = sResult.newPublicState
                  BY <8>2
                <8>4. inclSResult'.newPrivateState = sResult.newPrivateState
                  BY <8>2
                <8>5. inclNewPrivateStateEnc' = newPrivateStateEnc
                  BY <8>1, <8>4 DEF inclNewPrivateStateEnc, newPrivateStateEnc
                <8>6. inclNewStateHash' = newStateHash
                  BY <8>3, <8>5 DEF inclNewStateHash, newStateHash
                <8>7. LL1NVRAM.historySummary' = newHistorySummary
                  <10>1. LL1NVRAM' = [ historySummary |-> newHistorySummary,
                                   symmetricKey |-> LL1NVRAM.symmetricKey ]
                    BY <4>1 DEF newHistorySummary, newPrivateStateEnc, sResult, privateState
                  <10>2. QED
                    BY <10>1
                <8>8. QED
                  BY <8>6, <8>7 DEF inclNewHistoryStateBinding, newHistoryStateBinding
              (*************************************************************)
              (* Given that the state bindings are equal (and showing that *)
              (* the symmetric keys are equal, we can show that the new    *)
              (* authenticator defined by the LL1PerformOperation action   *)
              (* is equal to a MAC generated with the same inputs we are   *)
              (* attempting to validate.                                   *)
              (*************************************************************)
              <7>2. newAuthenticator =
                        GenerateMAC(LL1NVRAM.symmetricKey', inclNewHistoryStateBinding')
                <8>1. UNCHANGED LL1NVRAM.symmetricKey
                  BY <2>1, SymmetricKeyConstantLemma
                <8>2. QED
                  BY <7>1, <8>1 DEF newAuthenticator
              (*************************************************************)
              (* We can thus use the MACComplete property to show that the *)
              (* generated MAC validates appropriately.  To do this, we    *)
              (* first need to prove some types.                           *)
              (*************************************************************)
              <7>3. LL1NVRAM.symmetricKey' \in SymmetricKeyType
                <8>1. LL1TypeInvariant'
                  BY <2>1
                <8>2. QED
                  BY <8>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
              (*************************************************************)
              (* Since we have no lemma to prove the following type, we    *)
              (* include its entire type proof here.                       *)
              (*************************************************************)
              <7>4. inclNewHistoryStateBinding' \in HashType
                <8>1. inclSResult' \in ServiceResultType
                  <9>1. publicState \in PublicStateType
                    BY <5>1
                  <9>2. inclPrivateState' \in PrivateStateType
                    <10>1. LL1NVRAM.symmetricKey' \in SymmetricKeyType
                      <11>1. LL1TypeInvariant'
                        BY <2>1
                      <11>2. QED
                        BY <11>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
                    <10>2. privateStateEnc' \in PrivateStateEncType
                      BY <5>1
                    <10>3. QED
                      BY <10>1, <10>2, SymmetricDecryptionTypeSafe DEF inclPrivateState
                  <9>3. input2 \in InputType
                    BY <5>1
                  <9>4. QED
                    BY <9>1, <9>2, <9>3, ServiceTypeSafe DEF inclSResult
                <8>2. LL1NVRAM.historySummary' \in HashDomain
                  <9>1. LL1NVRAM.historySummary' \in HashType
                    <10>1. LL1TypeInvariant'
                      BY <2>1
                    <10>2. QED
                      BY <10>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
                  <9>2. QED
                    BY <9>1 DEF HashDomain
                <8>3. inclNewStateHash' \in HashDomain
                  <9>1. inclNewStateHash' \in HashType
                    <10>1. inclSResult.newPublicState' \in HashDomain
                      <11>1. inclSResult.newPublicState' \in PublicStateType
                        BY <8>1 DEF ServiceResultType
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>2. inclNewPrivateStateEnc' \in HashDomain
                      <11>1. inclNewPrivateStateEnc' \in PrivateStateEncType
                        <12>1. LL1NVRAM.symmetricKey' \in SymmetricKeyType
                          <13>1. LL1TypeInvariant'
                            BY <2>1
                          <13>2. QED
                            BY <13>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
                        <12>2. inclSResult.newPrivateState' \in PrivateStateType
                          BY <8>1 DEF ServiceResultType
                        <12>3. QED
                          BY <12>1, <12>2, SymmetricEncryptionTypeSafe DEF inclNewPrivateStateEnc
                      <11>2. QED
                        BY <11>1 DEF HashDomain
                    <10>3. QED
                      BY <10>1, <10>2, HashTypeSafe DEF inclNewStateHash
                  <9>2. QED
                    BY <9>1 DEF HashDomain
                <8>4. QED
                  BY <8>2, <8>3, HashTypeSafe DEF inclNewHistoryStateBinding
              (*************************************************************)
              (* Then, we appeal to the MACComplete property in a          *)
              (* straightforward way.                                      *)
              (*************************************************************)
              <7>5. QED
                BY <7>2, <7>3, <7>4, MACComplete
            <6>3. QED
              BY <6>1, <6>2 DEF LL1HistoryStateBindingAuthenticated
          (*****************************************************************)
          (* Since both conjuncts are true, the conjunction is true.       *)
          (*****************************************************************)
          <5>8. QED
            BY <5>6, <5>7
        (*******************************************************************)
        (* Second, we'll prove the CardinalityInvariant in the primed      *)
        (* state.                                                          *)
        (*******************************************************************)
        <4>7. CardinalityInvariant'
          (*****************************************************************)
          (* To prove the universally quantified expression, we take a new *)
          (* set of variables of the appropriate types.  For the TAKE step *)
          (* to be meaningful to the prover, first we have to tell the     *)
          (* prover to expand the definition of CardinalityInvariant, so   *)
          (* it will see the universally quantified expression therein.    *)
          (*****************************************************************)
          <5> USE DEF CardinalityInvariant
          <5>1. TAKE historySummary \in HashType, stateHash2 \in HashType
          (*****************************************************************)
          (* To simplify the writing of the proof, we re-state the         *)
          (* definition from the CardinalityInvariant.                     *)
          (*****************************************************************)
          <5> cardHistoryStateBinding == Hash(historySummary, stateHash2)
          (*****************************************************************)
          (* We then assert the type safety of these definitions, with the *)
          (* help of the CardinalityInvariantDefsTypeSafeLemma.            *)
          (*****************************************************************)
          <5>2. cardHistoryStateBinding \in HashType
              BY <5>1, CardinalityInvariantDefsTypeSafeLemma
          (*****************************************************************)
          (* The CardinalityInvariant states an implication.  To prove     *)
          (* this, it suffices to assume the antecedent and prove the      *)
          (* consequent.                                                   *)
          (*****************************************************************)
          <5>3. SUFFICES
                ASSUME
                    /\  LL1NVRAMHistorySummaryUncorrupted'
                    /\  LL1HistoryStateBindingAuthenticated(cardHistoryStateBinding)'
                PROVE
                    HashCardinality(historySummary) <= HashCardinality(LL1NVRAM.historySummary')
            OBVIOUS
          (*****************************************************************)
          (* We then hide the definitions.                                 *)
          (*****************************************************************)
          <5> HIDE DEF CardinalityInvariant
          <5> HIDE DEF cardHistoryStateBinding
          (*****************************************************************)
          (* First, we'll prove that the hash cardinality of the history   *)
          (* summary in the NVRAM does not decrease when a                 *)
          (* LL1PerformOperation action occurs.  This follows from the     *)
          (* fact that the hash cardinality of the history summary in the  *)
          (* LL1NVRAM increases by one when a LL1PerformOperation action   *)
          (* occurs.  This is straightforward though somewhat tedious.     *)
          (*****************************************************************)
          <5>4. HashCardinality(LL1NVRAM.historySummary) <=
                    HashCardinality(LL1NVRAM.historySummary')
            (***************************************************************)
            (* First, we prove a bunch of types that are needed by the     *)
            (* hash cardinality assumptions or for proving basic           *)
            (* arithmetic.                                                 *)
            (***************************************************************)
            <6>1. input1 \in InputType
              <7>1. input1 \in LL1AvailableInputs
                BY <4>1
              <7>2. LL1AvailableInputs \subseteq InputType
                <8>1. LL1TypeInvariant
                  BY <2>1
                <8>2. QED
                  BY <8>1 DEF LL1TypeInvariant
              <7>3. QED
                BY <7>1, <7>2
            <6>2. LL1NVRAM.historySummary \in HashDomain
              <7>1. LL1NVRAM.historySummary \in HashType
                BY <4>3
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>3. LL1NVRAM.historySummary' \in HashDomain
              <7>1. LL1NVRAM.historySummary' \in HashType
                BY <4>4
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>4. input1 \in HashDomain
                BY <6>1 DEF HashDomain
            <6>5. HashCardinality(LL1NVRAM.historySummary) \in Nat
              BY <6>2, HashCardinalityTypeSafe DEF HashDomain
            <6>6. HashCardinality(LL1NVRAM.historySummary') \in Nat
              BY <6>3, HashCardinalityTypeSafe DEF HashDomain
            (***************************************************************)
            (* With the type statements out of the way, we can construct a *)
            (* simple inequality.  We do this in four linear steps.        *)
            (***************************************************************)
            <6>7. HashCardinality(Hash(LL1NVRAM.historySummary, input1)) =
                       HashCardinality(LL1NVRAM.historySummary) + HashCardinality(input1) + 1
              BY <6>2, <6>4, HashCardinalityAccumulative
            <6>8. HashCardinality(LL1NVRAM.historySummary') =
                       HashCardinality(LL1NVRAM.historySummary) + HashCardinality(input1) + 1
              <7>1. LL1NVRAM.historySummary' = Hash(LL1NVRAM.historySummary, input1)
                <8>1. newHistorySummary = Hash(LL1NVRAM.historySummary, input1)
                  BY DEF newHistorySummary
                <8>2. LL1NVRAM.historySummary' = newHistorySummary
                  <9>1. LL1NVRAM' = [
                            historySummary |-> newHistorySummary,
                            symmetricKey |-> LL1NVRAM.symmetricKey ]
                    BY <4>1 DEF newHistorySummary, newPrivateStateEnc, sResult, privateState
                  <9>2. QED
                    BY <9>1
                <8>3. QED
                  BY <8>1, <8>2
              <7>2. QED
                BY <7>1, <6>7
            <6>9. HashCardinality(LL1NVRAM.historySummary') =
                      HashCardinality(LL1NVRAM.historySummary) + 1
              <7>3. HashCardinality(input1) = 0
                BY <6>1, InputCardinalityZero
              <7>4. QED
                BY <6>5, <6>6, <6>8, <7>3
            <6>10. QED
              BY <6>5, <6>6, <6>9
          (*****************************************************************)
          (* Next, we'll prove that either the history state binding is    *)
          (* authenticated in the unprimed state or the history state      *)
          (* binding is authenticated by the new authenticator defined by  *)
          (* the LL1PerformOperation action.  This follows from the fact   *)
          (* that the primed set of authenticators is constructed as the   *)
          (* union of the unprimed set and the new authenticator.          *)
          (*****************************************************************)
          <5>5. \/ LL1HistoryStateBindingAuthenticated(cardHistoryStateBinding)
                \/ ValidateMAC(LL1NVRAM.symmetricKey, cardHistoryStateBinding, newAuthenticator)
            <6>1. UNCHANGED LL1NVRAM.symmetricKey
              BY <2>1, SymmetricKeyConstantLemma
            <6>2. LL1ObservedAuthenticators' =
                      LL1ObservedAuthenticators \union {newAuthenticator}
              BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                      newHistorySummary, newPrivateStateEnc, sResult, privateState
            <6>3. LL1HistoryStateBindingAuthenticated(cardHistoryStateBinding)'
              BY <5>3
            <6>4. QED
              BY <6>1, <6>2, <6>3 DEF LL1HistoryStateBindingAuthenticated
          (*****************************************************************)
          (* Given the above disjunction, we proceed via case analysis.    *)
          (* First, we consider the case in which the history state        *)
          (* binding is authenticated in the unprimed state.               *)
          (*                                                               *)
          (* In this case, because the CardinalityInvariant is true in the *)
          (* unprimed state, we can prove that the hash cardinality of the *)
          (* history summary that satisfies the antecedent is less than or *)
          (* equal to the hash cardinality of the history summary in the   *)
          (* unprimed NVRAM.  Since the less-than-or-equal-to relation is  *)
          (* transitive, this leads directly to our proof goal.            *)
          (*****************************************************************)
          <5>6. CASE LL1HistoryStateBindingAuthenticated(cardHistoryStateBinding)
            <6>2. HashCardinality(historySummary) <= HashCardinality(LL1NVRAM.historySummary)
              (*************************************************************)
              (* We know the CardinalityInvariant to be true in the        *)
              (* unprimed state.                                           *)
              (*************************************************************)
              <7>1. CardinalityInvariant
                BY <2>1
              (*************************************************************)
              (* The CardinalityInvariant has two type assumptions.        *)
              (*************************************************************)
              <7>2. historySummary \in HashType
                BY <5>1
              <7>3. stateHash2 \in HashType
                BY <5>1
              (*************************************************************)
              (* The implication in the CardinalityInvariant has           *)
              (* antecedents.  One is equal to the present case condition. *)
              (* The other is that the LL1NVRAMHistorySummaryUncorrupted   *)
              (* predicate is true.                                        *)
              (*************************************************************)
              <7>4. LL1NVRAMHistorySummaryUncorrupted
                BY <4>5
              (*************************************************************)
              (* The consequent of the implication in the                  *)
              (* CardinalityInvariant follows directly.                    *)
              (*************************************************************)
              <7>5. QED
                BY <5>6, <7>1, <7>2, <7>3, <7>4
                DEF CardinalityInvariant, cardHistoryStateBinding
            <6>3. HashCardinality(historySummary) \in Nat
              <7>1. historySummary \in HashDomain
                <8>1. historySummary \in HashType
                  BY <5>1
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>2. QED
                BY <7>1, HashCardinalityTypeSafe
            <6>4. HashCardinality(LL1NVRAM.historySummary) \in Nat
              <7>1. LL1NVRAM.historySummary \in HashDomain
                <8>1. LL1NVRAM.historySummary \in HashType
                  BY <4>3
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>2. QED
                BY <7>1, HashCardinalityTypeSafe
            <6>5. HashCardinality(LL1NVRAM.historySummary') \in Nat
              <7>1. LL1NVRAM.historySummary' \in HashDomain
                <8>1. LL1NVRAM.historySummary' \in HashType
                  BY <4>4
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>2. QED
                BY <7>1, HashCardinalityTypeSafe
            <6>6. QED
              (*************************************************************)
              (* Ideally, this QED step should just read:                  *)
              (*                                                           *)
              (*  BY <<5>>3, <<6>>2, <<6>>3, <<6>>4, <<6>>5, LEQTransitive *)
              (*                                                           *)
              (* However, the prover seems to get a little confused in     *)
              (* this instance.  We make life easier for the prover by     *)
              (* definining some local variables and hiding their          *)
              (* definitions before appealing to the LEQTransitive         *)
              (* assumption.                                               *)
              (*************************************************************)
              <7> x == HashCardinality(historySummary)
              <7> y == HashCardinality(LL1NVRAM.historySummary)
              <7> z == HashCardinality(LL1NVRAM.historySummary')
              <7>1. x \in Nat
                BY <6>3
              <7>2. y \in Nat
                BY <6>4
              <7>3. z \in Nat
                BY <6>5
              <7>4. x <= y
                BY <6>2
              <7>5. y <= z
                BY <5>4
              <7>6. x <= z
                <8> HIDE DEF x, y, z
                <8>1. QED
                  BY <7>1, <7>2, <7>3, <7>4, <7>5, LEQTransitive
              <7>7. QED
                BY <7>6
          (*****************************************************************)
          (* In the other case, the history state binding is authenticated *)
          (* by the new authenticator defined by the LL1PerformOperation   *)
          (* action                                                        *)
          (*****************************************************************)
          <5>7. CASE ValidateMAC(LL1NVRAM.symmetricKey, cardHistoryStateBinding, newAuthenticator)
            (***************************************************************)
            (* Since the new authenticator authenticates the new history   *)
            (* state binding defined in the LL1PerformOperation action, it *)
            (* follows that the history state bindings are equal, by the   *)
            (* MACCollisionResistant property.                             *)
            (***************************************************************)
            <6>1. cardHistoryStateBinding = newHistoryStateBinding
              <7>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
                BY <4>3
              <7>3. cardHistoryStateBinding \in HashType
                BY <5>2
              <7>4. newHistoryStateBinding \in HashType
                BY <4>2
              <7>5. QED
                BY <5>7, <7>1, <7>3, <7>4, MACCollisionResistant DEF newAuthenticator
            (***************************************************************)
            (* Since the inputs-state bindings are equal, it follows that  *)
            (* the history summaries are equal, by the                     *)
            (* HashCollisionResistant property and the fact that the       *)
            (* history summary in the primed NVRAM equals the new history  *)
            (* summary defined in the LL1PerformOperation action.          *)
            (***************************************************************)
            <6>2. historySummary = LL1NVRAM.historySummary'
              <7>1. historySummary = newHistorySummary
                <8>1. historySummary \in HashDomain
                  <9>1. historySummary \in HashType
                    BY <5>5
                  <9>2. QED
                    BY <9>1 DEF HashDomain
                <8>2. stateHash2 \in HashDomain
                  <9>1. stateHash2 \in HashType
                    BY <5>1
                  <9>2. QED
                    BY <9>1 DEF HashDomain
                <8>3. newHistorySummary \in HashDomain
                  <9>1. newHistorySummary \in HashType
                    BY <4>2
                  <9>2. QED
                    BY <9>1 DEF HashDomain
                <8>4. newStateHash \in HashDomain
                  <9>1. newStateHash \in HashType
                    BY <4>2
                  <9>2. QED
                    BY <9>1 DEF HashDomain
                <8>5. QED
                  BY <6>1, <8>1, <8>2, <8>3, <8>4, HashCollisionResistant
                  DEF cardHistoryStateBinding, newHistoryStateBinding
              <7>2. LL1NVRAM.historySummary' = newHistorySummary
                <8>1. LL1NVRAM' = [ historySummary |-> newHistorySummary,
                                 symmetricKey |-> LL1NVRAM.symmetricKey ]
                  BY <4>1 DEF newHistorySummary, newPrivateStateEnc, sResult, privateState
                <8>2. QED
                  BY <8>1
              <7>3. QED
                BY <7>1, <7>2
            (***************************************************************)
            (* Since the history summaries are equal, their hash           *)
            (* cardinalities are equal.                                    *)
            (***************************************************************)
            <6>3. HashCardinality(historySummary) = HashCardinality(LL1NVRAM.historySummary')
              BY <6>2
            (***************************************************************)
            (* We then have to prove that the hash cardinalities are       *)
            (* natural numbers, to enable the prover to conclude that the  *)
            (* equality above implies the less-than-or-equal we are trying *)
            (* to prove.                                                   *)
            (***************************************************************)
            <6>4. HashCardinality(historySummary) \in Nat
              <7>1. historySummary \in HashType
                BY <5>1
              <7>2. historySummary \in HashDomain
                BY <7>1 DEF HashDomain
              <7>3. QED
                BY <7>2, HashCardinalityTypeSafe
            <6>5. HashCardinality(LL1NVRAM.historySummary') \in Nat
              <7>1. LL1NVRAM.historySummary' \in HashType
                BY <4>4
              <7>2. LL1NVRAM.historySummary' \in HashDomain
                BY <7>1 DEF HashDomain
              <7>3. QED
                BY <7>2, HashCardinalityTypeSafe
            <6>6. QED
              BY <6>3, <6>4, <6>5
          (*****************************************************************)
          (* The two cases are exhaustive, so the CardinalityInvariant is  *)
          (* proven.                                                       *)
          (*****************************************************************)
          <5>8. QED
            BY <5>5, <5>6, <5>7
        (*******************************************************************)
        (* Third, we'll prove the UniquenessInvariant in the primed state. *)
        (*******************************************************************)
        <4>8. UniquenessInvariant'
          (*****************************************************************)
          (* To prove the universally quantified expression, we take a new *)
          (* set of variables of the appropriate types.  For the TAKE step *)
          (* to be meaningful to the prover, first we have to tell the     *)
          (* prover to expand the definition of UniquenessInvariant, so it *)
          (* will see the universally quantified expression therein.       *)
          (*****************************************************************)
          <5> USE DEF UniquenessInvariant
          <5>2. TAKE stateHash1, stateHash2 \in HashType
          (*****************************************************************)
          (* To simplify the writing of the proof, we re-state the         *)
          (* definitions from the UniquenessInvariant.                     *)
          (*****************************************************************)
          <5> uniqHistoryStateBinding1 == Hash(LL1NVRAM.historySummary, stateHash1)
          <5> uniqHistoryStateBinding2 == Hash(LL1NVRAM.historySummary, stateHash2)
          (*****************************************************************)
          (* The UniquenessInvariant states an implication.  To prove      *)
          (* this, it suffices to assume the antecedent and prove the      *)
          (* consequent.                                                   *)
          (*****************************************************************)
          <5>3. SUFFICES
                ASSUME /\  LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding1)'
                       /\  LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding2)'
                PROVE stateHash1 = stateHash2
            OBVIOUS
          (*****************************************************************)
          (* We hide the definitions of UniquenessInvariant and the        *)
          (* definitions from the UniquenessInvariant.                     *)
          (*****************************************************************)
          <5> HIDE DEF UniquenessInvariant
          <5> HIDE DEF uniqHistoryStateBinding1, uniqHistoryStateBinding2
          (*****************************************************************)
          (* The proof of UniquenessInvariant' employs                     *)
          (* CardinalityInvariant.  To facilitate our extremely simple     *)
          (* arithmetic, we begin by proving the types of a couple of      *)
          (* critical hash cardinalities.                                  *)
          (*****************************************************************)
          <5>4. HashCardinality(LL1NVRAM.historySummary) \in Nat
            <6>1. LL1NVRAM.historySummary \in HashDomain
              <7>1. LL1NVRAM.historySummary \in HashType
                BY <4>3
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>2. QED
              BY <6>1, HashCardinalityTypeSafe
          <5>5. HashCardinality(LL1NVRAM.historySummary') \in Nat
            <6>1. LL1NVRAM.historySummary' \in HashDomain
              <7>1. LL1NVRAM.historySummary' \in HashType
                BY <4>4
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>2. QED
              BY <6>1, HashCardinalityTypeSafe
          (*****************************************************************)
          (* We'll prove that the hash cardinality of the history summary  *)
          (* in the LL1NVRAM increases when a LL1PerformOperation action   *)
          (* occurs.  This follows from the fact that the hash cardinality *)
          (* of the history summary in the NVRAM increases by one when a   *)
          (* LL1PerformOperation action occurs.  This is straightforward   *)
          (* though somewhat tedious.                                      *)
          (*****************************************************************)
          <5>6. HashCardinality(LL1NVRAM.historySummary) <
                    HashCardinality(LL1NVRAM.historySummary')
            (***************************************************************)
            (* First, we prove a bunch of types that are needed by the     *)
            (* hash cardinality assumptions or for proving basic           *)
            (* arithmetic.                                                 *)
            (***************************************************************)
            <6>1. LL1NVRAM.historySummary \in HashDomain
              <7>1. LL1NVRAM.historySummary \in HashType
                BY <4>3
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>2. LL1NVRAM.historySummary' \in HashDomain
              <7>1. LL1NVRAM.historySummary' \in HashType
                BY <4>4
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>3. input1 \in HashDomain
              <7>1. input1 \in InputType
                <8>1. input1 \in LL1AvailableInputs
                  BY <4>1
                <8>2. LL1AvailableInputs \subseteq InputType
                  <9>1. LL1TypeInvariant
                    BY <2>1
                  <9>2. QED
                    BY <9>1 DEF LL1TypeInvariant
                <8>3. QED
                  BY <8>1, <8>2
              <7>2. QED
                BY <7>1 DEF HashDomain
            (***************************************************************)
            (* With the type statements out of the way, we can construct a *)
            (* simple inequality.  We do this in four linear steps.        *)
            (***************************************************************)
            <6>4. HashCardinality(Hash(LL1NVRAM.historySummary, input1)) =
                       HashCardinality(LL1NVRAM.historySummary) + HashCardinality(input1) + 1
              BY <6>1, <6>3, HashCardinalityAccumulative
            <6>5. HashCardinality(LL1NVRAM.historySummary') =
                       HashCardinality(LL1NVRAM.historySummary) + HashCardinality(input1) + 1
              <7>1. LL1NVRAM.historySummary' = Hash(LL1NVRAM.historySummary, input1)
                <8>1. newHistorySummary = Hash(LL1NVRAM.historySummary, input1)
                  BY DEF newHistorySummary
                <8>2. LL1NVRAM.historySummary' = newHistorySummary
                  <9>1. LL1NVRAM' = [ historySummary |-> newHistorySummary,
                                   symmetricKey |-> LL1NVRAM.symmetricKey ]
                    BY <4>1 DEF newHistorySummary, newPrivateStateEnc, sResult, privateState
                  <9>2. QED
                    BY <9>1
                <8>3. QED
                  BY <8>1, <8>2
              <7>2. QED
                BY <7>1, <6>4
            <6>6. HashCardinality(LL1NVRAM.historySummary') =
                      HashCardinality(LL1NVRAM.historySummary) + 1
              <7>3. HashCardinality(input1) = 0
                <8>1. input1 \in InputType
                  <9>1. input1 \in LL1AvailableInputs
                    BY <4>1
                  <9>2. LL1AvailableInputs \subseteq InputType
                    <10>1. LL1TypeInvariant
                      BY <2>1
                    <10>2. QED
                      BY <10>1 DEF LL1TypeInvariant
                  <9>3. QED
                    BY <9>1, <9>2
                <8>2. QED
                  BY <8>1, InputCardinalityZero
              <7>4. QED
                BY <5>4, <5>5, <6>5, <7>3
            <6>7. QED
              BY <5>4, <5>5, <6>6
          (*****************************************************************)
          (* Next, we'll prove that the primed history state binding 1     *)
          (* defined by the UniquenessInvariant is equal to the new state  *)
          (* history state binding defined by the LL1PerformOperation      *)
          (* action.                                                       *)
          (*****************************************************************)
          <5>7. uniqHistoryStateBinding1' = newHistoryStateBinding
            <6>1. ValidateMAC(LL1NVRAM.symmetricKey, uniqHistoryStateBinding1', newAuthenticator)
              (*************************************************************)
              (* We start by proving that either the history state binding *)
              (* is authenticated in the unprimed state or the history     *)
              (* state binding is authenticated by the new authenticator   *)
              (* defined by the LL1PerformOperation action.  This follows  *)
              (* from the fact that the primed set of authenticators is    *)
              (* constructed as the union of the unprimed set and the new  *)
              (* authenticator.                                            *)
              (*************************************************************)
              <7>1. \/ LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding1')
                    \/ ValidateMAC(LL1NVRAM.symmetricKey, uniqHistoryStateBinding1', newAuthenticator)
                <8>1. LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding1)'
                  BY <5>3
                <8>2. UNCHANGED LL1NVRAM.symmetricKey
                  BY <2>1, SymmetricKeyConstantLemma
                <8>3. LL1ObservedAuthenticators' =
                          LL1ObservedAuthenticators \union {newAuthenticator}
                  BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                          newHistorySummary, newPrivateStateEnc, sResult, privateState
                <8>4. QED
                  BY <8>1, <8>2, <8>3 DEF LL1HistoryStateBindingAuthenticated
              (*************************************************************)
              (* We then prove that the history state binding is not       *)
              (* authenticated in the unprimed state.  We use proof by     *)
              (* contradiction.  We show that if the history state binding *)
              (* were authenticated in the unprimed state, we could        *)
              (* conclude a hash cardinality inequality that contradicts   *)
              (* the hash cardinality inequality we proved above.          *)
              (*************************************************************)
              <7>2. ~LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding1')
                <8>1. SUFFICES
                      ASSUME LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding1')
                      PROVE FALSE
                  OBVIOUS
                <8>2. HashCardinality(LL1NVRAM.historySummary') <=
                          HashCardinality(LL1NVRAM.historySummary)
                  <9>1. CardinalityInvariant
                    BY <2>1
                  <9>2. LL1NVRAM.historySummary' \in HashType
                   BY <4>4
                  <9>3. stateHash1 \in HashType
                    BY <5>2
                  <9>4. LL1NVRAMHistorySummaryUncorrupted
                    BY <4>5
                  <9>5. QED
                    BY <8>1, <9>1, <9>2, <9>3, <9>4 DEF CardinalityInvariant, uniqHistoryStateBinding1
                <8>3. QED
                  BY <5>4, <5>5, <5>6, <8>2, GEQorLT
              <7>3. QED
                BY <7>1, <7>2
            (***************************************************************)
            (* Since the new authenticator authenticates the new history   *)
            (* state binding defined in the LL1PerformOperation action, it *)
            (* follows that the history state bindings are equal, by the   *)
            (* MACCollisionResistant property.                             *)
            (*                                                             *)
            (* We first have to prove some basic types, then we can appeal *)
            (* to the MACCollisionResistant property.                      *)
            (***************************************************************)
            <6>2. LL1NVRAM.symmetricKey \in SymmetricKeyType
              BY <4>3
            <6>3. uniqHistoryStateBinding1' \in HashType
              <7>1. LL1NVRAM.historySummary' \in HashDomain
                <8>1. LL1NVRAM.historySummary' \in HashType
                  <9>1. LL1TypeInvariant'
                    BY <2>1
                  <9>2. QED
                    BY <9>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>2. stateHash1 \in HashDomain
                <8>1. stateHash1 \in HashType
                  BY <5>2
                <8>2. QED
                  BY <3>1 DEF HashDomain
              <7>3. QED
                BY <7>1, <7>2, HashTypeSafe DEF uniqHistoryStateBinding1
            <6>4. newHistoryStateBinding \in HashType
              BY <4>2
            <6>5. QED
              BY <6>1, <6>2, <6>3, <6>4, MACCollisionResistant DEF newAuthenticator
          (*****************************************************************)
          (* Next, we'll prove that the primed history state binding 2     *)
          (* defined by the UniquenessInvariant is equal to the new state  *)
          (* history state binding defined by the LL1PerformOperation      *)
          (* action.                                                       *)
          (*****************************************************************)
          <5>8. uniqHistoryStateBinding2' = newHistoryStateBinding
            <6>1. ValidateMAC(LL1NVRAM.symmetricKey, uniqHistoryStateBinding2', newAuthenticator)
              (*************************************************************)
              (* We start by proving that either the history state binding *)
              (* is authenticated in the unprimed state or the history     *)
              (* state binding is authenticated by the new authenticator   *)
              (* defined by the LL1PerformOperation action.  This follows  *)
              (* from the fact that the primed set of authenticators is    *)
              (* constructed as the union of the unprimed set and the new  *)
              (* authenticator.                                            *)
              (*************************************************************)
              <7>1. \/ LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding2')
                    \/ ValidateMAC(LL1NVRAM.symmetricKey, uniqHistoryStateBinding2', newAuthenticator)
                <8>1. LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding2)'
                  BY <5>3
                <8>2. UNCHANGED LL1NVRAM.symmetricKey
                  BY <2>1, SymmetricKeyConstantLemma
                <8>3. LL1ObservedAuthenticators' =
                          LL1ObservedAuthenticators \union {newAuthenticator}
                  BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                          newHistorySummary, newPrivateStateEnc, sResult, privateState
                <8>4. QED
                  BY <8>1, <8>2, <8>3 DEF LL1HistoryStateBindingAuthenticated
              (*************************************************************)
              (* We then prove that the history state binding is not       *)
              (* authenticated in the unprimed state.  We use proof by     *)
              (* contradiction.  We show that if the history state binding *)
              (* were authenticated in the unprimed state, we could        *)
              (* conclude a hash cardinality inequality that contradicts   *)
              (* the hash cardinality inequality we proved above.          *)
              (*************************************************************)
              <7>2. ~LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding2')
                <8>1. SUFFICES
                      ASSUME LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding2')
                      PROVE FALSE
                  OBVIOUS
                <8>2. HashCardinality(LL1NVRAM.historySummary') <=
                          HashCardinality(LL1NVRAM.historySummary)
                  <9>1. CardinalityInvariant
                    BY <2>1
                  <9>2. LL1NVRAM.historySummary' \in HashType
                   BY <4>4
                  <9>3. stateHash2 \in HashType
                    BY <5>2
                  <9>4. LL1NVRAMHistorySummaryUncorrupted
                    BY <4>5
                  <9>5. QED
                    BY <8>1, <9>1, <9>2, <9>3, <9>4 DEF CardinalityInvariant, uniqHistoryStateBinding2
                <8>3. QED
                  BY <5>4, <5>5, <5>6, <8>2, GEQorLT
              <7>3. QED
                BY <7>1, <7>2
            (***************************************************************)
            (* Since the new authenticator authenticates the new history   *)
            (* state binding defined in the LL1PerformOperation action, it *)
            (* follows that the history state bindings are equal, by the   *)
            (* MACCollisionResistant property.                             *)
            (*                                                             *)
            (* We first have to prove some basic types, then we can appeal *)
            (* to the MACCollisionResistant property.                      *)
            (***************************************************************)
            <6>2. LL1NVRAM.symmetricKey \in SymmetricKeyType
              BY <4>3
            <6>3. uniqHistoryStateBinding2' \in HashType
              <7>1. LL1NVRAM.historySummary' \in HashDomain
                <8>1. LL1NVRAM.historySummary' \in HashType
                  <9>1. LL1TypeInvariant'
                    BY <2>1
                  <9>2. QED
                    BY <9>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
                <8>2. QED
                  BY <8>1 DEF HashDomain
              <7>2. stateHash2 \in HashDomain
                <8>1. stateHash2 \in HashType
                  BY <5>2
                <8>2. QED
                  BY <3>1 DEF HashDomain
              <7>3. QED
                BY <7>1, <7>2, HashTypeSafe DEF uniqHistoryStateBinding2
            <6>4. newHistoryStateBinding \in HashType
              BY <4>2
            <6>5. QED
              BY <6>1, <6>2, <6>3, <6>4, MACCollisionResistant DEF newAuthenticator
          (*****************************************************************)
          (* Since each of the history state bindings is equal to the      *)
          (* history state binding defined by the LL1PerformOperation      *)
          (* action, the two history state bindings must be equal to each  *)
          (* other.                                                        *)
          (*****************************************************************)
          <5>9. uniqHistoryStateBinding1' = uniqHistoryStateBinding2'
            BY <5>7, <5>8
          (*****************************************************************)
          (* Since the inputs-state bindings are equal, it follows that    *)
          (* the state hashes are equal, by the HashCollisionResistant     *)
          (* property.                                                     *)
          (*****************************************************************)
          <5>10. QED
            <6>1. LL1NVRAM.historySummary' \in HashDomain
              <7>1. LL1NVRAM.historySummary' \in HashType
                BY <4>4
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>2. stateHash1 \in HashDomain
              <7>1. stateHash1 \in HashType
                BY <5>2
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>3. stateHash2 \in HashDomain
              <7>1. stateHash2 \in HashType
                BY <5>2
              <7>2. QED
                BY <7>1 DEF HashDomain
            <6>4. QED
              BY <5>9, <6>1, <6>2, <6>3, HashCollisionResistant
              DEF uniqHistoryStateBinding1, uniqHistoryStateBinding2
        <4>9. QED
          BY <4>6, <4>7, <4>8
      (*********************************************************************)
      (* The LL1RepeatOperation case is a straightforward application of   *)
      (* the InclusionUnchangedLemma, the CardinalityUnchangedLemma, and   *)
      (* the UniquenessUnchangedLemma, the preconditions for which follow  *)
      (* from the LL1RepeatOperationUnchangedObseredOutputsLemma and the   *)
      (* LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma. *)
      (*********************************************************************)
      <3>3. CASE LL1RepeatOperation
        <4>1. PICK input \in LL1AvailableInputs : LL1RepeatOperation!(input)!1 
          BY <3>3 DEF LL1RepeatOperation
        <4>2. UNCHANGED LL1NVRAM
          BY <4>1
        <4>3. UNCHANGED LL1ObservedOutputs
          <5>1. LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant
            BY <2>1
          <5>2. QED
            BY <3>3, <5>1, LL1RepeatOperationUnchangedObseredOutputsLemma
        <4>4. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant
            BY <2>1
          <5>2. QED
            BY <3>3, <5>1, LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma
        <4>5. InclusionInvariant'
          <5>1. InclusionInvariant /\ LL1TypeInvariant /\  LL1TypeInvariant'
            BY <2>1
          <5>2. QED
            BY <4>2, <4>3, <4>4, <5>1, InclusionUnchangedLemma
        <4>6. CardinalityInvariant'
          <5>1. CardinalityInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>2, <4>3, <4>4, <5>1, CardinalityUnchangedLemma
        <4>7. UniquenessInvariant'
          <5>1. UniquenessInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>2, <4>3, <4>4, <5>1, UniquenessUnchangedLemma
        <4>8. QED
          BY <4>5, <4>6, <4>7
      (*********************************************************************)
      (* The LL1Restart case is a straightforward application of the       *)
      (* InclusionUnchangedLemma, the CardinalityUnchangedLemma, and the   *)
      (* UniquenessUnchangedLemma.                                         *)
      (*********************************************************************)
      <3>4. CASE LL1Restart
        <4>1. UNCHANGED <<LL1NVRAM, LL1ObservedOutputs>>
          BY <3>4 DEF LL1Restart
        <4>2. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>4 DEF LL1Restart
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        <4>3. InclusionInvariant'
          <5>1. InclusionInvariant /\ LL1TypeInvariant /\  LL1TypeInvariant'
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, InclusionUnchangedLemma
        <4>4. CardinalityInvariant'
          <5>1. CardinalityInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, CardinalityUnchangedLemma
        <4>5. UniquenessInvariant'
          <5>1. UniquenessInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, UniquenessUnchangedLemma
        <4>6. QED
          BY <4>3, <4>4, <4>5
      (*********************************************************************)
      (* The LL1ReadDisk case is a straightforward application of the      *)
      (* InclusionUnchangedLemma, the CardinalityUnchangedLemma, and the   *)
      (* UniquenessUnchangedLemma.                                         *)
      (*********************************************************************)
      <3>5. CASE LL1ReadDisk
        <4>1. UNCHANGED <<LL1NVRAM, LL1ObservedOutputs>>
          BY <3>5 DEF LL1ReadDisk
        <4>2. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>5 DEF LL1ReadDisk
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        <4>3. InclusionInvariant'
          <5>1. InclusionInvariant /\ LL1TypeInvariant /\  LL1TypeInvariant'
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, InclusionUnchangedLemma
        <4>4. CardinalityInvariant'
          <5>1. CardinalityInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, CardinalityUnchangedLemma
        <4>5. UniquenessInvariant'
          <5>1. UniquenessInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, UniquenessUnchangedLemma
        <4>6. QED
          BY <4>3, <4>4, <4>5
      (*********************************************************************)
      (* The LL1WriteDisk case is a straightforward application of the     *)
      (* InclusionUnchangedLemma, the CardinalityUnchangedLemma, and the   *)
      (* UniquenessUnchangedLemma.                                         *)
      (*********************************************************************)
      <3>6. CASE LL1WriteDisk
        <4>1. UNCHANGED <<LL1NVRAM, LL1ObservedOutputs>>
          BY <3>6 DEF LL1WriteDisk
        <4>2. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>6 DEF LL1WriteDisk
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        <4>3. InclusionInvariant'
          <5>1. InclusionInvariant /\ LL1TypeInvariant /\  LL1TypeInvariant'
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, InclusionUnchangedLemma
        <4>4. CardinalityInvariant'
          <5>1. CardinalityInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, CardinalityUnchangedLemma
        <4>5. UniquenessInvariant'
          <5>1. UniquenessInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, UniquenessUnchangedLemma
        <4>6. QED
          BY <4>3, <4>4, <4>5
      (*********************************************************************)
      (* The LL1CorruptRAM case is a straightforward application of the    *)
      (* InclusionUnchangedLemma, the CardinalityUnchangedLemma, and the   *)
      (* UniquenessUnchangedLemma.                                         *)
      (*********************************************************************)
      <3>7. CASE LL1CorruptRAM
        <4>1. UNCHANGED <<LL1NVRAM, LL1ObservedOutputs>>
          BY <3>7 DEF LL1CorruptRAM
        <4>2. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>7 DEF LL1CorruptRAM
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        <4>3. InclusionInvariant'
          <5>1. InclusionInvariant /\ LL1TypeInvariant /\  LL1TypeInvariant'
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, InclusionUnchangedLemma
        <4>4. CardinalityInvariant'
          <5>1. CardinalityInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, CardinalityUnchangedLemma
        <4>5. UniquenessInvariant'
          <5>1. UniquenessInvariant /\ LL1TypeInvariant
            BY <2>1
          <5>2. QED
            BY <4>1, <4>2, <5>1, UniquenessUnchangedLemma
        <4>6. QED
          BY <4>3, <4>4, <4>5
      (*********************************************************************)
      (* The LL1RestrictedCorruption case is non-trivial, although nowhere *)
      (* near as involved as the LL1PerformOperation case.                 *)
      (*********************************************************************)
      <3>8. CASE LL1RestrictedCorruption
        (*******************************************************************)
        (* We pick a garbage history summary in the appropriate type.      *)
        (*******************************************************************)
        <4>1. PICK garbageHistorySummary \in HashType :
                  LL1RestrictedCorruption!nvram!(garbageHistorySummary)
          BY <3>8 DEF  LL1RestrictedCorruption
        (*******************************************************************)
        (* The primed value of the history summary in the NVRAM equals     *)
        (* this garbage history summary.                                   *)
        (*******************************************************************)
        <4>2. LL1NVRAM.historySummary' = garbageHistorySummary
          <5>1. LL1NVRAM' = [ historySummary |-> garbageHistorySummary,
                           symmetricKey |-> LL1NVRAM.symmetricKey ]
            BY <4>1
          <5>2. QED
            BY <5>1
        (*******************************************************************)
        (* We now prove each invariant separately, starting with the       *)
        (* InclusionInvariant.                                             *)
        (*******************************************************************)
        <4>3. InclusionInvariant'
          (*****************************************************************)
          (* To prove the universally quantified expression, we take a new *)
          (* set of variables of the appropriate types.  For the TAKE step *)
          (* to be meaningful to the prover, first we have to tell the     *)
          (* prover to expand the definition of InclusionInvariant, so it  *)
          (* will see the universally quantified expression therein.       *)
          (*****************************************************************)
          <5> USE DEF InclusionInvariant
          <5>1. TAKE input \in InputType,
                     historySummary \in HashType,
                     publicState \in PublicStateType,
                     privateStateEnc \in PrivateStateEncType
          (*****************************************************************)
          (* To simplify the writing of the proof, we re-state the         *)
          (* definitions from the InclusionInvariant.                      *)
          (*****************************************************************)
          <5> inclStateHash == Hash(publicState, privateStateEnc)
          <5> inclHistoryStateBinding == Hash(historySummary, inclStateHash)
          <5> inclPrivateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, privateStateEnc)
          <5> inclSResult == Service(publicState, inclPrivateState, input)
          <5> inclNewPrivateStateEnc ==
                  SymmetricEncrypt(LL1NVRAM.symmetricKey, inclSResult.newPrivateState)
          <5> inclNewStateHash == Hash(inclSResult.newPublicState, inclNewPrivateStateEnc)
          <5> inclNewHistoryStateBinding == Hash(LL1NVRAM.historySummary, inclNewStateHash)
          (*****************************************************************)
          (* We then assert the type safety of these definitions, with the *)
          (* help of the InclusionInvariantDefsTypeSafeLemma.              *)
          (*****************************************************************)
          <5>2. /\  inclStateHash \in HashType
                /\  inclHistoryStateBinding \in HashType
                /\  inclPrivateState \in PrivateStateType
                /\  inclSResult \in ServiceResultType
                /\  inclSResult.newPublicState \in PublicStateType
                /\  inclSResult.newPrivateState \in PrivateStateType
                /\  inclSResult.output \in OutputType
                /\  inclNewPrivateStateEnc \in PrivateStateEncType
                /\  inclNewStateHash \in HashType
                /\  inclNewHistoryStateBinding \in HashType
            <6>1. LL1TypeInvariant
              BY <2>1
            <6>2. QED
              BY <5>1, <6>1, InclusionInvariantDefsTypeSafeLemma
          (*****************************************************************)
          (* The InclusionInvariant states an implication.  To prove this, *)
          (* it suffices to prove that the the antecedent is false.  The   *)
          (* antecedent is a conjunction.  We will prove that the first    *)
          (* conjunct implies that the second conjunct is false.           *)
          (*****************************************************************)
          <5>3. SUFFICES
                ASSUME TRUE
                PROVE
                    LL1NVRAM.historySummary' = Hash(historySummary, input) =>
                        ~LL1HistoryStateBindingAuthenticated(inclHistoryStateBinding)'
            OBVIOUS
          (*****************************************************************)
          (* We hide the definition of InclusionInvariant and the          *)
          (* definitions from the InclusionInvariant.                      *)
          (*****************************************************************)
          <5> HIDE DEF InclusionInvariant
          <5> HIDE DEF inclStateHash, inclHistoryStateBinding, inclPrivateState, inclSResult,
                  inclNewPrivateStateEnc, inclNewStateHash, inclNewHistoryStateBinding
          (*****************************************************************)
          (* We assume the antecedent in this conjunction.                 *)
          (*****************************************************************)
          <5>4. HAVE LL1NVRAM.historySummary' = Hash(historySummary, input)
          (*****************************************************************)
          (* We prove that all of the authenticators in the set of         *)
          (* observed authenticators fail to validate the history state    *)
          (* binding defined by the inclusion invariant in the unprimed    *)
          (* state.                                                        *)
          (*****************************************************************)
          <5>5. \A authenticator \in LL1ObservedAuthenticators :
                    ~ValidateMAC(LL1NVRAM.symmetricKey, inclHistoryStateBinding, authenticator)
            (***************************************************************)
            (* We will make use of the conjunct in LL1RestrictedCorruption *)
            (* that prevents the garbage history summary from being a      *)
            (* predecessor to any history summary in an authenticated      *)
            (* history state binding.  This conjunct states a 4-way        *)
            (* universally quantifiied predicate.                          *)
            (***************************************************************)
            <6>1. LL1RestrictedCorruption!nvram!previous(garbageHistorySummary)
              BY <4>1
            (***************************************************************)
            (* We will pare the predicate down to a single universal       *)
            (* quantification by showing particular instances for three of *)
            (* the quantifiers.                                            *)
            (***************************************************************)
            <6>2. inclStateHash \in HashType
              BY <5>2
            <6>3. historySummary \in HashType
              BY <5>1
            <6>4. input \in InputType
              BY <5>1
            (***************************************************************)
            (* We also need to show that the antecedent of the implication *)
            (* in this predicate is satisfied.                             *)
            (***************************************************************)
            <6>5. garbageHistorySummary = Hash(historySummary, input)
              BY <4>2, <5>4
            (***************************************************************)
            (* The consequent of the implication in this predicate is      *)
            (* exactly what we need for our conclusion.                    *)
            (***************************************************************)
            <6>6. QED
              BY <6>1, <6>2, <6>3, <6>4, <6>5 DEF inclHistoryStateBinding
          (*****************************************************************)
          (* We show that the symmetric key in the NVRAM has not changed,  *)
          (* nor has the set of observed authenticators.                   *)
          (*****************************************************************)
          <5>6. UNCHANGED LL1NVRAM.symmetricKey
            BY <2>1, SymmetricKeyConstantLemma
          <5>7. UNCHANGED LL1ObservedAuthenticators
            BY <3>8 DEF  LL1RestrictedCorruption
          (*****************************************************************)
          (* Thus, we can conclude that all of the authenticators in the   *)
          (* set of observed authenticators fail to validate the history   *)
          (* state binding defined by the inclusion invariant in the       *)
          (* primed state.  This satisfies the definition of               *)
          (* LL1HistoryStateBindingAuthenticated.                          *)
          (*****************************************************************)
          <5>8. QED
            BY <5>5, <5>6, <5>7 DEF LL1HistoryStateBindingAuthenticated
        (*******************************************************************)
        (* We next prove the CardinalityInvariant.                         *)
        (*******************************************************************)
        <4>4. CardinalityInvariant'
          (*****************************************************************)
          (* To prove the universally quantified expression, we take a new *)
          (* set of variables of the appropriate types.  For the TAKE step *)
          (* to be meaningful to the prover, first we have to tell the     *)
          (* prover to expand the definition of CardinalityInvariant, so   *)
          (* it will see the universally quantified expression therein.    *)
          (*****************************************************************)
          <5> USE DEF CardinalityInvariant
          <5>1. TAKE historySummary \in HashType, stateHash \in HashType
          (*****************************************************************)
          (* The CardinalityInvariant states an implication.  To prove     *)
          (* this, it suffices to prove that the the antecedent is false.  *)
          (* The antecedent is a conjunction.  We will prove that the      *)
          (* first conjunct is false.                                      *)
          (*****************************************************************)
          <5>2. SUFFICES ASSUME TRUE PROVE ~LL1NVRAMHistorySummaryUncorrupted'
            OBVIOUS
          (*****************************************************************)
          (* We then hide the definition.                                  *)
          (*****************************************************************)
          <5> HIDE DEF CardinalityInvariant
          (*****************************************************************)
          (* We will make use of the conjunct in LL1RestrictedCorruption   *)
          (* that prevents the garbage history summary from being in an    *)
          (* authenticated history state binding.                          *)
          (*****************************************************************)
          <5>3. LL1RestrictedCorruption!nvram!current(garbageHistorySummary)
            BY <4>1
          (*****************************************************************)
          (* The following equivalence, plus the knowledge that the        *)
          (* symmetric key in the NVRAM and the set of observed            *)
          (* authenticators have not changed, are sufficient to prove the  *)
          (* conclusion.                                                   *)
          (*****************************************************************)
          <5>4. LL1NVRAM.historySummary' = garbageHistorySummary
            BY <4>2
          <5>5. UNCHANGED LL1NVRAM.symmetricKey
            BY <2>1, SymmetricKeyConstantLemma
          <5>6. UNCHANGED LL1ObservedAuthenticators
            BY <3>8 DEF  LL1RestrictedCorruption
          <5>7. QED
            BY <5>3, <5>4, <5>5, <5>6
            DEF LL1NVRAMHistorySummaryUncorrupted, LL1HistoryStateBindingAuthenticated
        (*******************************************************************)
        (* We last prove the UniquenessInvariant.                          *)
        (*******************************************************************)
        <4>5. UniquenessInvariant'
          (*****************************************************************)
          (* To prove the universally quantified expression, we take a new *)
          (* set of variables of the appropriate types.  For the TAKE step *)
          (* to be meaningful to the prover, first we have to tell the     *)
          (* prover to expand the definition of UniquenessInvariant, so it *)
          (* will see the universally quantified expression therein.       *)
          (*****************************************************************)
          <5> USE DEF UniquenessInvariant
          <5>1. TAKE stateHash1, stateHash2 \in HashType
          (*****************************************************************)
          (* To simplify the writing of the proof, we re-state the         *)
          (* definitions from the UniquenessInvariant.                     *)
          (*****************************************************************)
          <5> uniqHistoryStateBinding1 == Hash(LL1NVRAM.historySummary, stateHash1)
          <5> uniqHistoryStateBinding2 == Hash(LL1NVRAM.historySummary, stateHash2)
          (*****************************************************************)
          (* The UniquenessInvariant states an implication.  To prove      *)
          (* this, it suffices to prove that the the antecedent is false.  *)
          (* The antecedent is a conjunction.  We will prove that the      *)
          (* first conjunct is false.                                      *)
          (*****************************************************************)
          <5>2. SUFFICES
                ASSUME TRUE
                PROVE ~LL1HistoryStateBindingAuthenticated(uniqHistoryStateBinding1)'
            OBVIOUS
          (*****************************************************************)
          (* We hide the definitions of UniquenessInvariant and the        *)
          (* definitions from the UniquenessInvariant.                     *)
          (*****************************************************************)
          <5> HIDE DEF UniquenessInvariant
          <5> HIDE DEF uniqHistoryStateBinding1, uniqHistoryStateBinding2
          (*****************************************************************)
          (* We will make use of the conjunct in LL1RestrictedCorruption   *)
          (* that prevents the garbage history summary from being in an    *)
          (* authenticated history state binding.  This conjunct states a  *)
          (* 2-way universally quantifiied predicate.                      *)
          (*****************************************************************)
          <5>3. LL1RestrictedCorruption!nvram!current(garbageHistorySummary)
            BY <4>1
          (*****************************************************************)
          (* We will pare the predicate down to a single universal         *)
          (* quantification by showing particular instances for one of the *)
          (* quantifiers.                                                  *)
          (*****************************************************************)
          <5>4. stateHash1 \in HashType
            BY <5>1
          (*****************************************************************)
          (* The following equivalence, plus the knowledge that the        *)
          (* symmetric key in the NVRAM and the set of observed            *)
          (* authenticators have not changed, are sufficient to prove the  *)
          (* conclusion.                                                   *)
          (*****************************************************************)
          <5>5. LL1NVRAM.historySummary' = garbageHistorySummary
            BY <4>2
          <5>6. UNCHANGED LL1NVRAM.symmetricKey
            BY <2>1, SymmetricKeyConstantLemma
          <5>7. UNCHANGED LL1ObservedAuthenticators
            BY <3>8 DEF  LL1RestrictedCorruption
          <5>8. QED
            BY <5>3, <5>4, <5>5, <5>6, <5>7
            DEF uniqHistoryStateBinding1, LL1HistoryStateBindingAuthenticated
        <4>6. QED
          BY <4>3, <4>4, <4>5
      <3>9. QED
        BY <2>3, <3>1, <3>2, <3>3, <3>4, <3>5, <3>6, <3>7, <3>8 DEF LL1Next
    <2>4. QED
      BY <2>1, <2>2, <2>3
  <1>5. QED
    (***********************************************************************)
    (* Using the Inv1 proof rule, the base case and the induction step     *)
    (* together imply that the invariant always holds.                     *)
    (***********************************************************************)
    <2>1. ( /\ InclusionInvariant
            /\ CardinalityInvariant
            /\ UniquenessInvariant
            /\ [][LL1Next]_LL1Vars
            /\ []LL1TypeInvariant )
        =>
            []InclusionInvariant /\ []CardinalityInvariant /\ []UniquenessInvariant
      BY <1>4, Inv1
    <2>2. QED
      BY <1>1, <1>2, <1>3, <2>1 DEF LL1Spec

----

(***************************************************************************)
(* From the above proof and the previous proof that the                    *)
(* UnforgeabilityInvariant is an inductive invariant of the `^\basicname^' *)
(* spec, it proves that the set of CorrectnessInvariants are inductive     *)
(* invariants of the `^\basicname^' spec.                                  *)
(***************************************************************************)

THEOREM CorrectnessInvariance == LL1Spec => []CorrectnessInvariants \*`^\label{THEOREM-CorrectnessInvariance}^'
  <1>1. LL1Spec => []UnforgeabilityInvariant
    BY UnforgeabilityInvariance
  <1>2. LL1Spec => []InclusionInvariant /\ []UniquenessInvariant
    BY InclusionCardinalityUniquenessInvariance
  <1>3. UnforgeabilityInvariant /\ InclusionInvariant /\ UniquenessInvariant => CorrectnessInvariants
    BY DEF CorrectnessInvariants
  <1>4. QED
    BY <1>1, <1>2, <1>3

====
