---- MODULE MemoirLL1UnforgeabilityInvariance ----

(***************************************************************************)
(* This module proves that the UnforgeabilityInvariant is an inductive     *)
(* invariant of the `^\basicname^' spec.                                   *)
(***************************************************************************)

EXTENDS MemoirLL1InvarianceLemmas

(***************************************************************************)
(* Because the spec allows the data on the disk to be read into the RAM,   *)
(* proving UnforgeabilityInvariance of the RAM requires also proving an    *)
(* analagous property for the disk.  Thus, we first prove the that the     *)
(* ExtendedUnforgeabilityInvariant is an invariant of the spec.            *)
(***************************************************************************)

THEOREM ExtendedUnforgeabilityInvariance == LL1Spec => []ExtendedUnforgeabilityInvariant \*`^\label{THEOREM-ExtendedUnforgeabilityInvariance}^'
  (*************************************************************************)
  (* This proof will require the LL1TypeInvariant.  Fortunately, the       *)
  (* LL1TypeSafe theorem has already proven that the `^\basicname^' spec   *)
  (* satisfies its type invariant.                                         *)
  (*************************************************************************)
  <1>1. LL1Spec => []LL1TypeInvariant
    BY LL1TypeSafe
  (*************************************************************************)
  (* The top level of the proof is boilerplate TLA+ for an Inv1-style      *)
  (* proof.  First, we prove that the initial state satisfies              *)
  (* ExtendedUnforgeabilityInvariant.  Second, we prove that the LL1Next   *)
  (* predicate inductively preserves ExtendedUnforgeabilityInvariant.      *)
  (* Third, we use temporal induction to prove that these two conditions   *)
  (* satisfy the ExtendedUnforgeabilityInvariant over all behaviors.       *)
  (*************************************************************************)
  <1>2. LL1Init /\ LL1TypeInvariant => ExtendedUnforgeabilityInvariant
    (***********************************************************************)
    (* First, we assume the antecedents.                                   *)
    (***********************************************************************)
    <2>1. HAVE LL1Init /\ LL1TypeInvariant
    (***********************************************************************)
    (* Then, we pick some symmetrick key for which LL1Init is true.        *)
    (***********************************************************************)
    <2>2. PICK symmetricKey \in SymmetricKeyType : LL1Init!(symmetricKey)!1
      BY <2>1 DEF LL1Init
    (***********************************************************************)
    (* To simplify the writing of the proof, we re-state some of the       *)
    (* definitions from LL1Init.  We don't need all of them for this       *)
    (* proof, so we only re-state the ones we need.                        *)
    (***********************************************************************)
    <2> initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
    <2> initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
    <2> initialHistoryStateBinding == Hash(BaseHashValue, initialStateHash)
    <2> initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
    (***********************************************************************)
    (* We hide the definitions, so they don't overwhelm the prover.  We'll *)
    (* pull them in as necessary below.                                    *)
    (***********************************************************************)
    <2> HIDE DEF initialPrivateStateEnc, initialStateHash, initialHistoryStateBinding,
            initialAuthenticator
    (***********************************************************************)
    (* To prove the universally quantified expression, we take a new hash. *)
    (* For the TAKE step to be meaningful to the prover, first we have to  *)
    (* tell the prover to expand the definition of                         *)
    (* ExtendedUnforgeabilityInvariant, so it will see the universally     *)
    (* quantified expression therein.                                      *)
    (***********************************************************************)
    <2> USE DEF ExtendedUnforgeabilityInvariant
    <2>3. TAKE historyStateBinding \in HashType
    (***********************************************************************)
    (* We will prove each of the conjuncts separately.  Following is the   *)
    (* proof of the unforgeability invariant for the RAM.  It follows      *)
    (* directly from the definition of LL1Init.                            *)
    (***********************************************************************)
    <2>4. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator) =>
              LL1RAM.authenticator \in LL1ObservedAuthenticators
      <3>1. HAVE ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
      <3>2. LL1RAM.authenticator = initialAuthenticator
        BY <2>2 DEF initialAuthenticator, initialHistoryStateBinding,
                    initialStateHash, initialPrivateStateEnc
      <3>3. LL1ObservedAuthenticators = {initialAuthenticator}
        BY <2>2 DEF initialAuthenticator, initialHistoryStateBinding,
                    initialStateHash, initialPrivateStateEnc
      <3>4. QED
        BY <3>2, <3>3
    (***********************************************************************)
    (* Following is the proof of the unforgeability invariant for the      *)
    (* disk.  It follows directly from the definition of LL1Init.          *)
    (***********************************************************************)
    <2>5. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1Disk.authenticator) =>
              LL1Disk.authenticator \in LL1ObservedAuthenticators
      <3>1. HAVE ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1Disk.authenticator)
      <3>2. LL1Disk.authenticator = initialAuthenticator
        BY <2>2 DEF initialAuthenticator, initialHistoryStateBinding,
                    initialStateHash, initialPrivateStateEnc
      <3>3. LL1ObservedAuthenticators = {initialAuthenticator}
        BY <2>2 DEF initialAuthenticator, initialHistoryStateBinding,
                    initialStateHash, initialPrivateStateEnc
      <3>4. QED
        BY <3>2, <3>3
    <2>6. QED
      BY <2>4, <2>5
  (*************************************************************************)
  (* For the induction step, we will need the type invariant to be true in *)
  (* both the unprimed and primed states.                                  *)
  (*************************************************************************)
  <1>3. ( /\  ExtendedUnforgeabilityInvariant
          /\  [LL1Next]_LL1Vars
          /\  LL1TypeInvariant
          /\  LL1TypeInvariant' )
      =>
          ExtendedUnforgeabilityInvariant'
    (***********************************************************************)
    (* First, we assume the antecedents.                                   *)
    (***********************************************************************)
    <2>1. HAVE ExtendedUnforgeabilityInvariant /\ [LL1Next]_LL1Vars /\ LL1TypeInvariant /\ LL1TypeInvariant'
    (***********************************************************************)
    (* The induction step includes two cases: stuttering and LL1Next       *)
    (* actions.  The stuttering case is a straightforward application of   *)
    (* the LL1RAMUnforgeabilityUnchangedLemma and the                      *)
    (* LL1DiskUnforgeabilityUnchangedLemma.                                *)
    (***********************************************************************)
    <2>2. CASE UNCHANGED LL1Vars
      <3>1. \A historyStateBinding \in HashType :
                ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator') =>
                    LL1RAM.authenticator' \in LL1ObservedAuthenticators'
        <4>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
          BY <2>1
        <4>2. UNCHANGED LL1RAM
          BY <2>2 DEF LL1Vars
        <4>3. UNCHANGED LL1ObservedAuthenticators
          BY <2>2 DEF LL1Vars
        <4>4. UNCHANGED LL1NVRAM.symmetricKey
          <5>1. UNCHANGED LL1NVRAM
            BY <2>2 DEF LL1Vars
          <5>2. QED
            BY <5>1
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4, LL1RAMUnforgeabilityUnchangedLemma
      <3>2. \A historyStateBinding \in HashType :
                ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator') =>
                    LL1Disk.authenticator' \in LL1ObservedAuthenticators'
        <4>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
          BY <2>1
        <4>2. UNCHANGED LL1Disk
          BY <2>2 DEF LL1Vars
        <4>3. UNCHANGED LL1ObservedAuthenticators
          BY <2>2 DEF LL1Vars
        <4>4. UNCHANGED LL1NVRAM.symmetricKey
          <5>1. UNCHANGED LL1NVRAM
            BY <2>2 DEF LL1Vars
          <5>2. QED
            BY <5>1
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4, LL1DiskUnforgeabilityUnchangedLemma
      <3>3. QED
        BY <3>1, <3>2 DEF ExtendedUnforgeabilityInvariant
    (***********************************************************************)
    (* We break down the LL1Next case into eight separate cases for each   *)
    (* action.                                                             *)
    (***********************************************************************)
    <2>3. CASE LL1Next
      (*********************************************************************)
      (* The LL1MakeInputAvailable case is a straightforward application   *)
      (* of the LL1RAMUnforgeabilityUnchangedLemma and the                 *)
      (* LL1DiskUnforgeabilityUnchangedLemma.                              *)
      (*********************************************************************)
      <3>1. CASE LL1MakeInputAvailable
        <4>1. PICK input \in InputType : LL1MakeInputAvailable!(input)
          BY <3>1 DEF LL1MakeInputAvailable
        <4>2. \A historyStateBinding \in HashType :
                  ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator') =>
                      LL1RAM.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1RAM
            BY <4>1
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <4>1
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <4>1
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1RAMUnforgeabilityUnchangedLemma
        <4>3. \A historyStateBinding \in HashType :
                  ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator') =>
                      LL1Disk.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <4>1
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <4>1
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1DiskUnforgeabilityUnchangedLemma
        <4>4. QED
          BY <4>2, <4>3 DEF ExtendedUnforgeabilityInvariant
      (*********************************************************************)
      (* The LL1PerformOperation case is not terribly involved, but we     *)
      (* have to treat the RAM and disk conjuncts separately.              *)
      (*********************************************************************)
      <3>2. CASE LL1PerformOperation
        (*******************************************************************)
        (* We pick some input for which LL1PerformOperation is true.       *)
        (*******************************************************************)
        <4>1. PICK input \in LL1AvailableInputs : LL1PerformOperation!(input)!1 
          BY <3>2 DEF LL1PerformOperation
        (*******************************************************************)
        (* To simplify the writing of the proof, we re-state the           *)
        (* definitions from the LL1PerformOperation action.                *)
        (*******************************************************************)
        <4> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
        <4> historyStateBinding == Hash(LL1NVRAM.historySummary, stateHash)
        <4> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
        <4> sResult == Service(LL1RAM.publicState, privateState, input)
        <4> newPrivateStateEnc ==
                SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
        <4> newHistorySummary == Hash(LL1NVRAM.historySummary, input)
        <4> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
        <4> newHistoryStateBinding == Hash(newHistorySummary, newStateHash)
        <4> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
        (*******************************************************************)
        (* We hide the definitions, so they don't overwhelm the prover.    *)
        (* We'll pull them in as necessary below.                          *)
        (*******************************************************************)
        <4> HIDE DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
                newHistorySummary, newStateHash, newHistoryStateBinding, newAuthenticator
        (*******************************************************************)
        (* To prove the universally quantified expression, we take a new   *)
        (* hash.  For the TAKE step to be meaningful to the prover, first  *)
        (* we have to tell the prover to expand the definition of          *)
        (* ExtendedUnforgeabilityInvariant, so it will see the universally *)
        (* quantified expression therein.                                  *)
        (*******************************************************************)
        <4> USE DEF ExtendedUnforgeabilityInvariant
        <4>2. TAKE historyStateBinding1 \in HashType
        (*******************************************************************)
        (* Before proceeding to prove each of the conjuncts, we prove a    *)
        (* statement that will be useful in both of the sub-proofs below.  *)
        (* Namely, the new authenticator generated by the                  *)
        (* LL1PerformOperation action is unioned into the set of observed  *)
        (* authenticators.                                                 *)
        (*******************************************************************)
        <4>3. LL1ObservedAuthenticators' =
                  LL1ObservedAuthenticators \union {newAuthenticator}
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                   newHistorySummary, newPrivateStateEnc, sResult, privateState
        (*******************************************************************)
        (* For the RAM portion of the unforgeability invariant, we note    *)
        (* that the LL1PerformOperation action updates the authenticator   *)
        (* in the RAM with the new authenticator.  Since this new          *)
        (* authenticator is unioned into the set of observed               *)
        (* authenticators, the invariant holds in the primed state.        *)
        (*******************************************************************)
        <4>4. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding1, LL1RAM.authenticator') =>
                  LL1RAM.authenticator' \in LL1ObservedAuthenticators'
          <5>1. LL1RAM.authenticator' = newAuthenticator
            <6>1. newAuthenticator = LL1PerformOperation!(input)!newAuthenticator
              BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                     newHistorySummary, newPrivateStateEnc, sResult, privateState
            <6>2. LL1RAM' = [
                      publicState |-> LL1PerformOperation!(input)!sResult.newPublicState,
                      privateStateEnc |-> LL1PerformOperation!(input)!newPrivateStateEnc,
                      historySummary |-> LL1PerformOperation!(input)!newHistorySummary,
                      authenticator |-> newAuthenticator ]
              BY <4>1, <6>1
            <6>3. QED
              BY <6>2
          <5>2. QED
            BY <4>3, <5>1
        (*******************************************************************)
        (* For the disk portion of the unforgeability invariant, we employ *)
        (* the LL1DiskUnforgeabilityUnchangedLemma, since the disk is not  *)
        (* changed by the LL1PerformOperation action.                      *)
        (*******************************************************************)
        <4>5. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding1, LL1Disk.authenticator') =>
                  LL1Disk.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
            BY <4>3
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. LL1NVRAM' = [ historySummary |-> LL1PerformOperation!(input)!newHistorySummary,
                             symmetricKey |-> LL1NVRAM.symmetricKey ]
              BY <4>1
            <6>2. LL1NVRAM.symmetricKey' = LL1NVRAM.symmetricKey
              BY <6>1
            <6>3. QED
              BY <6>2
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1DiskUnforgeabilityUnchangedLemma
        <4>6. QED
          BY <4>4, <4>5
      (*********************************************************************)
      (* The LL1RepeatOperation case is not terribly involved, but we have *)
      (* to treat the RAM and disk conjuncts separately.                   *)
      (*********************************************************************)
      <3>3. CASE LL1RepeatOperation
        (*******************************************************************)
        (* We pick some input for which LL1RepeatOperation is true.        *)
        (*******************************************************************)
        <4>1. PICK input \in LL1AvailableInputs : LL1RepeatOperation!(input)!1 
          BY <3>3 DEF LL1RepeatOperation
        (*******************************************************************)
        (* To simplify the writing of the proof, we re-state the           *)
        (* definitions from the LL1RepeatOperation action.                 *)
        (*******************************************************************)
        <4> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
        <4> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
        <4> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
        <4> sResult == Service(LL1RAM.publicState, privateState, input)
        <4> newPrivateStateEnc ==
                SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
        <4> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
        <4> newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
        <4> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
        (*******************************************************************)
        (* We hide the definitions, so they don't overwhelm the prover.    *)
        (* We'll pull them in as necessary below.                          *)
        (*******************************************************************)
        <4> HIDE DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
                newStateHash, newHistoryStateBinding, newAuthenticator
        (*******************************************************************)
        (* To prove the universally quantified expression, we take a new   *)
        (* hash.  For the TAKE step to be meaningful to the prover, first  *)
        (* we have to tell the prover to expand the definition of          *)
        (* ExtendedUnforgeabilityInvariant, so it will see the universally *)
        (* quantified expression therein.                                  *)
        (*******************************************************************)
        <4> USE DEF ExtendedUnforgeabilityInvariant
        <4>2. TAKE historyStateBinding1 \in HashType
        (*******************************************************************)
        (* Before proceeding to prove each of the conjuncts, we prove a    *)
        (* statement that will be useful in both of the sub-proofs below.  *)
        (* Namely, the new authenticator generated by the                  *)
        (* LL1PerformOperation action is unioned into the set of observed  *)
        (* authenticators.                                                 *)
        (*******************************************************************)
        <4>3. LL1ObservedAuthenticators' =
                  LL1ObservedAuthenticators \union {newAuthenticator}
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                   newPrivateStateEnc, sResult, privateState
        (*******************************************************************)
        (* For the RAM portion of the unforgeability invariant, we note    *)
        (* that the LL1RepeatOperation action updates the authenticator in *)
        (* the RAM with the new authenticator.  Since this new             *)
        (* authenticator is unioned into the set of observed               *)
        (* authenticators, the invariant holds in the primed state.        *)
        (*******************************************************************)
        <4>4. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding1, LL1RAM.authenticator') =>
                  LL1RAM.authenticator' \in LL1ObservedAuthenticators'
          <5>1. LL1RAM.authenticator' = newAuthenticator
            <6>1. newAuthenticator = LL1RepeatOperation!(input)!newAuthenticator
              BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                     newPrivateStateEnc, sResult, privateState
            <6>2. LL1RAM' = [
                      publicState |-> LL1RepeatOperation!(input)!sResult.newPublicState,
                      privateStateEnc |-> LL1RepeatOperation!(input)!newPrivateStateEnc,
                      historySummary |-> LL1NVRAM.historySummary,
                      authenticator |-> newAuthenticator ]
              BY <4>1, <6>1
            <6>3. QED
              BY <6>2
          <5>2. QED
            BY <4>3, <5>1
        (*******************************************************************)
        (* For the disk portion of the unforgeability invariant, we employ *)
        (* the LL1DiskUnforgeabilityUnchangedLemma, since the disk is not  *)
        (* changed by the LL1RepeatOperation action.                       *)
        (*******************************************************************)
        <4>5. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding1, LL1Disk.authenticator') =>
                  LL1Disk.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
            BY <4>3
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <4>1
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1RAMUnforgeabilityUnchangedLemma
        <4>6. QED
          BY <4>4, <4>5
      (*********************************************************************)
      (* The LL1Restart case is moderately interesting.  The RAM portion   *)
      (* of the unforgeability invariant is where we pull in the           *)
      (* constraint on the set of authenticators that can wind up in the   *)
      (* randomized RAM after a LL1Restart.  This constraint is explicitly *)
      (* expressed in the definition of the LL1Restart action.  The disk   *)
      (* portion is a straightforward application of the                   *)
      (* LL1DiskUnforgeabilityUnchangedLemma.                              *)
      (*********************************************************************)
      <3>4. CASE LL1Restart
        (*******************************************************************)
        (* We pick some variables of the appropriate types for which       *)
        (* LL1Restart is true.                                             *)
        (*******************************************************************)
        <4>1. PICK untrustedStorage \in LL1UntrustedStorageType,
                   randomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                   hash \in HashType :
                   LL1Restart!(untrustedStorage, randomSymmetricKey, hash)
          BY <3>4 DEF LL1Restart
        (*******************************************************************)
        (* To prove the universally quantified expression, we take a new   *)
        (* hash.  For the TAKE step to be meaningful to the prover, first  *)
        (* we have to tell the prover to expand the definition of          *)
        (* ExtendedUnforgeabilityInvariant, so it will see the universally *)
        (* quantified expression therein.                                  *)
        (*******************************************************************)
        <4> USE DEF ExtendedUnforgeabilityInvariant
        <4>3. TAKE historyStateBinding \in HashType
        <4>4. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator') =>
                  LL1RAM.authenticator' \in LL1ObservedAuthenticators'
          (*****************************************************************)
          (* For the RAM portion, we first note that the definition of     *)
          (* LL1Restart tells us that the primed authenticator has been    *)
          (* constructed with a random symmetric key.  The antecedent of   *)
          (* the implication cannot be true, because the random symmetric  *)
          (* key does not match the symmetric key in the NVRAM.  We show   *)
          (* this using proof by contradiction.                            *)
          (*****************************************************************)
          <5>1. LL1RAM.authenticator' = GenerateMAC(randomSymmetricKey, hash)
            BY <4>1
          <5>2. ~ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator')
            <6>1. SUFFICES
                    ASSUME
                      ValidateMAC(
                          LL1NVRAM.symmetricKey',
                          historyStateBinding,
                          LL1RAM.authenticator')
                    PROVE FALSE
              BY <5>1
            <6>2. randomSymmetricKey = LL1NVRAM.symmetricKey'
              <7>1. ValidateMAC(
                        LL1NVRAM.symmetricKey',
                        historyStateBinding,
                        GenerateMAC(randomSymmetricKey, hash))
                BY <5>1, <6>1
              <7>2. randomSymmetricKey \in SymmetricKeyType
                BY <4>1
              <7>3. LL1NVRAM.symmetricKey' \in SymmetricKeyType
                <8>1. LL1TypeInvariant'
                  BY <2>1
                <8>2. QED
                  BY <8>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
              <7>4. hash \in HashType
                BY <4>1
              <7>5. historyStateBinding \in HashType
                BY <4>3
              <7>6. QED
                BY <7>1, <7>2, <7>3, <7>4, <7>5, MACUnforgeable
            <6>3. randomSymmetricKey # LL1NVRAM.symmetricKey'
              <7>1. randomSymmetricKey # LL1NVRAM.symmetricKey
                BY <4>1
              <7>2. UNCHANGED LL1NVRAM.symmetricKey
                BY <2>3, SymmetricKeyConstantLemma
              <7>3. QED
                BY <7>1, <7>2
            <6>4. QED
              BY <6>2, <6>3
          <5>3. QED
            BY <5>2
        (*******************************************************************)
        (* For the disk portion of the unforgeability invariant, we employ *)
        (* the LL1DiskUnforgeabilityUnchangedLemma, since the disk is not  *)
        (* changed by the LL1CorruptRAM action.                            *)
        (*******************************************************************)
        <4>5. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator') =>
                  LL1Disk.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <4>1
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <4>1
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1DiskUnforgeabilityUnchangedLemma
        <4>6. QED
          BY <4>4, <4>5
      (*********************************************************************)
      (* The LL1ReadDisk case is a straightforward application of the      *)
      (* LL1RAMUnforgeabilityUnchangedLemma and the                        *)
      (* LL1DiskUnforgeabilityUnchangedLemma.                              *)
      (*********************************************************************)
      <3>5. CASE LL1ReadDisk
        <4>1. \A historyStateBinding \in HashType :
                  ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator') =>
                      LL1RAM.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. LL1RAM' = LL1Disk
            BY <3>5 DEF LL1ReadDisk
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <3>5 DEF LL1ReadDisk
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <3>5 DEF LL1ReadDisk
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1RAMUnforgeabilityUnchangedLemma
        <4>2. \A historyStateBinding \in HashType :
                  ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator') =>
                      LL1Disk.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1Disk
            BY <3>5 DEF LL1ReadDisk
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <3>5 DEF LL1ReadDisk
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <3>5 DEF LL1ReadDisk
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1DiskUnforgeabilityUnchangedLemma
        <4>3. QED
          BY <4>1, <4>2 DEF ExtendedUnforgeabilityInvariant
      (*********************************************************************)
      (* The LL1WriteDisk case is a straightforward application of the     *)
      (* LL1RAMUnforgeabilityUnchangedLemma and the                        *)
      (* LL1DiskUnforgeabilityUnchangedLemma.                              *)
      (*********************************************************************)
      <3>6. CASE LL1WriteDisk
        <4>1. \A historyStateBinding \in HashType :
                  ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator') =>
                      LL1RAM.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1RAM
            BY <3>6 DEF LL1WriteDisk
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <3>6 DEF LL1WriteDisk
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <3>6 DEF LL1WriteDisk
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1RAMUnforgeabilityUnchangedLemma
        <4>2. \A historyStateBinding \in HashType :
                  ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator') =>
                      LL1Disk.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. LL1Disk' = LL1RAM
            BY <3>6 DEF LL1WriteDisk
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <3>6 DEF LL1WriteDisk
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <3>6 DEF LL1WriteDisk
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1DiskUnforgeabilityUnchangedLemma
        <4>3. QED
          BY <4>1, <4>2 DEF ExtendedUnforgeabilityInvariant
      (*********************************************************************)
      (* The LL1CorruptRAM case is moderately interesting.  The RAM        *)
      (* portion of the unforgeability invariant is where we pull in the   *)
      (* constraint on the set of authenticators the user can create.      *)
      (* This constraint is explicitly expressed in the definition of the  *)
      (* LL1CorruptRAM action.  The disk portion is a straightforward      *)
      (* application of the LL1DiskUnforgeabilityUnchangedLemma.           *)
      (*********************************************************************)
      <3>7. CASE LL1CorruptRAM
        (*******************************************************************)
        (* We pick some variables of the appropriate types for which       *)
        (* LL1CorruptRAM is true.                                          *)
        (*******************************************************************)
        <4>1. PICK untrustedStorage \in LL1UntrustedStorageType,
                   fakeSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                   hash \in HashType :
                   LL1CorruptRAM!(untrustedStorage, fakeSymmetricKey, hash)
          BY <3>7 DEF LL1CorruptRAM
        (*******************************************************************)
        (* To prove the universally quantified expression, we take a new   *)
        (* hash.  For the TAKE step to be meaningful to the prover, first  *)
        (* we have to tell the prover to expand the definition of          *)
        (* ExtendedUnforgeabilityInvariant, so it will see the universally *)
        (* quantified expression therein.                                  *)
        (*******************************************************************)
        <4> USE DEF ExtendedUnforgeabilityInvariant
        <4>3. TAKE historyStateBinding \in HashType
        <4>4. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator') =>
                  LL1RAM.authenticator' \in LL1ObservedAuthenticators'
          (*****************************************************************)
          (* For the RAM portion, we first note that the definition of     *)
          (* LL1CorruptRAM tells us that the primed authenticator in the   *)
          (* RAM is either in the unprimed set of observed authenticators  *)
          (* or has been constructed with a fake symmetric key.  We will   *)
          (* treat these two cases separately.                             *)
          (*****************************************************************)
          <5>1.  \/  LL1RAM.authenticator' \in LL1ObservedAuthenticators
                 \/  LL1RAM.authenticator' = GenerateMAC(fakeSymmetricKey, hash)
            BY <4>1
          (*****************************************************************)
          (* For the case in which the primed authenticator is in the      *)
          (* unprimed set of observed authenticators, the conclusion       *)
          (* directly follows because the set of observed authenticators   *)
          (* is not changed by the LL1CorruptRAM action.                   *)
          (*****************************************************************)
          <5>2. CASE LL1RAM.authenticator' \in LL1ObservedAuthenticators
            <6>1. HAVE ValidateMAC(
                      LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator')
            <6>2. UNCHANGED LL1ObservedAuthenticators
              BY <4>1
            <6>3. QED
              BY <5>2, <6>2
          (*****************************************************************)
          (* For the case in which the primed authenticator has been       *)
          (* constructed with a fake symmetric key, the antecedent of the  *)
          (* implication cannot be true, because the fake symmetric key    *)
          (* does not match the symmetric key in the NVRAM.  We show this  *)
          (* using proof by contradiction.                                 *)
          (*****************************************************************)
          <5>3. CASE LL1RAM.authenticator' = GenerateMAC(fakeSymmetricKey, hash)
            <6>1. ~ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator')
              <7>1. SUFFICES
                      ASSUME
                        ValidateMAC(
                            LL1NVRAM.symmetricKey',
                            historyStateBinding,
                            LL1RAM.authenticator')
                      PROVE FALSE
                BY <5>3
              <7>2. fakeSymmetricKey = LL1NVRAM.symmetricKey'
                <8>1. ValidateMAC(
                          LL1NVRAM.symmetricKey',
                          historyStateBinding,
                          GenerateMAC(fakeSymmetricKey, hash))
                  BY <5>3, <7>1
                <8>2. fakeSymmetricKey \in SymmetricKeyType
                  BY <4>1
                <8>3. LL1NVRAM.symmetricKey' \in SymmetricKeyType
                  <9>1. LL1TypeInvariant'
                    BY <2>1
                  <9>2. QED
                    BY <9>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
                <8>4. hash \in HashType
                  BY <4>1
                <8>5. historyStateBinding \in HashType
                  BY <4>3
                <8>6. QED
                  BY <8>1, <8>2, <8>3, <8>4, <8>5, MACUnforgeable
              <7>3. fakeSymmetricKey # LL1NVRAM.symmetricKey'
                <8>1. fakeSymmetricKey # LL1NVRAM.symmetricKey
                  BY <4>1
                <8>2. UNCHANGED LL1NVRAM.symmetricKey
                  BY <2>3, SymmetricKeyConstantLemma
                <8>3. QED
                  BY <8>1, <8>2
              <7>4. QED
                BY <7>2, <7>3
            <6>2. QED
              BY <6>1
          <5>4. QED
            BY <5>1, <5>2, <5>3
        (*******************************************************************)
        (* For the disk portion of the unforgeability invariant, we employ *)
        (* the LL1DiskUnforgeabilityUnchangedLemma, since the disk is not  *)
        (* changed by the LL1CorruptRAM action.                            *)
        (*******************************************************************)
        <4>5. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator') =>
                  LL1Disk.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1Disk
            BY <4>1
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <4>1
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            <6>1. UNCHANGED LL1NVRAM
              BY <4>1
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1DiskUnforgeabilityUnchangedLemma
        <4>6. QED
          BY <4>4, <4>5
      (*********************************************************************)
      (* The LL1RestrictedCorruption case is moderately interesting.  For  *)
      (* the RAM portion, we have two cases depending on whether the       *)
      (* LL1RestrictedCorruption action leaves the RAM unchanged or        *)
      (* trashes it in the same way an LL1Restart action does.  For the    *)
      (* latter case, we pull in the constraint on the set of state        *)
      (* authenticators that can wind up in the randomized RAM after the   *)
      (* action completes.  The disk portion is a straightforward          *)
      (* application of the LL1DiskUnforgeabilityUnchangedLemma.           *)
      (*********************************************************************)
      <3>8. CASE LL1RestrictedCorruption
        (*******************************************************************)
        (* To prove the universally quantified expression, we take a new   *)
        (* hash.  For the TAKE step to be meaningful to the prover, first  *)
        (* we have to tell the prover to expand the definition of          *)
        (* ExtendedUnforgeabilityInvariant, so it will see the universally *)
        (* quantified expression therein.                                  *)
        (*******************************************************************)
        <4> USE DEF ExtendedUnforgeabilityInvariant
        <4>1. TAKE historyStateBinding \in HashType
        <4>2. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator') =>
                  LL1RAM.authenticator' \in LL1ObservedAuthenticators'
          (*****************************************************************)
          (* First we consider the case in which the RAM is unchanged.     *)
          (* This is straightforward                                       *)
          (*****************************************************************)
          <5>1. CASE LL1RestrictedCorruption!ram!unchanged
            <6>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
              BY <2>1
            <6>2. UNCHANGED LL1RAM
              BY <6>1
            <6>3. UNCHANGED LL1ObservedAuthenticators
              BY <3>8 DEF LL1RestrictedCorruption
            <6>4. UNCHANGED LL1NVRAM.symmetricKey
              BY <2>3, SymmetricKeyConstantLemma
            <6>5. QED
              BY <6>1, <6>2, <6>3, <6>4, LL1RAMUnforgeabilityUnchangedLemma
          (*****************************************************************)
          (* Next we consider the case in which the RAM is trashed.        *)
          (*****************************************************************)
          <5>2. CASE LL1RestrictedCorruption!ram!trashed
            (***************************************************************)
            (* We pick some variables of the appropriate types for which   *)
            (* LL1RestrictedCorruption is true.                            *)
            (***************************************************************)
            <6>1. PICK untrustedStorage \in LL1UntrustedStorageType,
                      randomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                      hash \in HashType :
                      LL1RestrictedCorruption!ram!trashed!(
                          untrustedStorage, randomSymmetricKey, hash)
              BY <5>2
            (***************************************************************)
            (* The primed authenticator has been constructed with a random *)
            (* symmetric key.  The antecedent of the implication cannot be *)
            (* true, because the random symmetric key does not match the   *)
            (* symmetric key in the NVRAM.  We show this using proof by    *)
            (* contradiction.                                              *)
            (***************************************************************)
            <6>2. LL1RAM.authenticator' = GenerateMAC(randomSymmetricKey, hash)
              BY <6>1
            <6>3. ~ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator')
              <7>1. SUFFICES
                    ASSUME
                        ValidateMAC(
                            LL1NVRAM.symmetricKey',
                            historyStateBinding,
                            LL1RAM.authenticator')
                    PROVE FALSE
                BY <6>1
              <7>2. randomSymmetricKey = LL1NVRAM.symmetricKey'
                <8>1. ValidateMAC(
                          LL1NVRAM.symmetricKey',
                          historyStateBinding,
                          GenerateMAC(randomSymmetricKey, hash))
                  BY <6>1, <7>1
                <8>2. randomSymmetricKey \in SymmetricKeyType
                  BY <6>1
                <8>3. LL1NVRAM.symmetricKey' \in SymmetricKeyType
                  <9>1. LL1TypeInvariant'
                    BY <2>1
                  <9>2. QED
                    BY <9>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
                <8>4. hash \in HashType
                  BY <6>1
                <8>5. historyStateBinding \in HashType
                  BY <6>3
                <8>6. QED
                  BY <8>1, <8>2, <8>3, <8>4, <8>5, MACUnforgeable
              <7>3. randomSymmetricKey # LL1NVRAM.symmetricKey'
                <8>1. randomSymmetricKey # LL1NVRAM.symmetricKey
                  BY <6>1
                <8>2. UNCHANGED LL1NVRAM.symmetricKey
                  BY <2>3, SymmetricKeyConstantLemma
                <8>3. QED
                  BY <8>1, <8>2
              <7>4. QED
                BY <7>2, <7>3
            <6>4. QED
              BY <6>3
          <5>3. QED
            BY <3>8, <5>1, <5>2 DEF LL1RestrictedCorruption
        (*******************************************************************)
        (* For the disk portion of the unforgeability invariant, we employ *)
        (* the LL1DiskUnforgeabilityUnchangedLemma, since the disk is not  *)
        (* changed by the LL1CorruptRAM action.                            *)
        (*******************************************************************)
        <4>3. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator') =>
                  LL1Disk.authenticator' \in LL1ObservedAuthenticators'
          <5>1. ExtendedUnforgeabilityInvariant /\ LL1TypeInvariant /\ LL1TypeInvariant'
            BY <2>1
          <5>2. UNCHANGED LL1Disk
            BY <3>8 DEF LL1RestrictedCorruption
          <5>3. UNCHANGED LL1ObservedAuthenticators
            BY <3>8 DEF LL1RestrictedCorruption
          <5>4. UNCHANGED LL1NVRAM.symmetricKey
            BY <2>3, SymmetricKeyConstantLemma
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, LL1DiskUnforgeabilityUnchangedLemma
        <4>4. QED
          BY <4>2, <4>3

      <3>9. QED
        BY <2>3, <3>1, <3>2, <3>3, <3>4, <3>5, <3>6, <3>7, <3>8 DEF LL1Next
    <2>4. QED
      BY <2>1, <2>2, <2>3
  <1>4. QED
    (***********************************************************************)
    (* Using the Inv1 proof rule, the base case and the induction step     *)
    (* together imply that the invariant always holds.                     *)
    (***********************************************************************)
    <2>1. ( /\  ExtendedUnforgeabilityInvariant
            /\  [][LL1Next]_LL1Vars
            /\  []LL1TypeInvariant )
        =>
            []ExtendedUnforgeabilityInvariant
      BY <1>3, Inv1
    <2>2. QED
      BY <1>1, <1>2, <2>1 DEF LL1Spec

----

(***************************************************************************)
(* The UnforgeabilityInvariant follows directly from the                   *)
(* ExtendedUnforgeabilityInvariant.                                        *)
(***************************************************************************)

THEOREM UnforgeabilityInvariance == LL1Spec => []UnforgeabilityInvariant \*`^\label{THEOREM-UnforgeabilityInvariance}^'
  <1>1. LL1Spec => []ExtendedUnforgeabilityInvariant
    BY ExtendedUnforgeabilityInvariance
  <1>2. ExtendedUnforgeabilityInvariant => UnforgeabilityInvariant
    BY DEF ExtendedUnforgeabilityInvariant, UnforgeabilityInvariant
  <1>3. QED
    BY <1>1, <1>2

====
