---- MODULE MemoirLL2Refinement ----

(***************************************************************************)
(* This module describes how to interpret the state of `^\fastname^' as a  *)
(* state of `^\basicname^'.                                                *)
(*                                                                         *)
(* This module includes the following definitions:                         *)
(*     LL2HistorySummaryIsSuccessor                                        *)
(*     HistorySummariesMatch                                               *)
(*     AuthenticatorsMatch                                                 *)
(*     AuthenticatorSetsMatch                                              *)
(*     LL2NVRAMLogicalHistorySummary                                       *)
(*     LL2Refinement                                                       *)
(***************************************************************************)

EXTENDS MemoirLL2Specification, Sequences

(***************************************************************************)
(* The LL2HistorySummaryIsSuccessor predicate defines, in the              *)
(* `^\fastname^' spec, the conditions under which one history summary is a *)
(* successor of another history summary with a particular intervening      *)
(* input.                                                                  *)
(***************************************************************************)

LL2HistorySummaryIsSuccessor(historySummary, previousHistorySummary, input, hashBarrier) == \*`^\label{DEF-LL2HistorySummaryIsSuccessor}^'
    LET
        successorHistorySummary == Successor(previousHistorySummary, input, hashBarrier)
        checkpointedSuccessorHistorySummary == Checkpoint(successorHistorySummary)
    IN
        \/ historySummary = successorHistorySummary
        \/ historySummary = checkpointedSuccessorHistorySummary

----

(***************************************************************************)
(* The HistorySummariesMatch predicate defines the conditions under which  *)
(* a history summary in the `^\basicname^' spec semantically matches a     *)
(* history summary in the `^\fastname^' spec.                              *)
(*                                                                         *)
(* This requires a recursive definition, but the current version of the    *)
(* prover cannot handle recursive operators, nor can it tractably support  *)
(* proofs using recursive function definitions.  Therefore, we define the  *)
(* operator indirectly, by using an assumption.  Although Lamport has      *)
(* stated that this approach is ``not a satisfactory alternative to        *)
(* recursive definitions,'' he has also called it ``a reasonable hack to   *)
(* get a proof done.''                                                     *)
(***************************************************************************)

CONSTANT HistorySummariesMatch(_,_,_) \*`^\label{CONSTANT-HistorySummariesMatch}^'

HistorySummariesMatchRecursion(ll1HistorySummary, ll2HistorySummary, hashBarrier) == \*`^\label{DEF-HistorySummariesMatchRecursion}^'
    \E  input \in InputType,
        previousLL1HistorySummary \in HashType,
        previousLL2HistorySummary \in HistorySummaryType :
        /\  HistorySummariesMatch(previousLL1HistorySummary, previousLL2HistorySummary, hashBarrier)
        /\  ll1HistorySummary = Hash(previousLL1HistorySummary, input)
        /\  LL2HistorySummaryIsSuccessor(
                ll2HistorySummary, previousLL2HistorySummary, input, hashBarrier)

ASSUME HistorySummariesMatchDefinition == \*`^\label{ASSUME-HistorySummariesMatchDefinition}^'
    \A  ll1HistorySummary \in HashType,
        ll2HistorySummary \in HistorySummaryType,
        hashBarrier \in HashType :
        LET
            ll2InitialHistorySummary == [ anchor |-> BaseHashValue, extension |-> BaseHashValue ]
        IN
            IF ll2HistorySummary = ll2InitialHistorySummary
            THEN
                HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, hashBarrier) =
                    ( ll1HistorySummary = BaseHashValue )
            ELSE
                HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, hashBarrier) =
                    HistorySummariesMatchRecursion(ll1HistorySummary, ll2HistorySummary, hashBarrier)

----

(***************************************************************************)
(* The AuthenticatorsMatch predicate defines the conditions under which an *)
(* authenticator in the `^\basicname^' spec semantically matches an        *)
(* authenticator in the `^\fastname^' spec.                                *)
(***************************************************************************)

AuthenticatorsMatch(ll1Authenticator, ll2Authenticator, symmetricKey, hashBarrier) == \*`^\label{DEF-AuthenticatorsMatch}^'
    \E  stateHash \in HashType,
        ll1HistorySummary \in HashType,
        ll2HistorySummary \in HistorySummaryType :
        LET
            ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash)
            ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
            ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash)
        IN
                (***********************************************************)
                (* The `^\fastname^' authenticator is a valid MAC of an    *)
                (* `^\fastname^' history state binding that binds some     *)
                (* `^\fastname^' history summary to some state hash.       *)
                (***********************************************************)
            /\  ValidateMAC(symmetricKey, ll2HistoryStateBinding, ll2Authenticator)
                (***********************************************************)
                (* The `^\basicname^' authenticator is generated as a MAC  *)
                (* of a `^\basicname^' history state binding that binds    *)
                (* some `^\basicname^' history summary to the same state   *)
                (* hash as the previous conjunct.                          *)
                (***********************************************************)
            /\  ll1Authenticator = GenerateMAC(symmetricKey, ll1HistoryStateBinding)
                (***********************************************************)
                (* The `^\basicname^' history state binding matches the    *)
                (* `^\fastname^' history state binding.                    *)
                (***********************************************************)
            /\  HistorySummariesMatch(ll1HistorySummary, ll2HistorySummary, hashBarrier)

----

(***************************************************************************)
(* The AuthenticatorSetsMatch predicate defines the conditions under which *)
(* a set of authenticators in the `^\basicname^' spec semantically matches *)
(* a set of authenticators in the `^\fastname^' spec.                      *)
(***************************************************************************)

AuthenticatorSetsMatch(ll1Authenticators, ll2Authenticators, symmetricKey, hashBarrier) == \*`^\label{DEF-AuthenticatorSetsMatch}^'
        (*******************************************************************)
        (* For every authenticator in the `^\basicname^' set, there is     *)
        (* some authenticator in the `^\fastname^' set that matches it.    *)
        (*******************************************************************)
    /\  \A ll1Authenticator \in ll1Authenticators :
            \E ll2Authenticator \in ll2Authenticators :
                AuthenticatorsMatch(
                    ll1Authenticator, ll2Authenticator, symmetricKey, hashBarrier)
        (*******************************************************************)
        (* For every authenticator in the `^\fastname^' set, there is some *)
        (* authenticator in the `^\basicname^' set that matches it.        *)
        (*******************************************************************)
    /\  \A ll2Authenticator \in ll2Authenticators :
            \E ll1Authenticator \in ll1Authenticators :
                AuthenticatorsMatch(
                    ll1Authenticator, ll2Authenticator, symmetricKey, hashBarrier)

----

(***************************************************************************)
(* The LL2NVRAMLogicalHistorySummary is the history summary that is        *)
(* represented by the state of the NVRAM and the SPCR in the `^\fastname^' *)
(* spec.  The anchor always comes directly from the NVRAM, but the         *)
(* extension only comes from the SPCR if the NVRAM indicates that an       *)
(* extension is in progress; otherwise, the extension is set to the base   *)
(* hash value.  The reason for this is that a LL2TakeCheckpoint action     *)
(* clears the extensionInProgress flag but is unable to reset the SPCR, so *)
(* during the time between a LL2TakeCheckpoint action and a LL2Restart     *)
(* action, the extension is really the base hash value, even though the    *)
(* SPCR has not yet been reset.                                            *)
(*                                                                         *)
(* The other interesting aspect of LL2NVRAMLogicalHistorySummary is that   *)
(* if there is an extension in progress but the SPCR equals the base hash  *)
(* value, then the logical value of the extension is set to a crazy hash   *)
(* value.  This is necessary because if a Restart occurs when the          *)
(* LL2NVRAM.historySummaryAnchor equals the base hash value, the SPCR will *)
(* be set to the BaseHashValue, but we don't want the logical history      *)
(* summary to appear the same as the initial history summary, which it     *)
(* would if both the anchor and the extension were to equal the base hash  *)
(* value.                                                                  *)
(***************************************************************************)

CONSTANT CrazyHashValue \*`^\label{CONSTANT-CrazyHashValue}^'

ASSUME CrazyHashValueTypeSafe == CrazyHashValue \in HashType \*`^\label{ASSUME-CrazyHashValueTypeSafe}^'

ASSUME CrazyHashValueUnique == \*`^\label{ASSUME-CrazyHashValueUnique}^'
    /\  \A hashInput1, hashInput2 \in HashDomain :
            Hash(hashInput1, hashInput2) # CrazyHashValue
    /\  BaseHashValue # CrazyHashValue

LL2NVRAMLogicalHistorySummary == \*`^\label{DEF-LL2NVRAMLogicalHistorySummary}^'
    IF LL2NVRAM.extensionInProgress
    THEN
        IF LL2SPCR = BaseHashValue
        THEN
          [ anchor |-> LL2NVRAM.historySummaryAnchor,
            extension |-> CrazyHashValue ]
        ELSE
          [ anchor |-> LL2NVRAM.historySummaryAnchor,
            extension |-> LL2SPCR ]
    ELSE
      [ anchor |-> LL2NVRAM.historySummaryAnchor,
        extension |-> BaseHashValue ]

----

(***************************************************************************)
(* The LL2Refinement describes the relationship between the 2nd low-level  *)
(* spec (`^\fastname^') and the 1st low-level spec (`^\basicname^').       *)
(*                                                                         *)
(* The following variables are directly equal between the two specs:       *)
(*     LLxAvailableInputs                                                  *)
(*     LLxObservedOutputs                                                  *)
(*     LLxDisk.publicState                                                 *)
(*     LLxDisk.privateStateEnc                                             *)
(*     LLxRAM.publicState                                                  *)
(*     LLxRAM.privateStateEnc                                              *)
(*     LLxNVRAM.symmetricKey                                               *)
(*                                                                         *)
(* The following variables directly match according to the operators       *)
(* defined above:                                                          *)
(*     LLxObservedAuthenticators                                           *)
(*     LLxDisk.historySummary                                              *)
(*     LLxDisk.authenticator                                               *)
(*     LLxRAM.historySummary                                               *)
(*     LLxRAM.authenticator                                                *)
(* Note that the authenticators in the sets of observed authenticators     *)
(* match using the symmetric key in the LL2NVRAM.  By contrast, the        *)
(* authenticators in the disk and RAM match using some unspecified         *)
(* symmetric key, because the values in the disk and RAM may be set        *)
(* arbitrarily by the user.                                                *)
(*                                                                         *)
(* The following variable has a more involved matching process, involving  *)
(* both the LL2NVRAMLogicalHistorySummary operator and a match operator:   *)
(*     LL1NVRAM.historySummary                                             *)
(*                                                                         *)
(* For each of the variables that are refined via match predicates rather  *)
(* than through an equality relation, the refinement asserts that the      *)
(* variable has the appropriate type.                                      *)
(*                                                                         *)
(* For each of the record types (LL1Disk, LL1RAM, and LL1NVRAM), the       *)
(* refinement defines the mapping for each individual field within the     *)
(* record.                                                                 *)
(***************************************************************************)

LL2Refinement == \*`^\label{DEF-LL2Refinement}^'
    /\  LL1AvailableInputs = LL2AvailableInputs
    /\  LL1ObservedOutputs = LL2ObservedOutputs
    /\  LL1ObservedAuthenticators \subseteq MACType
    /\  AuthenticatorSetsMatch(
            LL1ObservedAuthenticators,
            LL2ObservedAuthenticators,
            LL2NVRAM.symmetricKey,
            LL2NVRAM.hashBarrier)
    /\  LL1Disk \in LL1UntrustedStorageType
    /\  LL1Disk.publicState = LL2Disk.publicState
    /\  LL1Disk.privateStateEnc = LL2Disk.privateStateEnc
    /\  HistorySummariesMatch(
            LL1Disk.historySummary,
            LL2Disk.historySummary,
            LL2NVRAM.hashBarrier)
    /\  \E symmetricKey \in SymmetricKeyType :
            AuthenticatorsMatch(
                LL1Disk.authenticator,
                LL2Disk.authenticator,
                symmetricKey,
                LL2NVRAM.hashBarrier)
    /\  LL1RAM \in LL1UntrustedStorageType
    /\  LL1RAM.publicState = LL2RAM.publicState
    /\  LL1RAM.privateStateEnc = LL2RAM.privateStateEnc
    /\  HistorySummariesMatch(
            LL1RAM.historySummary,
            LL2RAM.historySummary,
            LL2NVRAM.hashBarrier)
    /\  \E symmetricKey \in SymmetricKeyType :
            AuthenticatorsMatch(
                LL1RAM.authenticator,
                LL2RAM.authenticator,
                symmetricKey,
                LL2NVRAM.hashBarrier)
    /\  LL1NVRAM \in LL1TrustedStorageType
    /\  HistorySummariesMatch(
            LL1NVRAM.historySummary,
            LL2NVRAMLogicalHistorySummary,
            LL2NVRAM.hashBarrier)
    /\  LL1NVRAM.symmetricKey = LL2NVRAM.symmetricKey

====
