---- MODULE MemoirLL2Specification ----

(***************************************************************************)
(* This module defines the specification of `^\fastname^'.                 *)
(*                                                                         *)
(* There are nine actions:                                                 *)
(*     LL2MakeInputAvailable                                               *)
(*     LL2PerformOperation                                                 *)
(*     LL2RepeatOperation                                                  *)
(*     LL2TakeCheckpoint                                                   *)
(*     LL2Restart                                                          *)
(*     LL2ReadDisk                                                         *)
(*     LL2WriteDisk                                                        *)
(*     LL2CorruptRAM                                                       *)
(*     LL2CorruptSPCR                                                      *)
(***************************************************************************)

EXTENDS MemoirLL1Implementation

(***************************************************************************)
(* In `^\fastname^', each history summary (which is a composite chained    *)
(* hash of all inputs that the service has processed) is partitioned into  *)
(* two pieces: an anchor and an extension.                                 *)
(***************************************************************************)

HistorySummaryType == [ \*`^\label{DEF-HistorySummaryType}^'
    anchor : HashType,
    extension : HashType ]

(***************************************************************************)
(* The disk and RAM are untrusted.  They each can store the service's      *)
(* public state, encrypted private state, a history summary, and an        *)
(* authenticator (a MAC that binds the history summary to the public and   *)
(* private state).                                                         *)
(***************************************************************************)

LL2UntrustedStorageType == [ \*`^\label{DEF-LL2UntrustedStorageType}^'
    publicState : PublicStateType,
    privateStateEnc : PrivateStateEncType,
    historySummary : HistorySummaryType,
    authenticator : MACType ]

(***************************************************************************)
(* The NVRAM is trusted, since the TPM guarantees that it can only be read *)
(* or written by the code that implements `^\name^'.  The NVRAM stores the *)
(* current history summary anchor and the symmetric key that is used to    *)
(* encrypt the private state and to MAC the history summary and service    *)
(* state into an authenticator.  It also stores a hash barrier for         *)
(* securing the history summary and a guard bit that indicates whether the *)
(* current history summary has been extended, such that the history        *)
(* summary anchor is not a complete representation of the inputs summary.  *)
(***************************************************************************)

LL2TrustedStorageType == [ \*`^\label{DEF-LL2TrustedStorageType}^'
    historySummaryAnchor : HashType,
    symmetricKey : SymmetricKeyType,
    hashBarrier : HashType,
    extensionInProgress : BOOLEAN ]

(***************************************************************************)
(* LL2AvailableInputs and LL2ObservedOutputs are abstract variables that   *)
(* do not directly represent part of the implementation.  They correspond  *)
(* to the HLAvailableInputs and HLObservedOutputs variables in the         *)
(* `^\fastname^' spec.                                                     *)
(***************************************************************************)

VARIABLE LL2AvailableInputs \*`^\label{VARIABLE-LL2AvailableInputs}^'
VARIABLE LL2ObservedOutputs \*`^\label{VARIABLE-LL2ObservedOutputs}^'

(***************************************************************************)
(* The ObservedAuthenticators variable is also abstract.  It records the   *)
(* set of authenticators that the user has seen from `^\name^'.  A         *)
(* malicious user can attempt to use these authenticators in a replay      *)
(* attack against `^\name^'.                                               *)
(***************************************************************************)

VARIABLE LL2ObservedAuthenticators \*`^\label{VARIABLE-LL2ObservedAuthenticators}^'

(***************************************************************************)
(* The LL2Disk, LL2RAM, LL2NVRAM, and LL2SPCR variables represent concrete *)
(* state maintained by the `^\fastname^' implementation.                   *)
(*                                                                         *)
(* The LL2SPCR is semi-trusted.  Any party can write to it, but arbitrary  *)
(* writes are not allowed.  The only allowable updates are of the form     *)
(*  \E x : LL2SPCR' = Hash(LL2SPCR, x)                                     *)
(* We use the LL2SPCR to store a history summary extension.                *)
(***************************************************************************)

VARIABLE LL2Disk \*`^\label{VARIABLE-LL2Disk}^'
VARIABLE LL2RAM \*`^\label{VARIABLE-LL2RAM}^'
VARIABLE LL2NVRAM \*`^\label{VARIABLE-LL2NVRAM}^'
VARIABLE LL2SPCR \*`^\label{VARIABLE-LL2SPCR}^'

LL2TypeInvariant == \*`^\label{DEF-LL2TypeInvariant}^'
    /\  LL2AvailableInputs \subseteq InputType
    /\  LL2ObservedOutputs \subseteq OutputType
    /\  LL2ObservedAuthenticators \subseteq MACType
    /\  LL2Disk \in LL2UntrustedStorageType
    /\  LL2RAM \in LL2UntrustedStorageType
    /\  LL2NVRAM \in LL2TrustedStorageType
    /\  LL2SPCR \in HashType

LL2Vars == << \*`^\label{DEF-LL2Vars}^'
    LL2AvailableInputs,
    LL2ObservedOutputs,
    LL2ObservedAuthenticators,
    LL2Disk,
    LL2RAM,
    LL2NVRAM,
    LL2SPCR >>

----

(***************************************************************************)
(* The Checkpoint function takes a history summary that may or may not be  *)
(* checkpointed and produces a checkpointed history summary from it.       *)
(***************************************************************************)

Checkpoint(historySummary) == \*`^\label{DEF-Checkpoint}^'
    LET
        checkpointedAnchor == Hash(historySummary.anchor, historySummary.extension)
        checkpointedHistorySummary == [
            anchor |-> checkpointedAnchor,
            extension |-> BaseHashValue ]
    IN
        IF historySummary.extension = BaseHashValue
        THEN
            historySummary
        ELSE
            checkpointedHistorySummary

----

(***************************************************************************)
(* The Successor function defines the history summary that results from    *)
(* extending a given history summary with a given input.  It secures the   *)
(* input using a hash barrier to thwart forgery.                           *)
(***************************************************************************)

Successor(historySummary, input, hashBarrier) == \*`^\label{DEF-Successor}^'
    LET
        securedInput == Hash(hashBarrier, input)
        newAnchor == historySummary.anchor
        newExtension == Hash(historySummary.extension, securedInput)
        newHistorySummary == [
            anchor |-> newAnchor,
            extension |-> newExtension ]
    IN
        newHistorySummary

----

(***************************************************************************)
(* The LL2MakeInputAvailable action is a direct analog of the              *)
(* HLMakeInputAvailable action in the high-level spec.                     *)
(***************************************************************************)

LL2MakeInputAvailable == \*`^\label{DEF-LL2MakeInputAvailable}^'
    \E input \in InputType :
        /\  input \notin LL2AvailableInputs
        /\  LL2AvailableInputs' = LL2AvailableInputs \union {input}
        /\  UNCHANGED LL2Disk
        /\  UNCHANGED LL2RAM
        /\  UNCHANGED LL2NVRAM
        /\  UNCHANGED LL2SPCR
        /\  UNCHANGED LL2ObservedOutputs
        /\  UNCHANGED LL2ObservedAuthenticators

----

(***************************************************************************)
(* The LL2PerformOperation action is invoked by the user to perform a      *)
(* service operation.  It is intended to provide the semantics of the      *)
(* HLAdvanceService action in the high-level spec.                         *)
(***************************************************************************)

LL2PerformOperation == \*`^\label{DEF-LL2PerformOperation}^'
    \E input \in LL2AvailableInputs :
        LET
            historySummaryHash ==
                Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
            stateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
            historyStateBinding == Hash(historySummaryHash, stateHash)
            privateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
            sResult == Service(LL2RAM.publicState, privateState, input)
            newPrivateStateEnc ==
                SymmetricEncrypt(LL2NVRAM.symmetricKey, sResult.newPrivateState)
            currentHistorySummary == [
                anchor |-> LL2NVRAM.historySummaryAnchor,
                extension |-> LL2SPCR ]
            newHistorySummary == Successor(currentHistorySummary, input, LL2NVRAM.hashBarrier)
            newHistorySummaryHash == Hash(newHistorySummary.anchor, newHistorySummary.extension)
            newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
            newHistoryStateBinding == Hash(newHistorySummaryHash, newStateHash)
            newAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, newHistoryStateBinding)
        IN
                (***********************************************************)
                (* There are three enablement conditions:                  *)
                (*                                                         *)
                (* First, the authenticator supplied by the user           *)
                (* (LL2RAM.authenticator) must validly bind the            *)
                (* user-supplied public and encrypted private state to the *)
                (* user-supplied history summary.                          *)
                (*                                                         *)
                (* Second, the value in the SPCR must be consistent with   *)
                (* the flag that indicates whether an extension is in      *)
                (* progress.                                               *)
                (*                                                         *)
                (* Third, the user-supplied history summary                *)
                (* (LL2RAM.historySummary) must match the history summary  *)
                (* in the TPM (NVRAM and SPCR).  There are two different   *)
                (* ways to check this condition, depending on whether a    *)
                (* extension is in progress.                               *)
                (***********************************************************)
            /\  ValidateMAC(LL2NVRAM.symmetricKey, historyStateBinding, LL2RAM.authenticator)
            /\  IF LL2NVRAM.extensionInProgress = TRUE
                THEN
                    /\  LL2SPCR # BaseHashValue
                    /\  currentHistorySummary = LL2RAM.historySummary
                ELSE
                    /\  LL2SPCR = BaseHashValue
                    /\  currentHistorySummary = Checkpoint(LL2RAM.historySummary)
                (***********************************************************)
                (* At the conclusion of the action, the RAM contains the   *)
                (* new public and encrypted private state, the new history *)
                (* summary, and an authenticator that binds these          *)
                (* together.                                               *)
                (***********************************************************)
            /\  LL2RAM' = [
                    publicState |-> sResult.newPublicState,
                    privateStateEnc |-> newPrivateStateEnc,
                    historySummary |-> newHistorySummary,
                    authenticator |-> newAuthenticator ]
                (***********************************************************)
                (* The NVRAM is updated to indicate that the extension is  *)
                (* in progress.  The NVRAM may already indicate this, in   *)
                (* which case this step does not require writing to the    *)
                (* NVRAM.                                                  *)
                (***********************************************************)
            /\  LL2NVRAM' = [
                    historySummaryAnchor |-> LL2NVRAM.historySummaryAnchor,
                    symmetricKey |-> LL2NVRAM.symmetricKey,
                    hashBarrier |-> LL2NVRAM.hashBarrier,
                    extensionInProgress |-> TRUE ]
                (***********************************************************)
                (* The SPCR is updated with the new history summary        *)
                (* extension.                                              *)
                (***********************************************************)
            /\  LL2SPCR' = newHistorySummary.extension
                (***********************************************************)
                (* The output of the service is added to the set of        *)
                (* outputs that the user has observed.                     *)
                (***********************************************************)
            /\  LL2ObservedOutputs' = LL2ObservedOutputs \union {sResult.output}
                (***********************************************************)
                (* The disk is unchanged.                                  *)
                (***********************************************************)
            /\  UNCHANGED LL2Disk
                (***********************************************************)
                (* The set of available inputs is unchanged.               *)
                (***********************************************************)
            /\  UNCHANGED LL2AvailableInputs
                (***********************************************************)
                (* The new authenticator is added to the set of            *)
                (* authenticators that the user has observed.              *)
                (***********************************************************)
            /\  LL2ObservedAuthenticators' =
                    LL2ObservedAuthenticators \union {newAuthenticator}

----

(***************************************************************************)
(* The LL2RepeatOperation action is invoked by the user when the computer  *)
(* crashed after a LL2PerformOperation action was performed but before the *)
(* user had a chance to perform a LL2WriteDisk action to persistently      *)
(* record the new state and its authenticator.  This action enables the    *)
(* user to reproduce the result of the most-recent LL2PerformOperation     *)
(* action.                                                                 *)
(***************************************************************************)

LL2RepeatOperation == \*`^\label{DEF-LL2RepeatOperation}^'
    \E input \in LL2AvailableInputs :
        LET
            historySummaryHash ==
                Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
            stateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
            historyStateBinding == Hash(historySummaryHash, stateHash)
            newHistorySummary == Successor(LL2RAM.historySummary, input, LL2NVRAM.hashBarrier)
            checkpointedHistorySummary == Checkpoint(LL2RAM.historySummary)
            newCheckpointedHistorySummary ==
                Successor(checkpointedHistorySummary, input, LL2NVRAM.hashBarrier)
            checkpointedNewHistorySummary == Checkpoint(newHistorySummary)
            checkpointedNewCheckpointedHistorySummary ==
                Checkpoint(newCheckpointedHistorySummary)
            privateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
            sResult == Service(LL2RAM.publicState, privateState, input)
            newPrivateStateEnc ==
                SymmetricEncrypt(LL2NVRAM.symmetricKey, sResult.newPrivateState)
            currentHistorySummary == [
                anchor |-> LL2NVRAM.historySummaryAnchor,
                extension |-> LL2SPCR ]
            currentHistorySummaryHash == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
            newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
            newHistoryStateBinding == Hash(currentHistorySummaryHash, newStateHash)
            newAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, newHistoryStateBinding)
        IN
                (***********************************************************)
                (* There are three enablement conditions:                  *)
                (*                                                         *)
                (* First, the authenticator supplied by the user           *)
                (* (LL2RAM.authenticator) must validly bind the            *)
                (* user-supplied public and encrypted private state to the *)
                (* history summary supplied by the user.                   *)
                (*                                                         *)
                (* Second, a TakeCheckpoint action should always occur     *)
                (* immediately before a shutdown, power-off, or reboot;    *)
                (* therefore, an extension will not be in progress at the  *)
                (* time LL2RepeatOperation is needed.  This implies that   *)
                (* the flag in the NVRAM must indicate that an extension   *)
                (* is not in progress, and the value in the SPCR must      *)
                (* equal the BaseHashValue.                                *)
                (*                                                         *)
                (* Third, the user-supplied history summary                *)
                (* (LL2RAM.historySummary), extended with the              *)
                (* user-supplied input, must match the history summary in  *)
                (* the TPM (NVRAM and SPCR).  This condition ensures that  *)
                (* this action will invoke the service with the same input *)
                (* used with the most recent LL2PerformOperation action.   *)
                (* There are two different ways to check this condition,   *)
                (* depending on whether a checkpoint was taken before the  *)
                (* input was processed.                                    *)
                (***********************************************************)
            /\  ValidateMAC(LL2NVRAM.symmetricKey, historyStateBinding, LL2RAM.authenticator)
            /\  LL2NVRAM.extensionInProgress = FALSE
            /\  LL2SPCR = BaseHashValue
            /\      (*******************************************************)
                    (* no checkpoint before input                          *)
                    (*******************************************************)
                \/  currentHistorySummary = checkpointedNewHistorySummary
                    (*******************************************************)
                    (* checkpoints before input                            *)
                    (*******************************************************)
                \/  currentHistorySummary = checkpointedNewCheckpointedHistorySummary
                (***********************************************************)
                (* At the conclusion of the action, the RAM contains the   *)
                (* new public and encrypted private state, the new history *)
                (* summary, and an authenticator that binds these          *)
                (* together.  These should match the values produced by    *)
                (* the most recent LL2PerformOperation action.             *)
                (***********************************************************)
            /\  LL2RAM' = [
                    publicState |-> sResult.newPublicState,
                    privateStateEnc |-> newPrivateStateEnc,
                    historySummary |-> currentHistorySummary,
                    authenticator |-> newAuthenticator ]
                (***********************************************************)
                (* The output of the service is added to the set of        *)
                (* outputs that the user has observed.  If `^\name^' is    *)
                (* working correctly, the user already saw this output     *)
                (* when the previous LL2PerformOperation action was        *)
                (* executed.                                               *)
                (***********************************************************)
            /\  LL2ObservedOutputs' = LL2ObservedOutputs \union {sResult.output}
                (***********************************************************)
                (* The NVRAM is unchanged, because this action is not      *)
                (* supposed to change the state of the service.            *)
                (***********************************************************)
            /\  UNCHANGED LL2NVRAM
                (***********************************************************)
                (* The SPCR is unchanged, because this action is not       *)
                (* supposed to change the state of the service.            *)
                (***********************************************************)
            /\  UNCHANGED LL2SPCR
                (***********************************************************)
                (* The disk is unchanged.                                  *)
                (***********************************************************)
            /\  UNCHANGED LL2Disk
                (***********************************************************)
                (* The set of available inputs is unchanged.               *)
                (***********************************************************)
            /\  UNCHANGED LL2AvailableInputs
                (***********************************************************)
                (* The new authenticator is added to the set of            *)
                (* authenticators that the user has observed.  If          *)
                (* `^\name^' is working correctly, the user already saw    *)
                (* this authenticator when the previous                    *)
                (* LL2PerformOperation action was executed.                *)
                (***********************************************************)
            /\  LL2ObservedAuthenticators' =
                    LL2ObservedAuthenticators \union {newAuthenticator}

----

(***************************************************************************)
(* The LL2TakeCheckpoint action occurs in response to an NMI indicating    *)
(* that a shutdown, power-off, or reboot is imminent.                      *)
(***************************************************************************)

LL2TakeCheckpoint == \*`^\label{DEF-LL2TakeCheckpoint}^'
    LET
        newHistorySummaryAnchor == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
    IN
            (***************************************************************)
            (* There are two enablement conditions.  The guard bit in the  *)
            (* NVRAM must indicate that an extension is in progress, and   *)
            (* the SPCR must contain an extension.                         *)
            (***************************************************************)
        /\  LL2NVRAM.extensionInProgress = TRUE
        /\  LL2SPCR # BaseHashValue
            (***************************************************************)
            (* This action changes nothing other than the NVRAM.           *)
            (***************************************************************)
        /\  UNCHANGED LL2RAM
        /\  UNCHANGED LL2Disk
        /\  LL2NVRAM' = [
                historySummaryAnchor |-> newHistorySummaryAnchor,
                symmetricKey |-> LL2NVRAM.symmetricKey,
                hashBarrier |-> LL2NVRAM.hashBarrier,
                extensionInProgress |-> FALSE ]
        /\  UNCHANGED LL2SPCR
        /\  UNCHANGED LL2AvailableInputs
        /\  UNCHANGED LL2ObservedOutputs
        /\  UNCHANGED LL2ObservedAuthenticators

----

(***************************************************************************)
(* The LL2Restart action occurs when the computer restarts.                *)
(***************************************************************************)

LL2Restart == \*`^\label{DEF-LL2Restart}^'
    \E  untrustedStorage \in LL2UntrustedStorageType,
        randomSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey},
        hash \in HashType :
            (***************************************************************)
            (* The state of the RAM is garbaged by a restart, so we set it *)
            (* to some *almost* arbitrary value in                         *)
            (* LL2UntrustedStorageType.  The only condition we impose is   *)
            (* that the authenticator is not coincidentally equal to an    *)
            (* authenticator that could be computed with the symmetric key *)
            (* known only to `^\name^'.                                    *)
            (***************************************************************)
        /\  untrustedStorage.authenticator = GenerateMAC(randomSymmetricKey, hash)
        /\  LL2RAM' = untrustedStorage
        /\  UNCHANGED LL2Disk
        /\  UNCHANGED LL2NVRAM
            (***************************************************************)
            (* The value of the SPCR is set to a known starting value,     *)
            (* which we model with the BaseHashValue.                      *)
            (***************************************************************)
        /\  LL2SPCR' = BaseHashValue
        /\  UNCHANGED LL2AvailableInputs
        /\  UNCHANGED LL2ObservedOutputs
        /\  UNCHANGED LL2ObservedAuthenticators

----

(***************************************************************************)
(* The LL2ReadDisk action copies the state of the disk into the RAM.       *)
(***************************************************************************)

LL2ReadDisk == \*`^\label{DEF-LL2ReadDisk}^'
    /\  LL2RAM' = LL2Disk
    /\  UNCHANGED LL2Disk
    /\  UNCHANGED LL2NVRAM
    /\  UNCHANGED LL2SPCR
    /\  UNCHANGED LL2AvailableInputs
    /\  UNCHANGED LL2ObservedOutputs
    /\  UNCHANGED LL2ObservedAuthenticators

----

(***************************************************************************)
(* The LL2WriteDisk action copies the state of the RAM onto the disk.      *)
(***************************************************************************)

LL2WriteDisk == \*`^\label{DEF-LL2WriteDisk}^'
    /\  LL2Disk' = LL2RAM
    /\  UNCHANGED LL2RAM
    /\  UNCHANGED LL2NVRAM
    /\  UNCHANGED LL2SPCR
    /\  UNCHANGED LL2AvailableInputs
    /\  UNCHANGED LL2ObservedOutputs
    /\  UNCHANGED LL2ObservedAuthenticators

----

(***************************************************************************)
(* The LL2CorruptRAM action models the ability of a malicious user to      *)
(* attack `^\name^' by supplying *almost* arbitrary data to `^\name^'.     *)
(* The data is not completely arbitrary, because the user is assumed to be *)
(* unable to forge authenticators using the symmetric key stored in the    *)
(* NVRAM of the TPM.                                                       *)
(***************************************************************************)

LL2CorruptRAM == \*`^\label{DEF-LL2CorruptRAM}^'
    \E  untrustedStorage \in LL2UntrustedStorageType,
        fakeSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey},
        hash \in HashType :
                (***********************************************************)
                (* The user can launch a replay attack by re-using any     *)
                (* authenticator previously observed.                      *)
                (***********************************************************)
        /\  \/  untrustedStorage.authenticator \in LL2ObservedAuthenticators
                (***********************************************************)
                (* Or the user can create a fake authenticator using some  *)
                (* other symmetric key.                                    *)
                (***********************************************************)
            \/  untrustedStorage.authenticator = GenerateMAC(fakeSymmetricKey, hash)
        /\  LL2RAM' = untrustedStorage
        /\  UNCHANGED LL2Disk
        /\  UNCHANGED LL2NVRAM
        /\  UNCHANGED LL2SPCR
        /\  UNCHANGED LL2AvailableInputs
        /\  UNCHANGED LL2ObservedOutputs
        /\  UNCHANGED LL2ObservedAuthenticators

----

(***************************************************************************)
(* The LL2CorruptSPCR action models the ability of a malicious user to     *)
(* attack `^\name^' by extending the SPCR with *almost* arbitrary data.    *)
(* The data is not completely arbitrary, because the user is assumed not   *)
(* to know the hash barrier stored in the NVRAM of the TPM, so the user    *)
(* has negligible probability of being able to correctly guess this value  *)
(* and use it to extend the SPCR.                                          *)
(***************************************************************************)

LL2CorruptSPCR == \*`^\label{DEF-LL2CorruptSPCR}^'
    \E fakeHash \in HashDomain :
        LET
            (***************************************************************)
            (* The SPCR can only be modified by extending it.              *)
            (***************************************************************)
            newHistorySummaryExtension == Hash(LL2SPCR, fakeHash)
        IN
            /\  \A fakeInput \in InputType : fakeHash # Hash(LL2NVRAM.hashBarrier, fakeInput)
            /\  UNCHANGED LL2RAM
            /\  UNCHANGED LL2Disk
            /\  UNCHANGED LL2NVRAM
            /\  LL2SPCR' = newHistorySummaryExtension
            /\  UNCHANGED LL2AvailableInputs
            /\  UNCHANGED LL2ObservedOutputs
            /\  UNCHANGED LL2ObservedAuthenticators

----

(***************************************************************************)
(* In the initial state of the `^\fastname^' spec, some symmetric key and  *)
(* some hash barrier are generated and stored in the NVRAM.  The initial   *)
(* history summary anchor and extension are both the base hash value,      *)
(* indicating that no inputs have been supplied yet.  An initial           *)
(* authenticator binds the initial history summary to the initial public   *)
(* and encrypted private state.                                            *)
(***************************************************************************)

LL2Init == \*`^\label{DEF-LL2Init}^'
    \E symmetricKey \in SymmetricKeyType, hashBarrier \in HashType :
        LET
            initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
            initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
            initialHistorySummary == [
                anchor |-> BaseHashValue,
                extension |-> BaseHashValue ]
            initialHistorySummaryHash == Hash(BaseHashValue, BaseHashValue)
            initialHistoryStateBinding == Hash(initialHistorySummaryHash, initialStateHash)
            initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
            initialUntrustedStorage == [
                publicState |-> InitialPublicState,
                privateStateEnc |-> initialPrivateStateEnc,
                historySummary |-> initialHistorySummary,
                authenticator |-> initialAuthenticator ]
            initialTrustedStorage == [
                historySummaryAnchor |-> BaseHashValue,
                symmetricKey |-> symmetricKey,
                hashBarrier |-> hashBarrier,
                extensionInProgress |-> FALSE ]
        IN
            /\  LL2Disk = initialUntrustedStorage
            /\  LL2RAM = initialUntrustedStorage
            /\  LL2NVRAM = initialTrustedStorage
            /\  LL2SPCR = BaseHashValue
            /\  LL2AvailableInputs = InitialAvailableInputs
            /\  LL2ObservedOutputs = {}
            /\  LL2ObservedAuthenticators = {initialAuthenticator}

LL2Next == \*`^\label{DEF-LL2Next}^'
    \/  LL2MakeInputAvailable
    \/  LL2PerformOperation
    \/  LL2RepeatOperation
    \/  LL2TakeCheckpoint
    \/  LL2Restart
    \/  LL2ReadDisk
    \/  LL2WriteDisk
    \/  LL2CorruptRAM
    \/  LL2CorruptSPCR

LL2Spec == LL2Init /\ [][LL2Next]_LL2Vars \*`^\label{DEF-LL2Spec}^'

====
