---- MODULE MemoirLL2TypeLemmas ----

(***************************************************************************)
(* This module states and proves sevaral lemmas that are useful for        *)
(* proving type safety.  Since type safety is an important part of the     *)
(* implementation proof, these lemmas also will be used in theorems other  *)
(* than the `^\fastname^' type-safety theorem.                             *)
(*                                                                         *)
(* The lemmas in this module are:                                          *)
(*     CheckpointDefsTypeSafeLemma                                         *)
(*     CheckpointTypeSafe                                                  *)
(*     SuccessorDefsTypeSafeLemma                                          *)
(*     SuccessorTypeSafe                                                   *)
(*     LL2SubtypeImplicationLemma                                          *)
(*     LL2InitDefsTypeSafeLemma                                            *)
(*     LL2PerformOperationDefsTypeSafeLemma                                *)
(*     LL2RepeatOperationDefsTypeSafeLemma                                 *)
(*     LL2TakeCheckpointDefsTypeSafeLemma                                  *)
(*     LL2CorruptSPCRDefsTypeSafeLemma                                     *)
(*     AuthenticatorsMatchDefsTypeSafeLemma                                *)
(***************************************************************************)

EXTENDS MemoirLL2Refinement

(***************************************************************************)
(* The CheckpointDefsTypeSafeLemma proves that the definitions within the  *)
(* LET of the Checkpoint function all have the appropriate type.  This is  *)
(* a trivial proof that merely walks through the definitions.              *)
(***************************************************************************)

THEOREM CheckpointDefsTypeSafeLemma == \*`^\label{THEOREM-CheckpointDefsTypeSafeLemma}^'
    \A historySummary \in HistorySummaryType :
        LET
            checkpointedAnchor == Hash(historySummary.anchor, historySummary.extension)
            checkpointedHistorySummary == [
                anchor |-> checkpointedAnchor,
                extension |-> BaseHashValue ]
        IN
            /\  checkpointedAnchor \in HashType
            /\  checkpointedHistorySummary \in HistorySummaryType
            /\  checkpointedHistorySummary.anchor \in HashType
            /\  checkpointedHistorySummary.extension \in HashType
  <1>1. TAKE historySummary \in HistorySummaryType
  <1> checkpointedAnchor == Hash(historySummary.anchor, historySummary.extension)
  <1> checkpointedHistorySummary == [
          anchor |-> checkpointedAnchor,
          extension |-> BaseHashValue ]
  <1> HIDE DEF checkpointedAnchor, checkpointedHistorySummary
  <1>2. checkpointedAnchor \in HashType
    <2>1. historySummary.anchor \in HashDomain
      <3>1. historySummary.anchor \in HashType
        BY <1>1 DEF HistorySummaryType
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. historySummary.extension \in HashDomain
      <3>1. historySummary.extension \in HashType
        BY <1>1 DEF HistorySummaryType
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF checkpointedAnchor
  <1>3. checkpointedHistorySummary \in HistorySummaryType
    <2>1. BaseHashValue \in HashType
      BY BaseHashValueTypeSafe
    <2>2. QED
      BY <1>2, <2>1 DEF checkpointedHistorySummary, HistorySummaryType
  <1>4. /\  checkpointedHistorySummary.anchor \in HashType
        /\  checkpointedHistorySummary.extension \in HashType
    BY <1>3 DEF HistorySummaryType
  <1>5. QED
    BY <1>2, <1>3, <1>4 DEF checkpointedAnchor, checkpointedHistorySummary

(***************************************************************************)
(* Type safety of the Checkpoint function.                                 *)
(***************************************************************************)

THEOREM CheckpointTypeSafe == \*`^\label{THEOREM-CheckpointTypeSafe}^'
    \A historySummary \in HistorySummaryType : Checkpoint(historySummary) \in HistorySummaryType
  <1>1. TAKE historySummary \in HistorySummaryType
  <1> checkpointedAnchor == Hash(historySummary.anchor, historySummary.extension)
  <1> checkpointedHistorySummary == [
          anchor |-> checkpointedAnchor,
          extension |-> BaseHashValue ]
  <1>2. /\  checkpointedAnchor \in HashType
        /\  checkpointedHistorySummary \in HistorySummaryType
        /\  checkpointedHistorySummary.anchor \in HashType
        /\  checkpointedHistorySummary.extension \in HashType
    BY <1>1, CheckpointDefsTypeSafeLemma
  <1> HIDE DEF checkpointedAnchor, checkpointedHistorySummary
  <1>3. \/  Checkpoint(historySummary) = historySummary
        \/  Checkpoint(historySummary) = checkpointedHistorySummary
    BY <1>1 DEF Checkpoint, checkpointedHistorySummary, checkpointedAnchor
  <1>4. historySummary \in HistorySummaryType
    BY <1>1
  <1>5. checkpointedHistorySummary \in HistorySummaryType
    BY <1>2
  <1>6. QED
    BY <1>3, <1>4, <1>5

(***************************************************************************)
(* The SuccessorDefsTypeSafeLemma proves that the definitions within the   *)
(* LET of the Successor function all have the appropriate type.  This is a *)
(* trivial proof that merely walks through the definitions.                *)
(***************************************************************************)

THEOREM SuccessorDefsTypeSafeLemma == \*`^\label{THEOREM-SuccessorDefsTypeSafeLemma}^'
    \A historySummary \in HistorySummaryType, input \in InputType, hashBarrier \in HashType :
        LET
            securedInput == Hash(hashBarrier, input)
            newAnchor == historySummary.anchor
            newExtension == Hash(historySummary.extension, securedInput)
            newHistorySummary == [
                anchor |-> newAnchor,
                extension |-> newExtension ]
        IN
            /\  securedInput \in HashType
            /\  newAnchor \in HashType
            /\  newExtension \in HashType
            /\  newHistorySummary \in HistorySummaryType
            /\  newHistorySummary.anchor \in HashType
            /\  newHistorySummary.extension \in HashType
  <1>1. TAKE historySummary \in HistorySummaryType, input \in InputType, hashBarrier \in HashType
  <1> securedInput == Hash(hashBarrier, input)
  <1> newAnchor == historySummary.anchor
  <1> newExtension == Hash(historySummary.extension, securedInput)
  <1> newHistorySummary == [
          anchor |-> newAnchor,
          extension |-> newExtension ]
  <1> HIDE DEF securedInput, newAnchor, newExtension, newHistorySummary
  <1>2. securedInput \in HashType
    <2>1. hashBarrier \in HashDomain
      <3>1. hashBarrier \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. input \in HashDomain
      <3>1. input \in InputType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF securedInput
  <1>3. newAnchor \in HashType
    <2>1. historySummary \in HistorySummaryType
      BY <1>1
    <2>2. QED
      BY <2>1 DEF newAnchor, HistorySummaryType
  <1>4. newExtension \in HashType
    <2>1. historySummary.extension \in HashDomain
      <3>1. historySummary.extension \in HashType
        <4>1. historySummary \in HistorySummaryType
          BY <1>1
        <4>2. QED
          BY <4>1 DEF HistorySummaryType
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. securedInput \in HashDomain
      BY <1>2 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newExtension
  <1>5. newHistorySummary \in HistorySummaryType
    BY <1>3, <1>4 DEF newHistorySummary, HistorySummaryType
  <1>6. /\  newHistorySummary.anchor \in HashType
        /\  newHistorySummary.extension \in HashType
    BY <1>5 DEF HistorySummaryType
  <1>7. QED
    BY <1>2, <1>3, <1>4, <1>5, <1>6
    DEF securedInput, newAnchor, newExtension, newHistorySummary

(***************************************************************************)
(* Type safety of the Successor function.                                  *)
(***************************************************************************)

THEOREM SuccessorTypeSafe == \*`^\label{THEOREM-SuccessorTypeSafe}^'
    \A historySummary \in HistorySummaryType, input \in InputType, hashBarrier \in HashType :
        Successor(historySummary, input, hashBarrier) \in HistorySummaryType
  <1>1. TAKE historySummary \in HistorySummaryType, input \in InputType, hashBarrier \in HashType
  <1> securedInput == Hash(hashBarrier, input)
  <1> newAnchor == historySummary.anchor
  <1> newExtension == Hash(historySummary.extension, securedInput)
  <1> newHistorySummary == [
          anchor |-> newAnchor,
          extension |-> newExtension ]
  <1>2. /\  securedInput \in HashType
        /\  newAnchor \in HashType
        /\  newExtension \in HashType
        /\  newHistorySummary \in HistorySummaryType
        /\  newHistorySummary.anchor \in HashType
        /\  newHistorySummary.extension \in HashType
    BY <1>1, SuccessorDefsTypeSafeLemma
  <1> HIDE DEF securedInput, newAnchor, newExtension, newHistorySummary
  <1>3. Successor(historySummary, input, hashBarrier) = newHistorySummary
    BY <1>1 DEF Successor, newHistorySummary, newExtension, newAnchor, securedInput
  <1>4. newHistorySummary \in HistorySummaryType
    BY <1>2
  <1>5. QED
    BY <1>3, <1>4

(***************************************************************************)
(* The LL2SubtypeImplicationLemma proves that when the LL2TypeInvariant    *)
(* holds, the subtypes of LL2Disk, LL2RAM, and LL2NVRAM also hold.  This   *)
(* is asserted and proven for both the unprimed and primed states.         *)
(*                                                                         *)
(* The proof itself is completely trivial.  It follows directly from the   *)
(* type definitions LL2UntrustedStorageType and LL2TrustedStorageType.     *)
(***************************************************************************)

LL2SubtypeImplication == \*`^\label{DEF-LL2SubtypeImplication}^'
    LL2TypeInvariant =>
        /\  LL2Disk.publicState \in PublicStateType
        /\  LL2Disk.privateStateEnc \in PrivateStateEncType
        /\  LL2Disk.historySummary \in HistorySummaryType
        /\  LL2Disk.historySummary.anchor \in HashType
        /\  LL2Disk.historySummary.extension \in HashType
        /\  LL2Disk.authenticator \in MACType
        /\  LL2RAM.publicState \in PublicStateType
        /\  LL2RAM.privateStateEnc \in PrivateStateEncType
        /\  LL2RAM.historySummary \in HistorySummaryType
        /\  LL2RAM.historySummary.anchor \in HashType
        /\  LL2RAM.historySummary.extension \in HashType
        /\  LL2RAM.authenticator \in MACType
        /\  LL2NVRAM.historySummaryAnchor \in HashType
        /\  LL2NVRAM.symmetricKey \in SymmetricKeyType
        /\  LL2NVRAM.hashBarrier \in HashType
        /\  LL2NVRAM.extensionInProgress \in BOOLEAN

THEOREM LL2SubtypeImplicationLemma == \*`^\label{THEOREM-LL2SubtypeImplicationLemma}^'
    /\ LL2SubtypeImplication
    /\ LL2SubtypeImplication'
  <1>1. LL2SubtypeImplication
    <2>1. SUFFICES
          ASSUME LL2TypeInvariant
          PROVE
            /\  LL2Disk.publicState \in PublicStateType
            /\  LL2Disk.privateStateEnc \in PrivateStateEncType
            /\  LL2Disk.historySummary \in HistorySummaryType
            /\  LL2Disk.historySummary.anchor \in HashType
            /\  LL2Disk.historySummary.extension \in HashType
            /\  LL2Disk.authenticator \in MACType
            /\  LL2RAM.publicState \in PublicStateType
            /\  LL2RAM.privateStateEnc \in PrivateStateEncType
            /\  LL2RAM.historySummary \in HistorySummaryType
            /\  LL2RAM.historySummary.anchor \in HashType
            /\  LL2RAM.historySummary.extension \in HashType
            /\  LL2RAM.authenticator \in MACType
            /\  LL2NVRAM.historySummaryAnchor \in HashType
            /\  LL2NVRAM.symmetricKey \in SymmetricKeyType
            /\  LL2NVRAM.hashBarrier \in HashType
            /\  LL2NVRAM.extensionInProgress \in BOOLEAN
      BY DEF LL2SubtypeImplication, LL2TypeInvariant
    <2>2. LL2Disk \in LL2UntrustedStorageType
      BY <2>1 DEF LL2TypeInvariant
    <2>3. LL2RAM \in LL2UntrustedStorageType
      BY <2>1 DEF LL2TypeInvariant
    <2>4. LL2NVRAM \in LL2TrustedStorageType
      BY <2>1 DEF LL2TypeInvariant
    <2>5. LL2Disk.publicState \in PublicStateType
      BY <2>2 DEF LL2UntrustedStorageType
    <2>6. LL2Disk.privateStateEnc \in PrivateStateEncType
      BY <2>2 DEF LL2UntrustedStorageType
    <2>7. LL2Disk.historySummary \in HistorySummaryType
      BY <2>2 DEF LL2UntrustedStorageType
    <2>8. LL2Disk.historySummary.anchor \in HashType
      BY <2>7 DEF HistorySummaryType
    <2>9. LL2Disk.historySummary.extension \in HashType
      BY <2>7 DEF HistorySummaryType
    <2>10. LL2Disk.authenticator \in MACType
      BY <2>2 DEF LL2UntrustedStorageType
    <2>11. LL2RAM.publicState \in PublicStateType
      BY <2>3 DEF LL2UntrustedStorageType
    <2>12. LL2RAM.privateStateEnc \in PrivateStateEncType
      BY <2>3 DEF LL2UntrustedStorageType
    <2>13. LL2RAM.historySummary \in HistorySummaryType
      BY <2>3 DEF LL2UntrustedStorageType
    <2>14. LL2RAM.historySummary.anchor \in HashType
      BY <2>13 DEF HistorySummaryType
    <2>15. LL2RAM.historySummary.extension \in HashType
      BY <2>13 DEF HistorySummaryType
    <2>16. LL2RAM.authenticator \in MACType
      BY <2>3 DEF LL2UntrustedStorageType
    <2>17. LL2NVRAM.historySummaryAnchor \in HashType
      BY <2>4 DEF LL2TrustedStorageType
    <2>18. LL2NVRAM.symmetricKey \in SymmetricKeyType
      BY <2>4 DEF LL2TrustedStorageType
    <2>19. LL2NVRAM.hashBarrier \in HashType
      BY <2>4 DEF LL2TrustedStorageType
    <2>20. LL2NVRAM.extensionInProgress \in BOOLEAN
      BY <2>4 DEF LL2TrustedStorageType
    <2>21. QED
      BY <2>5, <2>6, <2>7, <2>8, <2>9, <2>10, <2>11, <2>12,
         <2>13, <2>14, <2>15, <2>16, <2>17, <2>18, <2>19, <2>20
  <1>2. LL2SubtypeImplication'
    <2>1. SUFFICES
          ASSUME LL2TypeInvariant'
          PROVE
            /\  LL2Disk.publicState' \in PublicStateType
            /\  LL2Disk.privateStateEnc' \in PrivateStateEncType
            /\  LL2Disk.historySummary' \in HistorySummaryType
            /\  LL2Disk.historySummary.anchor' \in HashType
            /\  LL2Disk.historySummary.extension' \in HashType
            /\  LL2Disk.authenticator' \in MACType
            /\  LL2RAM.publicState' \in PublicStateType
            /\  LL2RAM.privateStateEnc' \in PrivateStateEncType
            /\  LL2RAM.historySummary' \in HistorySummaryType
            /\  LL2RAM.historySummary.anchor' \in HashType
            /\  LL2RAM.historySummary.extension' \in HashType
            /\  LL2RAM.authenticator' \in MACType
            /\  LL2NVRAM.historySummaryAnchor' \in HashType
            /\  LL2NVRAM.symmetricKey' \in SymmetricKeyType
            /\  LL2NVRAM.hashBarrier' \in HashType
            /\  LL2NVRAM.extensionInProgress' \in BOOLEAN
      BY DEF LL2SubtypeImplication, LL2TypeInvariant
    <2>2. LL2Disk' \in LL2UntrustedStorageType
      BY <2>1 DEF LL2TypeInvariant
    <2>3. LL2RAM' \in LL2UntrustedStorageType
      BY <2>1 DEF LL2TypeInvariant
    <2>4. LL2NVRAM' \in LL2TrustedStorageType
      BY <2>1 DEF LL2TypeInvariant
    <2>5. LL2Disk.publicState' \in PublicStateType
      BY <2>2 DEF LL2UntrustedStorageType
    <2>6. LL2Disk.privateStateEnc' \in PrivateStateEncType
      BY <2>2 DEF LL2UntrustedStorageType
    <2>7. LL2Disk.historySummary' \in HistorySummaryType
      BY <2>2 DEF LL2UntrustedStorageType
    <2>8. LL2Disk.historySummary.anchor' \in HashType
      BY <2>7 DEF HistorySummaryType
    <2>9. LL2Disk.historySummary.extension' \in HashType
      BY <2>7 DEF HistorySummaryType
    <2>10. LL2Disk.authenticator' \in MACType
      BY <2>2 DEF LL2UntrustedStorageType
    <2>11. LL2RAM.publicState' \in PublicStateType
      BY <2>3 DEF LL2UntrustedStorageType
    <2>12. LL2RAM.privateStateEnc' \in PrivateStateEncType
      BY <2>3 DEF LL2UntrustedStorageType
    <2>13. LL2RAM.historySummary' \in HistorySummaryType
      BY <2>3 DEF LL2UntrustedStorageType
    <2>14. LL2RAM.historySummary.anchor' \in HashType
      BY <2>13 DEF HistorySummaryType
    <2>15. LL2RAM.historySummary.extension' \in HashType
      BY <2>13 DEF HistorySummaryType
    <2>16. LL2RAM.authenticator' \in MACType
      BY <2>3 DEF LL2UntrustedStorageType
    <2>17. LL2NVRAM.historySummaryAnchor' \in HashType
      BY <2>4 DEF LL2TrustedStorageType
    <2>18. LL2NVRAM.symmetricKey' \in SymmetricKeyType
      BY <2>4 DEF LL2TrustedStorageType
    <2>19. LL2NVRAM.hashBarrier' \in HashType
      BY <2>4 DEF LL2TrustedStorageType
    <2>20. LL2NVRAM.extensionInProgress' \in BOOLEAN
      BY <2>4 DEF LL2TrustedStorageType
    <2>21. QED
      BY <2>5, <2>6, <2>7, <2>8, <2>9, <2>10, <2>11, <2>12,
         <2>13, <2>14, <2>15, <2>16, <2>17, <2>18, <2>19, <2>20
  <1>3. QED
    BY <1>1, <1>2

(***************************************************************************)
(* The LL2InitDefsTypeSafeLemma proves that the definitions within the LET *)
(* of the LL2Init action all have the appropriate type.  This is a trivial *)
(* proof that merely walks through the definitions.                        *)
(***************************************************************************)

THEOREM LL2InitDefsTypeSafeLemma == \*`^\label{THEOREM-LL2InitDefsTypeSafeLemma}^'
    \A symmetricKey \in SymmetricKeyType, hashBarrier \in HashType :
        LET
            initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
            initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
            initialHistorySummary == [
                anchor |-> BaseHashValue,
                extension |-> BaseHashValue ]
            initialHistorySummaryHash == Hash(BaseHashValue, BaseHashValue)
            initialHistoryStateBinding == Hash(initialHistorySummaryHash, initialStateHash)
            initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
            initialUntrustedStorage == [
                publicState |-> InitialPublicState,
                privateStateEnc |-> initialPrivateStateEnc,
                historySummary |-> initialHistorySummary,
                authenticator |-> initialAuthenticator ]
            initialTrustedStorage == [
                historySummaryAnchor |-> BaseHashValue,
                symmetricKey |-> symmetricKey,
                hashBarrier |-> hashBarrier,
                extensionInProgress |-> FALSE ]
        IN
            /\  initialPrivateStateEnc \in PrivateStateEncType
            /\  initialStateHash \in HashType
            /\  initialHistorySummary \in HistorySummaryType
            /\  initialHistorySummaryHash \in HashType
            /\  initialHistoryStateBinding \in HashType
            /\  initialAuthenticator \in MACType
            /\  initialUntrustedStorage \in LL2UntrustedStorageType
            /\  initialTrustedStorage \in LL2TrustedStorageType
  <1>1. TAKE symmetricKey \in SymmetricKeyType, hashBarrier \in HashType
  <1> initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
  <1> initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
  <1> initialHistorySummary == [
          anchor |-> BaseHashValue,
          extension |-> BaseHashValue ]
  <1> initialHistorySummaryHash == Hash(BaseHashValue, BaseHashValue)
  <1> initialHistoryStateBinding == Hash(initialHistorySummaryHash, initialStateHash)
  <1> initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
  <1> initialUntrustedStorage == [
          publicState |-> InitialPublicState,
          privateStateEnc |-> initialPrivateStateEnc,
          historySummary |-> initialHistorySummary,
          authenticator |-> initialAuthenticator ]
  <1> initialTrustedStorage == [
          historySummaryAnchor |-> BaseHashValue,
          symmetricKey |-> symmetricKey,
          hashBarrier |-> hashBarrier,
          extensionInProgress |-> FALSE ]
  <1> HIDE DEF initialPrivateStateEnc, initialStateHash, initialHistorySummary,
          initialHistorySummaryHash, initialHistoryStateBinding, initialAuthenticator,
          initialUntrustedStorage, initialTrustedStorage
  <1>2. initialPrivateStateEnc \in PrivateStateEncType
    <2>1. symmetricKey \in SymmetricKeyType
      BY <1>1
    <2>2. InitialPrivateState \in PrivateStateType
      BY ConstantsTypeSafe
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF initialPrivateStateEnc
  <1>3. initialStateHash \in HashType
    <2>1. InitialPublicState \in HashDomain
      <3>1. InitialPublicState \in PublicStateType
        BY ConstantsTypeSafe
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. initialPrivateStateEnc \in HashDomain
      BY <1>2 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF initialStateHash
  <1>4. initialHistorySummary \in HistorySummaryType
    <2>1. BaseHashValue \in HashType
      BY BaseHashValueTypeSafe
    <2>2. QED
      BY <2>1 DEF initialHistorySummary, HistorySummaryType
  <1>5. initialHistorySummaryHash \in HashType
    <2>1. BaseHashValue \in HashDomain
      <3>1. BaseHashValue \in HashType
        BY BaseHashValueTypeSafe
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. QED
      BY <2>1, HashTypeSafe DEF initialHistorySummaryHash
  <1>6. initialHistoryStateBinding \in HashType
    <2>1. initialHistorySummaryHash \in HashDomain
      BY <1>5 DEF HashDomain
    <2>2. initialStateHash \in HashDomain
      BY <1>3 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF initialHistoryStateBinding
  <1>7. initialAuthenticator \in MACType
    <2>1. symmetricKey \in SymmetricKeyType
      BY <1>1
    <2>2. initialHistoryStateBinding \in HashType
      BY <1>6
    <2>3. QED
      BY <2>1, <2>2, GenerateMACTypeSafe DEF initialAuthenticator
  <1>8. initialUntrustedStorage \in LL2UntrustedStorageType
    <2>1. InitialPublicState \in PublicStateType
      BY ConstantsTypeSafe
    <2>2. initialPrivateStateEnc \in PrivateStateEncType
      BY <1>2
    <2>3. initialHistorySummary \in HistorySummaryType
      BY <1>4
    <2>4. initialAuthenticator \in MACType
      BY <1>7
    <2>5. QED
      BY <2>1, <2>2, <2>3, <2>4 DEF initialUntrustedStorage, LL2UntrustedStorageType
  <1>9. initialTrustedStorage \in LL2TrustedStorageType
    <2>1. BaseHashValue \in HashType
      BY BaseHashValueTypeSafe
    <2>2. symmetricKey \in SymmetricKeyType
      BY <1>1
    <2>3. hashBarrier \in HashType
      BY <1>1
    <2>4. FALSE \in BOOLEAN
      OBVIOUS
    <2>5. QED
      BY <2>1, <2>2, <2>3, <2>4 DEF initialTrustedStorage, LL2TrustedStorageType
  <1>10. QED
    BY <1>2, <1>3, <1>4, <1>5, <1>6, <1>7, <1>8, <1>9
    DEF initialPrivateStateEnc, initialStateHash, initialHistorySummary,
        initialHistorySummaryHash, initialHistoryStateBinding, initialAuthenticator,
        initialUntrustedStorage, initialTrustedStorage

(***************************************************************************)
(* The LL2PerformOperationDefsTypeSafeLemma proves that the definitions    *)
(* within the LET of the LL2PerformOperation action all have the           *)
(* appropriate type.  This is a trivial proof that merely walks through    *)
(* the definitions.                                                        *)
(***************************************************************************)

THEOREM LL2PerformOperationDefsTypeSafeLemma == \*`^\label{THEOREM-LL2PerformOperationDefsTypeSafeLemma}^'
    \A input \in LL2AvailableInputs :
        LL2TypeInvariant =>
            LET
                historySummaryHash ==
                    Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
                stateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
                historyStateBinding == Hash(historySummaryHash, stateHash)
                privateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
                sResult == Service(LL2RAM.publicState, privateState, input)
                newPrivateStateEnc ==
                    SymmetricEncrypt(LL2NVRAM.symmetricKey, sResult.newPrivateState)
                currentHistorySummary == [
                    anchor |-> LL2NVRAM.historySummaryAnchor,
                    extension |-> LL2SPCR ]
                newHistorySummary == Successor(currentHistorySummary, input, LL2NVRAM.hashBarrier)
                newHistorySummaryHash ==
                    Hash(newHistorySummary.anchor, newHistorySummary.extension)
                newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
                newHistoryStateBinding == Hash(newHistorySummaryHash, newStateHash)
                newAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, newHistoryStateBinding)
            IN
                /\  historySummaryHash \in HashType
                /\  stateHash \in HashType
                /\  historyStateBinding \in HashType
                /\  privateState \in PrivateStateType
                /\  sResult \in ServiceResultType
                /\  sResult.newPublicState \in PublicStateType
                /\  sResult.newPrivateState \in PrivateStateType
                /\  sResult.output \in OutputType
                /\  newPrivateStateEnc \in PrivateStateEncType
                /\  currentHistorySummary \in HistorySummaryType
                /\  currentHistorySummary.anchor \in HashType
                /\  currentHistorySummary.extension \in HashType
                /\  newHistorySummary \in HistorySummaryType
                /\  newHistorySummary.anchor \in HashType
                /\  newHistorySummary.extension \in HashType
                /\  newHistorySummaryHash \in HashType
                /\  newStateHash \in HashType
                /\  newHistoryStateBinding \in HashType
                /\  newAuthenticator \in MACType
  <1>1. TAKE input \in LL2AvailableInputs
  <1> historySummaryHash == Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
  <1> stateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
  <1> historyStateBinding == Hash(historySummaryHash, stateHash)
  <1> privateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
  <1> sResult == Service(LL2RAM.publicState, privateState, input)
  <1> newPrivateStateEnc ==
          SymmetricEncrypt(LL2NVRAM.symmetricKey, sResult.newPrivateState)
  <1> currentHistorySummary == [
          anchor |-> LL2NVRAM.historySummaryAnchor,
          extension |-> LL2SPCR ]
  <1> newHistorySummary == Successor(currentHistorySummary, input, LL2NVRAM.hashBarrier)
  <1> newHistorySummaryHash == Hash(newHistorySummary.anchor, newHistorySummary.extension)
  <1> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
  <1> newHistoryStateBinding == Hash(newHistorySummaryHash, newStateHash)
  <1> newAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, newHistoryStateBinding)
  <1> HIDE DEF historySummaryHash, stateHash, historyStateBinding, privateState,
          sResult, newPrivateStateEnc, currentHistorySummary, newHistorySummary,
          newHistorySummaryHash, newStateHash, newHistoryStateBinding, newAuthenticator
  <1>2. HAVE LL2TypeInvariant
  <1>3. historySummaryHash \in HashType
    <2>1. LL2RAM.historySummary.anchor \in HashDomain
      <3>1. LL2RAM.historySummary.anchor \in HashType
        BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. LL2RAM.historySummary.extension \in HashDomain
      <3>1. LL2RAM.historySummary.extension \in HashType
        BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF historySummaryHash
  <1>4. stateHash \in HashType
    <2>1. /\  LL2RAM.publicState \in PublicStateType
          /\  LL2RAM.privateStateEnc \in PrivateStateEncType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. /\  LL2RAM.publicState \in HashDomain
          /\  LL2RAM.privateStateEnc \in HashDomain
      BY <2>1 DEF HashDomain
    <2>3. QED
      BY <2>2, HashTypeSafe DEF stateHash
  <1>5. historyStateBinding \in HashType
    <2>1. historySummaryHash \in HashDomain
      BY <1>3 DEF HashDomain
    <2>2. stateHash \in HashDomain
      BY <1>4 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF historyStateBinding
  <1>6. privateState \in PrivateStateType
    <2>1. /\  LL2NVRAM.symmetricKey \in SymmetricKeyType
          /\  LL2RAM.privateStateEnc \in PrivateStateEncType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. QED
      BY <2>1, SymmetricDecryptionTypeSafe DEF privateState
  <1>7. sResult \in ServiceResultType
    <2>1. LL2RAM.publicState \in PublicStateType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. privateState \in PrivateStateType
      BY <1>6
    <2>3. input \in InputType
      <3>1. LL2AvailableInputs \subseteq InputType
        BY <1>2 DEF LL2TypeInvariant
      <3>2. QED
        BY <1>1, <3>1
    <2>4. QED
      BY <2>1, <2>2, <2>3, ServiceTypeSafe DEF sResult
  <1>8. /\  sResult.newPublicState \in PublicStateType
        /\  sResult.newPrivateState \in PrivateStateType
        /\  sResult.output \in OutputType
    BY <1>7 DEF ServiceResultType
  <1>9. newPrivateStateEnc \in PrivateStateEncType
    <2>1. LL2NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. sResult.newPrivateState \in PrivateStateType
      BY <1>8
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF newPrivateStateEnc
  <1>10. currentHistorySummary \in HistorySummaryType
    <2>1. LL2NVRAM.historySummaryAnchor \in HashType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. LL2SPCR \in HashType
      BY <1>2 DEF LL2TypeInvariant
    <2>3. QED
      BY <2>1, <2>2 DEF currentHistorySummary, HistorySummaryType
  <1>11. /\  currentHistorySummary.anchor \in HashType
         /\  currentHistorySummary.extension \in HashType
    BY <1>10 DEF HistorySummaryType
  <1>12. newHistorySummary \in HistorySummaryType
    <2>1. currentHistorySummary \in HistorySummaryType
      BY <1>10
    <2>2. input \in InputType
      <3>1. input \in LL2AvailableInputs
        BY <1>1
      <3>2. LL2AvailableInputs \subseteq InputType
        BY <1>2 DEF LL2TypeInvariant
      <3>3. QED
        BY <3>1, <3>2
    <2>3. LL2NVRAM.hashBarrier \in HashType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>4. QED
      BY <2>1, <2>2, <2>3, SuccessorTypeSafe DEF newHistorySummary
  <1>13. /\  newHistorySummary.anchor \in HashType
         /\  newHistorySummary.extension \in HashType
    BY <1>12 DEF HistorySummaryType
  <1>14. newHistorySummaryHash \in HashType
    <2>1. newHistorySummary.anchor \in HashDomain
      <3>1. newHistorySummary.anchor \in HashType
        BY <1>12 DEF HistorySummaryType
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. newHistorySummary.extension \in HashDomain
      <3>1. newHistorySummary.extension \in HashType
        BY <1>12 DEF HistorySummaryType
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistorySummaryHash
  <1>15. newStateHash \in HashType
    <2>1. sResult.newPublicState \in HashDomain
      <3>1. sResult.newPublicState \in PublicStateType
        BY <1>8
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. newPrivateStateEnc \in HashDomain
      BY <1>9 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newStateHash
  <1>16. newHistoryStateBinding \in HashType
    <2>1. newHistorySummaryHash \in HashDomain
      BY <1>14 DEF HashDomain
    <2>2. newStateHash \in HashDomain
      BY <1>15 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistoryStateBinding
  <1>17. newAuthenticator \in MACType
    <2>1. LL2NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. newHistoryStateBinding \in HashType
      BY <1>16
    <2>3. QED
      BY <2>1, <2>2, GenerateMACTypeSafe DEF newAuthenticator
  <1>18. QED
    BY <1>3, <1>4, <1>5, <1>6, <1>7, <1>8, <1>9,
       <1>10, <1>11, <1>12, <1>13, <1>14, <1>15, <1>16, <1>17
    DEF historySummaryHash, stateHash, historyStateBinding, privateState,
        sResult, newPrivateStateEnc, currentHistorySummary, newHistorySummary,
        newHistorySummaryHash, newStateHash, newHistoryStateBinding, newAuthenticator

(***************************************************************************)
(* The LL2RepeatOperationDefsTypeSafeLemma proves that the definitions     *)
(* within the LET of the LL2RepeatOperation action all have the            *)
(* appropriate type.  This is a trivial proof that merely walks through    *)
(* the definitions.                                                        *)
(***************************************************************************)

THEOREM LL2RepeatOperationDefsTypeSafeLemma == \*`^\label{THEOREM-LL2RepeatOperationDefsTypeSafeLemma}^'
    \A input \in LL2AvailableInputs :
        LL2TypeInvariant =>
            LET
                historySummaryHash ==
                    Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
                stateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
                historyStateBinding == Hash(historySummaryHash, stateHash)
                newHistorySummary ==
                    Successor(LL2RAM.historySummary, input, LL2NVRAM.hashBarrier)
                checkpointedHistorySummary == Checkpoint(LL2RAM.historySummary)
                newCheckpointedHistorySummary ==
                    Successor(checkpointedHistorySummary, input, LL2NVRAM.hashBarrier)
                checkpointedNewHistorySummary == Checkpoint(newHistorySummary)
                checkpointedNewCheckpointedHistorySummary ==
                    Checkpoint(newCheckpointedHistorySummary)
                privateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
                sResult == Service(LL2RAM.publicState, privateState, input)
                newPrivateStateEnc ==
                    SymmetricEncrypt(LL2NVRAM.symmetricKey, sResult.newPrivateState)
                currentHistorySummary == [
                    anchor |-> LL2NVRAM.historySummaryAnchor,
                    extension |-> LL2SPCR ]
                currentHistorySummaryHash == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
                newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
                newHistoryStateBinding == Hash(currentHistorySummaryHash, newStateHash)
                newAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, newHistoryStateBinding)
            IN
                /\  historySummaryHash \in HashType
                /\  stateHash \in HashType
                /\  historyStateBinding \in HashType
                /\  newHistorySummary \in HistorySummaryType
                /\  newHistorySummary.anchor \in HashType
                /\  newHistorySummary.extension \in HashType
                /\  checkpointedHistorySummary \in HistorySummaryType
                /\  checkpointedHistorySummary.anchor \in HashType
                /\  checkpointedHistorySummary.extension \in HashType
                /\  newCheckpointedHistorySummary \in HistorySummaryType
                /\  newCheckpointedHistorySummary.anchor \in HashType
                /\  newCheckpointedHistorySummary.extension \in HashType
                /\  checkpointedNewHistorySummary \in HistorySummaryType
                /\  checkpointedNewHistorySummary.anchor \in HashType
                /\  checkpointedNewHistorySummary.extension \in HashType
                /\  checkpointedNewCheckpointedHistorySummary \in HistorySummaryType
                /\  checkpointedNewCheckpointedHistorySummary.anchor \in HashType
                /\  checkpointedNewCheckpointedHistorySummary.extension \in HashType
                /\  privateState \in PrivateStateType
                /\  sResult \in ServiceResultType
                /\  sResult.newPublicState \in PublicStateType
                /\  sResult.newPrivateState \in PrivateStateType
                /\  sResult.output \in OutputType
                /\  newPrivateStateEnc \in PrivateStateEncType
                /\  currentHistorySummary \in HistorySummaryType
                /\  currentHistorySummary.anchor \in HashType
                /\  currentHistorySummary.extension \in HashType
                /\  currentHistorySummaryHash \in HashType
                /\  newStateHash \in HashType
                /\  newHistoryStateBinding \in HashType
                /\  newAuthenticator \in MACType
  <1>1. TAKE input \in LL2AvailableInputs
  <1> historySummaryHash ==
          Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
  <1> stateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
  <1> historyStateBinding == Hash(historySummaryHash, stateHash)
  <1> newHistorySummary == Successor(LL2RAM.historySummary, input, LL2NVRAM.hashBarrier)
  <1> checkpointedHistorySummary == Checkpoint(LL2RAM.historySummary)
  <1> newCheckpointedHistorySummary ==
          Successor(checkpointedHistorySummary, input, LL2NVRAM.hashBarrier)
  <1> checkpointedNewHistorySummary == Checkpoint(newHistorySummary)
  <1> checkpointedNewCheckpointedHistorySummary ==
          Checkpoint(newCheckpointedHistorySummary)
  <1> privateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
  <1> sResult == Service(LL2RAM.publicState, privateState, input)
  <1> newPrivateStateEnc ==
          SymmetricEncrypt(LL2NVRAM.symmetricKey, sResult.newPrivateState)
  <1> currentHistorySummary == [
          anchor |-> LL2NVRAM.historySummaryAnchor,
          extension |-> LL2SPCR ]
  <1> currentHistorySummaryHash == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
  <1> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
  <1> newHistoryStateBinding == Hash(currentHistorySummaryHash, newStateHash)
  <1> newAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, newHistoryStateBinding)
  <1> HIDE DEF historySummaryHash, stateHash, historyStateBinding, newHistorySummary,
          checkpointedHistorySummary, newCheckpointedHistorySummary,
          checkpointedNewHistorySummary, checkpointedNewCheckpointedHistorySummary,
          privateState, sResult, newPrivateStateEnc, currentHistorySummary,
          currentHistorySummaryHash, newStateHash, newHistoryStateBinding,
          newAuthenticator
  <1>2. HAVE LL2TypeInvariant
  <1>3. input \in InputType
    <2>1. input \in LL2AvailableInputs
      BY <1>1
    <2>2. LL2AvailableInputs \subseteq InputType
      BY <1>2 DEF LL2TypeInvariant
    <2>3. QED
      BY <2>1, <2>2
  <1>4. LL2NVRAM.hashBarrier \in HashType
    BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
  <1>5. historySummaryHash \in HashType
    <2>1. LL2RAM.historySummary.anchor \in HashDomain
      <3>1. LL2RAM.historySummary.anchor \in HashType
        BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. LL2RAM.historySummary.extension \in HashDomain
      <3>1. LL2RAM.historySummary.extension \in HashType
        BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF historySummaryHash
  <1>6. stateHash \in HashType
    <2>1. /\  LL2RAM.publicState \in PublicStateType
          /\  LL2RAM.privateStateEnc \in PrivateStateEncType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. /\  LL2RAM.publicState \in HashDomain
          /\  LL2RAM.privateStateEnc \in HashDomain
      BY <2>1 DEF HashDomain
    <2>3. QED
      BY <2>2, HashTypeSafe DEF stateHash
  <1>7. historyStateBinding \in HashType
    <2>1. historySummaryHash \in HashDomain
      BY <1>5 DEF HashDomain
    <2>2. stateHash \in HashDomain
      BY <1>6 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF historyStateBinding
  <1>8. newHistorySummary \in HistorySummaryType
    <2>1. LL2RAM.historySummary \in HistorySummaryType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. QED
      BY <1>3, <1>4, <2>1, SuccessorTypeSafe DEF newHistorySummary
  <1>9. /\  newHistorySummary.anchor \in HashType
         /\  newHistorySummary.extension \in HashType
    BY <1>8 DEF HistorySummaryType
  <1>10. checkpointedHistorySummary \in HistorySummaryType
    <2>1. LL2RAM.historySummary \in HistorySummaryType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. QED
      BY <2>1, CheckpointTypeSafe DEF checkpointedHistorySummary
  <1>11. /\  checkpointedHistorySummary.anchor \in HashType
         /\  checkpointedHistorySummary.extension \in HashType
    BY <1>10 DEF HistorySummaryType
  <1>12. newCheckpointedHistorySummary \in HistorySummaryType
    BY <1>3, <1>4, <1>10, SuccessorTypeSafe DEF newCheckpointedHistorySummary
  <1>13. /\  newCheckpointedHistorySummary.anchor \in HashType
         /\  newCheckpointedHistorySummary.extension \in HashType
    BY <1>12 DEF HistorySummaryType
  <1>14. checkpointedNewHistorySummary \in HistorySummaryType
      BY <1>8, CheckpointTypeSafe DEF checkpointedNewHistorySummary
  <1>15. /\  checkpointedNewHistorySummary.anchor \in HashType
         /\  checkpointedNewHistorySummary.extension \in HashType
    BY <1>14 DEF HistorySummaryType
  <1>16. checkpointedNewCheckpointedHistorySummary \in HistorySummaryType
      BY <1>12, CheckpointTypeSafe DEF checkpointedNewCheckpointedHistorySummary
  <1>17. /\  checkpointedNewCheckpointedHistorySummary.anchor \in HashType
         /\  checkpointedNewCheckpointedHistorySummary.extension \in HashType
    BY <1>16 DEF HistorySummaryType
  <1>18. privateState \in PrivateStateType
    <2>1. /\  LL2NVRAM.symmetricKey \in SymmetricKeyType
          /\  LL2RAM.privateStateEnc \in PrivateStateEncType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. QED
      BY <2>1, SymmetricDecryptionTypeSafe DEF privateState
  <1>19. sResult \in ServiceResultType
    <2>1. LL2RAM.publicState \in PublicStateType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. privateState \in PrivateStateType
      BY <1>18
    <2>3. input \in InputType
      <3>1. LL2AvailableInputs \subseteq InputType
        BY <1>2 DEF LL2TypeInvariant
      <3>2. QED
        BY <1>1, <3>1
    <2>4. QED
      BY <2>1, <2>2, <2>3, ServiceTypeSafe DEF sResult
  <1>20. /\  sResult.newPublicState \in PublicStateType
         /\  sResult.newPrivateState \in PrivateStateType
         /\  sResult.output \in OutputType
    BY <1>19 DEF ServiceResultType
  <1>21. newPrivateStateEnc \in PrivateStateEncType
    <2>1. LL2NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. sResult.newPrivateState \in PrivateStateType
      BY <1>20
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF newPrivateStateEnc
  <1>22. currentHistorySummary \in HistorySummaryType
    <2>1. LL2NVRAM.historySummaryAnchor \in HashType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. LL2SPCR \in HashType
      BY <1>2 DEF LL2TypeInvariant
    <2>3. QED
      BY <2>1, <2>2 DEF currentHistorySummary, HistorySummaryType
  <1>23. /\  currentHistorySummary.anchor \in HashType
         /\  currentHistorySummary.extension \in HashType
    BY <1>22 DEF HistorySummaryType
  <1>24. currentHistorySummaryHash \in HashType
    <2>1. LL2NVRAM.historySummaryAnchor \in HashDomain
      <3>1. LL2NVRAM.historySummaryAnchor \in HashType
        BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. LL2SPCR \in HashDomain
      <3>1. LL2SPCR \in HashType
        BY <1>2 DEF LL2TypeInvariant
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF currentHistorySummaryHash
  <1>25. newStateHash \in HashType
    <2>1. sResult.newPublicState \in HashDomain
      <3>1. sResult.newPublicState \in PublicStateType
        BY <1>20
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. newPrivateStateEnc \in HashDomain
      BY <1>21 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newStateHash
  <1>26. newHistoryStateBinding \in HashType
    <2>1. currentHistorySummaryHash \in HashDomain
      BY <1>24 DEF HashDomain
    <2>2. newStateHash \in HashDomain
      BY <1>25 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistoryStateBinding
  <1>27. newAuthenticator \in MACType
    <2>1. LL2NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
    <2>2. newHistoryStateBinding \in HashType
      BY <1>26
    <2>3. QED
      BY <2>1, <2>2, GenerateMACTypeSafe DEF newAuthenticator
  <1>28. QED
    BY <1>5, <1>6, <1>7, <1>8, <1>9, <1>10, <1>11, <1>12, <1>13, <1>14, <1>15,
       <1>16, <1>17, <1>18, <1>19, <1>20, <1>21, <1>22, <1>23, <1>24, <1>25, <1>26, <1>27
    DEF historySummaryHash, stateHash, historyStateBinding, newHistorySummary,
        checkpointedHistorySummary, newCheckpointedHistorySummary,
        checkpointedNewHistorySummary, checkpointedNewCheckpointedHistorySummary,
        privateState, sResult, newPrivateStateEnc, currentHistorySummary,
        currentHistorySummaryHash, newStateHash, newHistoryStateBinding,
        newAuthenticator

----

(***************************************************************************)
(* The LL2TakeCheckpointDefsTypeSafeLemma proves that the definition       *)
(* within the LET of the LL2TakeCheckpoint action has the appropriate      *)
(* type.  This is a trivial proof that merely walks through the            *)
(* definitions.                                                            *)
(***************************************************************************)

THEOREM LL2TakeCheckpointDefsTypeSafeLemma == \*`^\label{THEOREM-LL2TakeCheckpointDefsTypeSafeLemma}^'
    LL2TypeInvariant =>
        LET
            newHistorySummaryAnchor == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
        IN
            newHistorySummaryAnchor \in HashType
  <1> newHistorySummaryAnchor == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
  <1> HIDE DEF newHistorySummaryAnchor
  <1>1. HAVE LL2TypeInvariant
  <1>2. newHistorySummaryAnchor \in HashType
    <2>1. LL2NVRAM.historySummaryAnchor \in HashDomain
      <3>1. LL2NVRAM.historySummaryAnchor \in HashType
        BY <1>1, LL2SubtypeImplicationLemma DEF LL2SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. LL2SPCR \in HashDomain
      <3>1. LL2SPCR \in HashType
        BY <1>1 DEF LL2TypeInvariant
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistorySummaryAnchor
  <1>3. QED
    BY <1>2 DEF newHistorySummaryAnchor

----

(***************************************************************************)
(* The LL2CorruptSPCRDefsTypeSafeLemma proves that the definition within   *)
(* the LET of the LL2CorruptSPCR action has the appropriate type.  This is *)
(* a trivial proof that merely walks through the definitions.              *)
(***************************************************************************)

THEOREM LL2CorruptSPCRDefsTypeSafeLemma == \*`^\label{THEOREM-LL2CorruptSPCRDefsTypeSafeLemma}^'
    \A fakeHash \in HashDomain :
        LL2TypeInvariant =>
            LET
                newHistorySummaryExtension == Hash(LL2SPCR, fakeHash)
            IN
                newHistorySummaryExtension \in HashType
  <1>1. TAKE fakeHash \in HashDomain
  <1> newHistorySummaryExtension == Hash(LL2SPCR, fakeHash)
  <1> HIDE DEF newHistorySummaryExtension
  <1>2. HAVE LL2TypeInvariant
  <1>3. newHistorySummaryExtension \in HashType
    <2>1. LL2SPCR \in HashDomain
      <3>1. LL2SPCR \in HashType
        BY <1>2 DEF LL2TypeInvariant
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. fakeHash \in HashDomain
      BY <1>1
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistorySummaryExtension
  <1>4. QED
    BY <1>3 DEF newHistorySummaryExtension

----

(***************************************************************************)
(* The AuthenticatorsMatchDefsTypeSafeLemma proves that the definitions    *)
(* within the LET of the AuthenticatorsMatch predicate have the            *)
(* appropriate type.  This is a trivial proof that merely walks through    *)
(* the definitions.                                                        *)
(***************************************************************************)

THEOREM AuthenticatorsMatchDefsTypeSafeLemma == \*`^\label{THEOREM-AuthenticatorsMatchDefsTypeSafeLemma}^'
    \A  stateHash \in HashType,
        ll1HistorySummary \in HashType,
        ll2HistorySummary \in HistorySummaryType :
        LET
            ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash)
            ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
            ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash)
        IN
            /\  ll1HistoryStateBinding \in HashType
            /\  ll2HistorySummaryHash \in HashType
            /\  ll2HistoryStateBinding \in HashType
  <1>1. TAKE stateHash \in HashType,
             ll1HistorySummary \in HashType,
             ll2HistorySummary \in HistorySummaryType
  <1> ll1HistoryStateBinding == Hash(ll1HistorySummary, stateHash)
  <1> ll2HistorySummaryHash == Hash(ll2HistorySummary.anchor, ll2HistorySummary.extension)
  <1> ll2HistoryStateBinding == Hash(ll2HistorySummaryHash, stateHash)
  <1> HIDE DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding
  <1>2. ll1HistoryStateBinding \in HashType
    <2>1. ll1HistorySummary \in HashDomain
      <3>1. ll1HistorySummary \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. stateHash \in HashDomain
      <3>1. stateHash \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF ll1HistoryStateBinding
  <1>3. ll2HistorySummaryHash \in HashType
    <2>1. ll2HistorySummary \in HistorySummaryType
      BY <1>1
    <2>2. ll2HistorySummary.anchor \in HashDomain
      <3>1. ll2HistorySummary.anchor \in HashType
        BY <2>1 DEF HistorySummaryType
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. ll2HistorySummary.extension \in HashDomain
      <3>1. ll2HistorySummary.extension \in HashType
        BY <2>1 DEF HistorySummaryType
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>4. QED
      BY <2>2, <2>3, HashTypeSafe DEF ll2HistorySummaryHash
  <1>4. ll2HistoryStateBinding \in HashType
    <2>1. ll2HistorySummaryHash \in HashDomain
      <3>1. ll2HistorySummaryHash \in HashType
        BY <1>3
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. stateHash \in HashDomain
      <3>1. stateHash \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF ll2HistoryStateBinding
  <1>5. QED
    BY <1>2, <1>3, <1>4
    DEF ll1HistoryStateBinding, ll2HistorySummaryHash, ll2HistoryStateBinding

----

(***************************************************************************)
(* The TypeSafetyRefinementLemma states that if the `^\fastname^' type     *)
(* invariant holds, and the refinement holds, then the `^\basicname^' type *)
(* invariant holds.                                                        *)
(***************************************************************************)

THEOREM TypeSafetyRefinementLemma == \*`^\label{THEOREM-TypeSafetyRefinementLemma}^'
    LL2TypeInvariant /\ LL2Refinement => LL1TypeInvariant
  <1>1. HAVE LL2TypeInvariant /\ LL2Refinement
  <1>2. LL1AvailableInputs \subseteq InputType
    <2>1. LL1AvailableInputs = LL2AvailableInputs
      BY <1>1 DEF LL2Refinement
    <2>2. LL2AvailableInputs \subseteq InputType
      BY <1>1 DEF LL2TypeInvariant
    <2>3. QED
      BY <2>1, <2>2
  <1>3. LL1ObservedOutputs \subseteq OutputType
    <2>1. LL1ObservedOutputs = LL2ObservedOutputs
      BY <1>1 DEF LL2Refinement
    <2>2. LL2ObservedOutputs \subseteq OutputType
      BY <1>1 DEF LL2TypeInvariant
    <2>3. QED
      BY <2>1, <2>2
  <1>4. LL1ObservedAuthenticators \subseteq MACType
    BY <1>1 DEF LL2Refinement
  <1>5. LL1Disk \in LL1UntrustedStorageType
    BY <1>1 DEF LL2Refinement
  <1>6. LL1RAM \in LL1UntrustedStorageType
    BY <1>1 DEF LL2Refinement
  <1>7. LL1NVRAM \in LL1TrustedStorageType
    BY <1>1 DEF LL2Refinement
  <1>8. QED
    BY <1>2, <1>3, <1>4, <1>5, <1>6, <1>7 DEF LL1TypeInvariant

====
